/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU General Public License                              */

/*
 * $Id: device_file.c,v 1.16 2002/06/15 15:16:14 mulix Exp $
 */ 

/* This code was initially based upon code from the book "Linux Device */
/* Drivers" by Alessandro Rubini and Jonathan Corbet, published by */
/* O'Reilly & Associates.*/

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/version.h>      /* module things */
#include <linux/module.h>       /* module things */
#include <linux/kernel.h>       /* printk() */
#include <linux/slab.h>         /* kmalloc() */
#include <linux/fs.h>           /* everything... */
#include <linux/errno.h>        /* error codes */
#include <linux/types.h>        /* size_t */
#include <linux/proc_fs.h>      /* /proc stuff */ 
#include <linux/fcntl.h>        /* O_ACCMODE */
#include <asm/atomic.h>         /* test_and_set_bit() */ 
#include <asm/uaccess.h>        /* copy_to/from_user */

#include "update_rules.h"
#include "rule_engine_public.h"
#include "device_file.h"
#include "module_interface.h"

#include "membuf.h"

#define SCT_MAJOR 0   /* dynamic major by default */

/* the major number we use */
int sct_major = SCT_MAJOR;

/* the device struct */
sct_dev_t sct_dev;

/* 1 is the device file is in use, 0 if we're free at the moment */ 
static volatile long sct_in_use; 

/* forward declarations */

void handle_request_packet(sct_dev_t* dev);

void build_response_packet(sct_membuf_t* respbuf, unsigned int packet_id, 
			   unsigned int command, int result_code, 
			   size_t datasz, unsigned char* data); 

/* does this request buffer contain a whole request packet, or not yet? */
int req_has_whole_packet(sct_membuf_t* req);

#define INITIAL_REQ_SIZE 16 * 1024

/* we can't use copy_to/from_user, as they're macros, so we wrap them */ 
unsigned long cp_to_user(void *to, const void *from, unsigned long n)
{
	return copy_to_user(to, from, n); 
}

unsigned long cp_from_user(void *to, const void *from, unsigned long n)
{
	return copy_from_user(to, from, n); 
}

/* utility functions */ 
int init_dev_file(sct_dev_t* dev)
{
	int retval; 
	/*
	 * Register your major, and accept a dynamic number
	 */
	retval = register_chrdev(sct_major, "sct_rules", &sct_fops);
	if (retval < 0)
		return retval; 
	
	if (sct_major == 0) 
		sct_major = retval; /* dynamic */
	
	sema_init(&dev->sem, 1);

	if (!(dev->req = sct_membuf_create(INITIAL_REQ_SIZE, cp_from_user, NULL)))
		goto unregister_chrdev; 

	if (!(dev->resp = sct_membuf_create(INITIAL_REQ_SIZE, NULL, cp_to_user))) 
		goto destroy_request; 

	return 0; 

 destroy_request:
	sct_membuf_destroy(dev->req); 

 unregister_chrdev:
	unregister_chrdev(sct_major, "sct_rules");
	
	return retval; 
}

void destroy_dev_file(sct_dev_t* dev)
{
	if (!dev)
		BUG(); 

	sct_membuf_destroy(dev->req); 
	sct_membuf_destroy(dev->resp); 
	unregister_chrdev(sct_major, "sct_rules");
}
/*
 * Open and close
 */
int sct_open (struct inode *inode, struct file *filp)
{
	/* only one process make talk to us. this is ok for the configuration   */
	/* device file, but we'll have to redo this to support multiple readers */
	/* for the logging device file                                          */
	if (test_and_set_bit(0, &sct_in_use))
		return -EBUSY; /* already open */ 

	if (!capable(CAP_SYS_PTRACE)) {
		clear_bit(0, &sct_in_use); 
		return -EPERM;
	}
	
	MOD_INC_USE_COUNT;

	sct_membuf_clear(sct_dev.req); 
	sct_membuf_clear(sct_dev.resp); 
	
	/* and use filp->private_data to point to the device data */
	filp->private_data = &sct_dev;
	
	return 0;          /* success */
}

int sct_release (struct inode *inode, struct file *filp)
{
	if (!sct_in_use)
		BUG(); 

	clear_bit(0, &sct_in_use); 

	MOD_DEC_USE_COUNT;
	return 0;
}


/*
 * Data management: read and write
 */

ssize_t 
sct_write(struct file *filp, const char *buf, size_t count,
	  loff_t *f_pos)
{
	sct_dev_t *dev = filp->private_data;
	ssize_t retval; 

	if (down_interruptible (&dev->sem))
		return -ERESTARTSYS;

	if ((retval = sct_membuf_add_chunk(dev->req, buf, count)) < 0) {
		retval = -EINVAL; 
		goto add_chunk_error;
	}
	
	if (!req_has_whole_packet(dev->req)) {
		printk(KERN_DEBUG "sct_write: whoohoo! got a whole packet!\n"); 
		handle_request_packet(dev);
	}

	up (&dev->sem);    
	return retval; 
	
 add_chunk_error:
	printk("sct_write: req_add_chunk returned '%d'\n", retval); 
	up (&dev->sem);
	return retval;
}

ssize_t 
sct_read (struct file *filp, char *buf, size_t count,
	  loff_t *f_pos)
{
	sct_dev_t *dev = filp->private_data;
	ssize_t retval = 0;
	
	if (down_interruptible (&dev->sem))
		return -ERESTARTSYS;
	
	retval = sct_membuf_get_chunk(dev->resp, buf, count);
	if (retval == -1)
		retval = -EINVAL; 

	up (&dev->sem);
	return retval; 
}

/*
 * The fops
 */

struct file_operations sct_fops = {
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0) )
        owner: THIS_MODULE, 
#endif
	read: sct_read,
	write: sct_write, 
	open: sct_open,
	release: sct_release
};
static inline void update_size(dev_req_t* req)
{
	if (!req)
		return; 

	req->size = sizeof(req->size) + 
		sizeof(req->id) + 
		sizeof(req->cmd) +
		sizeof(req->result) + 
		sizeof(req->data_sz) + 
		req->data_sz + 
		sizeof(req->magic);
}

void build_response_packet(sct_membuf_t* respbuf, unsigned int packet_id, 
			   unsigned int command, int result_code, 
			   size_t datasz, unsigned char* data)
{
	dev_req_t resp; 
	int retval; 

	resp.size = 0; /* calculated last */ 
	resp.id = packet_id; 
	resp.cmd = command | REPLY; 
	resp.result = result_code; 
	resp.data_sz = datasz; 
	resp.data = data; 
	resp.magic = DEV_FILE_MAGIC_COOKIE; 

	update_size(&resp); 

	printk(KERN_DEBUG "build_response_packet: returning a packet of size %d,  "
	       "buffer available space is %d\n", resp.size, 
	       sct_membuf_get_available(respbuf)); 
	/* TODO: enlarge the buffer if necessary */ 

	/* BETTER TODO: dont prepare the response in advance - 
	   feed it to uspace in chunks, as the reads come in. */ 
	if ((retval  = sct_membuf_add_chunk(respbuf, &resp.size, 
					    sizeof(resp.size))) < 0)
		goto failed; 

	if ((retval = sct_membuf_add_chunk(respbuf, &resp.id, 
					   sizeof(resp.id)) < 0))
		goto failed; 
	
	if ((retval = sct_membuf_add_chunk(respbuf, &resp.cmd, 
					   sizeof(resp.cmd))) < 0)
		goto failed; 
	
	if ((retval = sct_membuf_add_chunk(respbuf, &resp.result, 
					   sizeof(resp.result))) < 0)
		goto failed; 
	
	if ((retval = sct_membuf_add_chunk(respbuf, &resp.data_sz, 
					   sizeof(resp.data_sz))) < 0)
		goto failed; 
	
	if ((retval = sct_membuf_add_chunk(respbuf, resp.data, 
					   resp.data_sz)) < 0)
		goto failed; 
	
	if ((retval = sct_membuf_add_chunk(respbuf, &resp.magic, 
					   sizeof(resp.magic))) < 0)
		goto failed; 

	printk(KERN_DEBUG "build_response_packet: returning a packet of size %d, "
	       "buffer serialized size is %d\n", resp.size, 
	       sct_membuf_get_serialized_size(respbuf)); 

	return; 
 failed:
	printk(KERN_DEBUG "build_response_packet failed, retval from add_chunk is %d\n", 
	       retval); 
}

void handle_update_packet(sct_dev_t* dev, int id, int cmd)
{
	int rule_id;
	sct_rule_when_t when;
	int syscall_id, mux_func_id;
	sct_action_t* a = NULL;
	sct_filter_node_t* f = NULL;
	unsigned int rule_len; 
	char* rule; 
	int packet_rc; 
	sct_membuf_t* req = dev->req; 
	int retval; 
	int dummy; 

	/* we've already checked that it's a valid packet of the valid size, */ 
	/* so there's no need for the usual size checks here */ 
	/* skip the result field */ 
	sct_membuf_get_chunk(req, &dummy, sizeof(dummy)); 
	printk(KERN_DEBUG "packet result is %d\n", dummy); 

	sct_membuf_get_chunk(req, &rule_len, sizeof(rule_len)); 
	printk(KERN_DEBUG "packet rule len is %d\n", dummy); 

	rule = kmalloc(rule_len, GFP_KERNEL); 
	if (!rule) {
		packet_rc = CTL_TRACKER_ENOMEM; 
		goto err_oom; 
	}

	sct_membuf_get_chunk(req, rule, rule_len); 
	
	/* are sanity checks necessary here? tracker_build_rule */
	/* should perform them for us */
	if ((retval = tracker_build_rule(rule, rule_len, 
					 &rule_id, &when, 
					 &syscall_id, &mux_func_id, 
					 &a, &f)) < 0) {
		printk(KERN_DEBUG "handle_request_packet: tracker_build_rule "
		       "returned %d\n", retval); 
		packet_rc = CTL_TRACKER_EBADTRACKINGRULE;
		goto err_build_rule; 
	}
	
	if ((retval = sanity_check_user_input(syscall_id, mux_func_id, when)) < 0) {
		/* not good enough, but all sanity_check_user_input will */
		/* return to us is -EINVAL on error. it needs serious */
		/* surgery */
		printk(KERN_DEBUG "handle_request_packet: sanity_check returned %d\n", 
		       retval); 
		packet_rc = CTL_TRACKER_EBADTRACKINGRULE;
		goto err_failed_sanity_check;
	}
	
	printk(KERN_NOTICE "handle_request_packet: about to modify rules!\n");
	
	/* modify rules */
	switch (cmd) {
	case(CTL_TRACKER_CMD_ADD_RULE):
		retval = tracker_add_rule(rule_id, when,
					  syscall_id, mux_func_id,
					  f, a);
		a = NULL; f = NULL; /* mark that we consumed the action and the filter */ 
		break;
	
	case(CTL_TRACKER_CMD_DEL_RULE):
		retval = tracker_del_rule(rule_id, when,
					  syscall_id, mux_func_id);
		break;
	
	case(CTL_TRACKER_CMD_DEL_ALL_RULES_FOR_SYSCALL):
		retval = tracker_del_all_rules_for_syscall(syscall_id,
							   mux_func_id);
		break;
	
	case(CTL_TRACKER_CMD_DEL_ALL_MODULE_RULES):
		retval = tracker_del_all_module_rules();
		break;
	
	case(CTL_TRACKER_CMD_PRINT_ALL_RULES):
		retval = tracker_print_all_rules();
		break;
	
	default:
		printk(KERN_NOTICE "device_file.c:handle_packet(): invalid or unsupported "
		       "type of request - %d\n", cmd);
		packet_rc = CTL_TRACKER_EINVAL;
		goto err_invalid_request;
	}

	switch (retval) {
	case(-EINTR):
		packet_rc = CTL_TRACKER_EINTR;
		break;
	case(-ENOMEM):
		packet_rc = CTL_TRACKER_ENOMEM;
		break;
	case (0):
		packet_rc = CTL_TRACKER_SUCCESS;
		break;
	default: /* catch all */
		packet_rc = CTL_TRACKER_EINVAL;
	}

	/* build response packet */
 err_invalid_request:
 err_failed_sanity_check:
	if (a)
		sct_free_action(a); 
	if (f)
		sct_free_filter_node(f); 
 err_build_rule:
 err_oom: 
	build_response_packet(dev->resp, id, cmd, packet_rc, 0, NULL);
	/* and we're done! */
}

void handle_get_packet(sct_dev_t* dev, unsigned int id, unsigned int cmd)
{
	int retval = -1, packet_rc = CTL_TRACKER_EINVAL; 
	sct_membuf_t* mbuf = NULL; 
	
	switch(cmd) {
	case CTL_TRACKER_CMD_GET_ALL_RULE_COUNT: 
		if (!(mbuf = sct_membuf_create(sizeof(int), NULL, NULL))) {
			retval = -ENOMEM; 
			goto done; 
		}
		retval = tracker_get_all_rule_count(mbuf); 
		break; 
	case CTL_TRACKER_CMD_GET_ALL_RULES:
		if (!(mbuf = sct_membuf_create(1024 * 16, NULL, NULL))) {
			retval = -ENOMEM; 
			goto done; 
		}
		retval = tracker_get_all_rules(mbuf); 
		break; 
	default: 
		; /* nothing */ 
	}

	switch (retval) {
	case(-EINTR):
	        packet_rc = CTL_TRACKER_EINTR;
		break;
	case(-ENOMEM):
		packet_rc = CTL_TRACKER_ENOMEM;
		break;
	case (0):
		packet_rc = CTL_TRACKER_SUCCESS;
		break;
	default: /* catch all */
		packet_rc = CTL_TRACKER_EINVAL;
	}
 done:
	build_response_packet(dev->resp, id, cmd, packet_rc, 
			      sct_membuf_get_serialized_size(mbuf), 
			      sct_membuf_get_data(mbuf)); 
	
	printk(KERN_DEBUG "get returned a packet with data sz %d and data %d\n",
	       sct_membuf_get_serialized_size(mbuf), 
	       *(int*)sct_membuf_get_data(mbuf)); 
		
	if (mbuf) 
		sct_membuf_destroy(mbuf); 
}

void handle_request_packet(sct_dev_t* dev)
{
	int dummy, packet_id, cmd;
	sct_membuf_t* req = dev->req; 
	
	/* parse the user space supplied buffer */
	/* we are guaranteesd the packet is valid */
	/* when we get called, since req_has_whole_packet */
	/* has validated it */
	/* no checking, since we know the mbuf is big enough */ 
	sct_membuf_get_chunk(req, &dummy, sizeof(dummy)); /* skip the size field */
	printk(KERN_DEBUG "packet size is %d\n", dummy); 
	sct_membuf_get_chunk(req, &packet_id, sizeof(packet_id)); /* id field */ 
	printk(KERN_DEBUG "packet id is %d\n", packet_id); 
	sct_membuf_get_chunk(req, &cmd, sizeof(cmd)); /* the type of request field */
	printk(KERN_DEBUG "packet command is %d\n", cmd); 

	/* now, what kind of command is it? */ 
	switch (cmd) {
	case(CTL_TRACKER_CMD_ADD_RULE):
	case(CTL_TRACKER_CMD_DEL_RULE):
	case(CTL_TRACKER_CMD_DEL_ALL_RULES_FOR_SYSCALL):
	case(CTL_TRACKER_CMD_DEL_ALL_MODULE_RULES):	
	case(CTL_TRACKER_CMD_PRINT_ALL_RULES):
		handle_update_packet(dev, packet_id, cmd); 
		break;
	case(CTL_TRACKER_CMD_GET_ALL_RULES):
	case(CTL_TRACKER_CMD_GET_ALL_RULE_COUNT):
		handle_get_packet(dev, packet_id, cmd); 
		break; 
	default:
		printk(KERN_NOTICE "device_file.c:handle_packet(): invalid or unsupported "
		       "type of request - %d\n", cmd);
		build_response_packet(dev->resp, packet_id, cmd, CTL_TRACKER_EINVAL, 0, NULL); 
	}	
	/* the response packet was built by one of the packt handlers or by us - return */ 
}

/* does this request buffer contain a whole request packet, or not yet? */
int req_has_whole_packet(sct_membuf_t* req)
{
	unsigned int req_sz;
	unsigned int cookie = DEV_FILE_MAGIC_COOKIE;
	unsigned int packet_cookie;
	unsigned char* cookie_loc; 
	unsigned char* data; 
	size_t datasz; 
    
	if (!req)
		return -EINVAL;

	/* get the size of this request packet, which is the first 'int' in */
	/* the data buffer */
	data = sct_membuf_get_data(req); 
	datasz = sct_membuf_get_serialized_size(req); 

	if (datasz < sizeof(req_sz))
		return -EINVAL; 

	memcpy(&req_sz, data, sizeof(req_sz));
	
	/* now, sanity checks on req_sz */

	/* we dont have quite enough data yet */
	if (req_sz > datasz) {
		return -EINVAL; /* not a full packet yet */
	}
	
	/* ok, so req_sz must be smaller or equal to the data_size. cool! */
    
	/* let's check the last int from the end, if it's the magic cookie */
	/* we expect */

#ifdef SCT_MODULE_DEBUG
	printk(KERN_DEBUG "request packet dump: "); 
	for (cookie_loc = data; cookie_loc < data + req_sz; cookie_loc += sizeof(int))
		printk("%0x ", *(unsigned int*)cookie_loc);
	printk("\n"); 
#endif
	
	cookie_loc = &data[req_sz - sizeof(cookie)]; 

	memcpy(&packet_cookie, cookie_loc, sizeof(cookie));
	if (packet_cookie != cookie) {
		printk("req_has_whole_packet got invalid cookie - wanted %0x, got %0x\n", 
		       cookie, packet_cookie); 
		return -EINVAL; 
	}
	
	/* whee, success */
	return 0;
}
