#!/usr/bin/perl -W

# This file is part of syscalltrack, a GNU/Linux kernel module and
# user space utilities for logging or modifying any system call    
# invocation.                                                      
#                                                                  
# Copyright (C) 2000-2002 guy keren, choo@actcom.co.il      
# License: GNU General Public License                              

#
# $Id: gen_syscalls.pl,v 1.58 2002/06/15 15:16:14 mulix Exp $
#

use strict;

use warnings;

use Data::Dumper;

sub mywarning_hook {
    die $_[0];
}

BEGIN {

    #$SIG{__WARN__} = \&mywarning_hook;
}

# input/output files.

# list of struct types and syscalls to support.
my $data_file = "syscalls.dat";

# generated files
my $hijack_stubs_file  = "syscall_hijack_autogen.c";
my $tracker_stubs_file = "syscall_tracker_autogen.c";

# template files used in the generation of the output files.
my $templates_dir            = "./templates";
my $syscall_hijack_stub_tmpl =
  $templates_dir . '/' . "syscall_hijack_stub.tmpl";
my $syscall_tracker_stub_tmpl =
  $templates_dir . '/' . "syscall_tracker_stub.tmpl";
my $syscall_tracker_stub_after_tmpl =
  $templates_dir . '/' . "syscall_tracker_stub_after.tmpl";
my $syscall_tracker_mux_stub_tmpl =
  $templates_dir . '/' . "syscall_tracker_mux_stub.tmpl";
my $syscall_tracker_mux_stub_after_tmpl =
  $templates_dir . '/' . "syscall_tracker_mux_stub_after.tmpl";
my $syscall_tracker_struct_offsets_tmpl =
  $templates_dir . '/' . "syscall_tracker_struct_offsets.tmpl";
my $syscall_tracker_struct_log_tmpl =
  $templates_dir . '/' . "syscall_tracker_struct_log.tmpl";
my $syscall_tracker_struct_data_init_tmpl =
  $templates_dir . '/' . "syscall_tracker_struct_data_init.tmpl";

# data types mappings.
my %types_map = (
    'char'           => 'char',
    'uchar'          => 'unsigned char',
    'schar'          => 'signed char',
    'short'          => 'short',
    'ushort'         => 'unsigned short',
    'unsigned_short' => 'unsigned short',
    'int'            => 'int',
    'uint'           => 'unsigned int',
    'unsigned_int'   => 'unsigned int',
    'long'           => 'long',
    'ulong'          => 'unsigned long',
    'unsigned_long'  => 'unsigned long',
    'string'         => 'char*',
    'const_string'   => 'const char*',
    'buffer'         => 'void*',
    'const_buffer'   => 'const void*',
    'pointer'        => 'void*',
    'mode_t'         => 'mode_t',
    'size_t'         => 'size_t',
    'buffer_size'    => 'size_t',
    'ssize_t'        => 'ssize_t',
    'pid_t'          => 'pid_t',
    'uid_t'          => 'uid_t',
    'gid_t'          => 'gid_t',
    'off_t'          => 'off_t',
    'time_t'         => 'time_t',
    'dev_t'          => 'dev_t',
    'ino_t'          => 'ino_t',
    'nlink_t'        => 'nlink_t',
    'blksize_t'      => 'blksize_t',
    'blkcnt_t'       => 'blkcnt_t',
    'clock_t'        => 'clock_t', 
    'net_ushort'     => 'unsigned short',
    'net_ip'         => 'unsigned long',
);

my %types_data_map = (
    'char'           => 'DT_CHAR',
    'short'          => 'DT_SHORT',
    'ushort'         => 'DT_USHORT',
    'unsigned_short' => 'DT_USHORT',
    'int'            => 'DT_INT',
    'uint'           => 'DT_UINT',
    'unsigned_int'   => 'DT_UINT',
    'long'           => 'DT_LONG',
    'ulong'          => 'DT_ULONG',
    'unsigned_long'  => 'DT_ULONG',
    'string'         => 'DT_STRING',
    'const_string'   => 'DT_STRING',
    'buffer'         => 'DT_BUFFER',
    'const_buffer'   => 'DT_BUFFER',
    'mode_t'         => 'DT_USHORT',
    'size_t'         => 'DT_ULONG',
    'buffer_size'    => 'DT_ULONG',
    'ssize_t'        => 'DT_LONG',
    'pid_t'          => 'DT_INT',
    'uid_t'          => 'DT_UINT',
    'gid_t'          => 'DT_UINT',
    'off_t'          => 'DT_LONG',
    'time_t'         => 'DT_LONG',
    'dev_t'          => 'DT_USHORT',
    'ino_t'          => 'DT_ULONG',
    'nlink_t'        => 'DT_USHORT',
    'blksize_t'      => 'DT_ULONG',
    'blkcnt_t'       => 'DT_ULONG',
    'clock_t'        => 'DT_LONG', 
    'net_ushort'     => 'DT_USHORT',
    'net_ip'         => 'DT_ULONG',
);

my %types_makeval_map = (
    'DT_CHAR'   => 'sct_make_char_val',
    'DT_SHORT'  => 'sct_make_short_val',
    'DT_USHORT' => 'sct_make_ushort_val',
    'DT_INT'    => 'sct_make_int_val',
    'DT_UINT'   => 'sct_make_uint_val',
    'DT_LONG'   => 'sct_make_long_val',
    'DT_ULONG'  => 'sct_make_ulong_val',
    'DT_STRING' => 'sct_make_string_val',
    'DT_BUFFER' => 'sct_make_buffer_val',
);

# this mapping should be indexed according to the DT_XXX data type
# only add new types here if they dont have a DT_XXX type
my %types_specifier_map = (
    'DT_CHAR'    => 'c',
    'DT_SHORT'   => 'h',
    'DT_USHORT'  => 'uh',
    'DT_INT'     => 'i',
    'DT_UINT'    => 'ui',
    'DT_LONG'    => 'l',
    'DT_ULONG'   => 'ul',
    'DT_STRING'  => 'P',
    'DT_BUFFER'  => 'B',
    'void'       => 'v',
    'net_ushort' => 'U',
    'net_ip'     => 'U',
);

my %struct_field_logger_helper_map = (
    'net_ushort' => 'LHF_NTOHS',
    'net_ip'     => 'LHF_INET_NTOA',
);

my %basic_types_dummy_data_specifier_map = (
    'c'  => "'\\0'",
    'h'  => '0',
    'uh' => '0',
    'i'  => '0',
    'ui' => '0',
    'l'  => '0L',
    'ul' => '0L',
    's'  => 'NULL',
    'v'  => 'NULL',
    'B'  => 'NULL'
);

my ($matchable) = "";
my ($param_names_for_logging_string);
my ($syscall_param_matchable_ref);
my ($syscall_param_types_ref);
my ($syscall_param_names_ref);

my $os = `uname -s`;
chomp($os);

if ( $#ARGV != 0 ) {
    die "Usage - $0 <syscalls number>";
}

my $num_syscalls = $ARGV[0];

my $cvs_id = "/*\n * \$Id\$\n */\n\n";

# read template files first.
open( TMPL, "<$syscall_hijack_stub_tmpl.$os" )
  || die "Cannot open file '$syscall_hijack_stub_tmpl.$os' "
  . "for reading - $!";
my @syscall_hijack_stub_tmpl = <TMPL>;
close(TMPL);

open( TMPL, "<$syscall_tracker_stub_tmpl.$os" )
  || die "Cannot open file '$syscall_tracker_stub_tmpl.$os' "
  . "for reading - $!";
my @syscall_tracker_stub_tmpl = <TMPL>;
close(TMPL);

open( TMPL, "<$syscall_tracker_stub_after_tmpl.$os" )
  || die "Cannot open file '$syscall_tracker_stub_after_tmpl.$os'"
  . " for reading - $!";
my @syscall_tracker_stub_after_tmpl = <TMPL>;
close(TMPL);

open( TMPL, "<$syscall_tracker_mux_stub_tmpl.$os" )
  || die "Cannot open file '$syscall_tracker_mux_stub_tmpl.$os' "
  . "for reading - $!";
my @syscall_tracker_mux_stub_tmpl = <TMPL>;
close(TMPL);

open( TMPL, "<$syscall_tracker_mux_stub_after_tmpl.$os" )
  || die "Cannot open file '$syscall_tracker_mux_stub_after_tmpl.$os' "
  . "for reading - $!";
my @syscall_tracker_mux_stub_after_tmpl = <TMPL>;
close(TMPL);

open( TMPL, "<$syscall_tracker_struct_offsets_tmpl.$os" )
  || die "Cannot open file '$syscall_tracker_struct_offsets_tmpl.$os' "
  . "for reading - $!";
my @syscall_tracker_struct_offsets_tmpl = <TMPL>;
close(TMPL);

open( TMPL, "<$syscall_tracker_struct_log_tmpl.$os" )
  || die "Cannot open file '$syscall_tracker_struct_log_tmpl.$os' "
  . "for reading - $!";
my @syscall_tracker_struct_log_tmpl = <TMPL>;
close(TMPL);

open( TMPL, "<$syscall_tracker_struct_data_init_tmpl.$os" )
  || die "Cannot open file '$syscall_tracker_struct_data_init_tmpl.$os' "
  . "for reading - $!";
my @syscall_tracker_struct_data_init_tmpl = <TMPL>;
close(TMPL);

# open data file, scan and check input, generate output along the way.
open( DAT, "<$data_file" )
  || die "Cannot open file '$data_file' for reading - $!";
print "generating system calls stubs - $num_syscalls system calls\n";

if ( -e $hijack_stubs_file ) {
    rename( "$hijack_stubs_file", "${hijack_stubs_file}.old" )
      || die "Cannot rename '$hijack_stubs_file' "
      . "to '${hijack_stubs_file}.old' - $!";
}
if ( -e $tracker_stubs_file ) {
    rename( "$tracker_stubs_file", "${tracker_stubs_file}.old" )
      || die "Cannot rename '$tracker_stubs_file' "
      . "to '${tracker_stubs_file}.old' - $!";
}
open( HIJACK_AUTOGEN, ">$hijack_stubs_file" )
  || die "Cannot open '$hijack_stubs_file' for writing - $!";
open( TRACKER_AUTOGEN, ">$tracker_stubs_file" )
  || die "Cannot open '$tracker_stubs_file' for writing - $!";

print HIJACK_AUTOGEN $cvs_id;
print HIJACK_AUTOGEN "/* This file is automatically generated - "
  . "do not edit. */\n";
print HIJACK_AUTOGEN "/* built by '$0 $num_syscalls' for $os. */\n\n";

print HIJACK_AUTOGEN "\n#include <asm/errno.h>\n\n";

print TRACKER_AUTOGEN $cvs_id;
print TRACKER_AUTOGEN "/* This file is automatically generated - "
  . "do not edit. */\n";
print TRACKER_AUTOGEN "/* built by '$0 $num_syscalls' for $os. */\n\n";
print TRACKER_AUTOGEN "\n#include <asm/errno.h>\n\n";
print TRACKER_AUTOGEN "#include \"var_callbacks.h\"\n";
print TRACKER_AUTOGEN "#include \"struct_handling.h\"\n";
print TRACKER_AUTOGEN "\n";

my @syscall_ids               = ();
my @syscall_functions         = ();
my %structs_fields            = ();
my %structs_types_indexes     = ();
my %syscall_track_mux_init    = ();
my %syscall_track_mux_cleanup = ();

my ($include_file);
my ($asis);
my ($cpp);
my ( $struct_type, $struct_fields, $kernel_function );
my ( $syscall_id, $syscall_name, $syscall_params );

my ($first_stub_function) = 'yes';
my ($struct_type_index)   = 0;

while (<DAT>) {
    chomp;

    # strip comments, from a # mark to the end of the line
    s/#.*$//;
    /^\s*$/ && next;    # skip empty lines.
                        # handle 'include' lines:
    if (/^\s*include\s*:\s*(\S+)\s*$/) {
        $include_file = $1;
        print HIJACK_AUTOGEN "#include $include_file\n";
        print TRACKER_AUTOGEN "#include $include_file\n";
        next;
    }

    # handle 'asis' lines:
    if (/^\s*asis:\s*(.*)$/) {
        $asis = $1;
        print HIJACK_AUTOGEN "$asis",  "\n";
        print TRACKER_AUTOGEN "$asis", "\n";
        next;
    }

    # handle 'cpp' lines:
    if (/^\s*cpp:\s*(.*)$/) {
        $cpp = $1;
        print HIJACK_AUTOGEN "#${cpp}",  "\n";
        print TRACKER_AUTOGEN "#${cpp}", "\n";
        next;
    }

    # handle 'struct' lines:
    #in the format of struct struct_type (struct_field1 struct_filed2...);
    if (/^\s*struct\s+([^\(\s]+)\s*\(([^\)]+)\)\s*;\s*$/) {
        $struct_type   = $1;
        $struct_fields = $2;
        print "got struct '$struct_type', fields '$struct_fields'\n";
        $structs_fields{$struct_type}        = $struct_fields;
        $structs_types_indexes{$struct_type} = $struct_type_index;
        $struct_type_index++;

        # encountering the first struct, print headline
        if ( $struct_type_index == 1 ) {
            print TRACKER_AUTOGEN "/*\n";
            print TRACKER_AUTOGEN " * struct-offsets "
              . "evaluation functions for dynamic type-casting.\n";
            print TRACKER_AUTOGEN " */\n\n";
        }
        &emit_struct_function( $data_file, $., $struct_type );
        next;
    }

    # syscall format:
    # kernel_function    :syscall_ID     : syscall_name    (param_type param_name...);
    if (/^\s*([^:\s]+)\s*:\s*([^:\s]+)\s*:\s*([^\(\s]+)\s*\(([^\)]+)\)\s*;\s*$/)
    {
        $kernel_function = $1;
        $syscall_id      = $2;
        $syscall_name    = $3;
        $syscall_params  = $4;

        # if the kernel function begins with a '!', ignore it
        # beautify it (remove '!' for debug printing)
        if ( $kernel_function =~ s/^\!// ) {
            print "kernel function '$kernel_function' "
              . "shall not be autogenerated\n";
            next;
        }

        # handle the case where the syscall id is for a muxed syscall,
        # in the "sys.func" format
        if ( $syscall_id =~ /^(\d+)\.(\d+)/ ) {
            my $syscall = $1;
            my $func    = $2;

            # hold the $func in the upper 2 bytes,
            # and the $syscall in the lower two.
            $syscall_id = ( ( $func << 16 ) | $syscall );
        }

        push ( @syscall_ids,       $syscall_id );
        push ( @syscall_functions, $syscall_name );
        print "kernel_function - '$kernel_function'\n";
        print "    syscall_id - '$syscall_id'\n";
        print "    syscall_name - '$syscall_name'\n";

        #print "    syscall_params - '$syscall_params'\n";

        if ( $first_stub_function eq 'yes' ) {
            print HIJACK_AUTOGEN "/*\n";
            print HIJACK_AUTOGEN " * Stub functions for syscall hijacking.\n";
            print HIJACK_AUTOGEN " */\n\n";
            print TRACKER_AUTOGEN "/*\n";
            print TRACKER_AUTOGEN
              " * Tracking functions for syscall hijacking.\n";
            print TRACKER_AUTOGEN " */\n\n";
            $first_stub_function = 'no';
        }

        &emit_tracker_and_hijack_stubs(
            $data_file,               $.,
            $kernel_function,         $syscall_id,
            $syscall_name,            $syscall_params,
            \%syscall_track_mux_init, \%syscall_track_mux_cleanup
        );
        next;
    }
    else {
        die "Invalid syntax in '$data_file' line $.";
    }
}

# make 'init_mux_functions'
print TRACKER_AUTOGEN "\n";
print TRACKER_AUTOGEN "void\n";
print TRACKER_AUTOGEN "init_mux_functions(void)\n";
print TRACKER_AUTOGEN "{\n";
foreach $syscall_id ( keys %syscall_track_mux_init ) {
    print TRACKER_AUTOGEN "$syscall_track_mux_init{$syscall_id}";
    print TRACKER_AUTOGEN "\n";
}
print TRACKER_AUTOGEN "}\n";

# make 'cleanup_mux_functions'
print TRACKER_AUTOGEN "\n";
print TRACKER_AUTOGEN "void\n";
print TRACKER_AUTOGEN "cleanup_mux_functions(void)\n";
print TRACKER_AUTOGEN "{\n";
foreach $syscall_id ( keys %syscall_track_mux_cleanup ) {
    print TRACKER_AUTOGEN "$syscall_track_mux_cleanup{$syscall_id}";
    print TRACKER_AUTOGEN "\n";
}
print TRACKER_AUTOGEN "}\n";

print HIJACK_AUTOGEN "\n";
print TRACKER_AUTOGEN "\n";

&emit_structs_data_array( \%structs_fields, \%structs_types_indexes );
&emit_stubs_arrays( $num_syscalls, \@syscall_ids );

close(HIJACK_AUTOGEN)
  || die "Error closing '$hijack_stubs_file' - $!";
close(TRACKER_AUTOGEN)
  || die "Error closing '$tracker_stubs_file' - $!";

exit(0);

# replace macros which come in the form of %%$key%% by $value
sub replace_macros {
    my $line_copy_ptr     = shift;
    my $mappings_no_chomp = shift;
    my $mappings_w_chomp  = shift;

    while ( my ( $key, $value ) = each(%$mappings_no_chomp) ) {
        ${$line_copy_ptr} =~ s/\%\%$key\%\%/$value/g;
    }
    while ( my ( $key, $value ) = each(%$mappings_w_chomp) ) {
        if ( ${$line_copy_ptr} =~ s/\%\%$key\%\%/$value/g ) {
            chomp( ${$line_copy_ptr} );
        }
    }

    return 0;
}


sub call_replace_macros{
    my ($mappings_no_chomp, $mappings_w_chomp,@tmpl)=@_;

    # generate the functions using a template file.
    # avoid modifying the arrays - thus,
    # work on a copy of that whole array
    my $line;
    my $accumulator='';
    foreach $line (@tmpl){
	&replace_macros( \$line, $mappings_no_chomp, $mappings_w_chomp );
	$accumulator.="$line";
    }

    return $accumulator;
}



# emit functions that allow calculating offsets of fields in the given struct.
sub emit_struct_function {
    my ( $data_file, $data_file_line, $struct_type ) = @_;
    my ($struct_params) = $structs_fields{$struct_type};
    my (@struct_params_strings) = split ( ',', $struct_params );
    my ( $pre_num_struct_fields, $num_struct_fields );
    my ($field_offsets_calc) = "";
    my ($field_types_calc)   = "";

    my @struct_params_breakdown = ();

    &break_down_params_list( $data_file, $data_file_line,
        \@struct_params_strings, \@struct_params_breakdown );

    my @struct_params_expanded = ();

    # handle struct-in-struct
    &expand_struct_in_struct(
        $data_file, $data_file_line,
        \@struct_params_breakdown,
        \@struct_params_expanded,
    );

    $num_struct_fields = scalar(@struct_params_expanded);

    # prepare code to calculate the offset+type for each field of the struct.
    for ( my $i = 0 ; $i < $num_struct_fields ; $i++ ) {
        my $record = $struct_params_expanded[$i];
        if ( $record->{'matchable'} eq 'matchable' ) {
            $field_offsets_calc .= "\t\toffsets_vec[$i] = (long)( ((char*)&s."
              . $record->{'name'}
              . ") - base);\n";
            $field_types_calc .= "\t\ttypes_vec[$i] = "
              . $types_data_map{ $record->{'type'} } . ";\n";
        }
        else {
            $field_offsets_calc .= "\t\toffsets_vec[$i] = -1;\n";
            $field_types_calc .= "\t\ttypes_vec[$i] = DT_LAST_TYPE;\n";
        }
    }

    my (@struct_param_names_deref) = ();
    for ( my $i = 0 ; $i < $num_struct_fields ; $i++ ) {
        my $record = $struct_params_expanded[$i];
	defined($record->{'fmt_spec'}) or
	    die "undefined format $data_file $data_file_line";
        if ( defined( $struct_field_logger_helper_map{ $record->{'type'} } ) ) {
            my $lhf_func = $struct_field_logger_helper_map{ $record->{'type'} };
            $struct_param_names_deref[$i] =
              '&s->' . $record->{'name'} . ", (void*)$lhf_func";
        }
        else {
            $struct_param_names_deref[$i] = 's->' . $record->{'name'};
        }
    } 
    my ($struct_param_fmt_spec_str) = join (
        ' ',
        (
            map {  $_->{'fmt_spec'} }
              @struct_params_expanded
        )
    );

    my ($struct_fields_names) = join ( ', ', @struct_param_names_deref );

    my %mapping = (
		   'STRUCT_TYPE'               => $struct_type,
		   'STRUCT_TYPE_NUM_FIELDS'    => $num_struct_fields,
		   'STRUCT_FIELDS_OFFSET_CALC' => $field_offsets_calc,
		   'STRUCT_FIELDS_TYPES_CALC'  => $field_types_calc,
		   'STRUCT_FIELDS_NAMES'       => $struct_fields_names,
		   'STRUCT_LOG_SPEC'           => $struct_param_fmt_spec_str,
		  );

    my $str = &call_replace_macros
	(\%mapping, {},@syscall_tracker_struct_offsets_tmpl );
    print (TRACKER_AUTOGEN $str);

    $str = &call_replace_macros
	(\%mapping, {}, @syscall_tracker_struct_log_tmpl);
    print (TRACKER_AUTOGEN $str);

    print TRACKER_AUTOGEN "\n";
}

sub get_param_type_specifier {
    my ($param_type) = @_; 

    # first, look for this param type in the type specifier map,
    # and then look for it based on its DT_XXXX data type in 
    # the types data map
    my $ret; 
    if ( exists($types_specifier_map{$param_type})) {
	$ret = $types_specifier_map{$param_type};
    }
    elsif ( exists($types_data_map{$param_type}) &&
	    exists($types_specifier_map{$types_data_map{$param_type}}) )
    {
	$ret = $types_specifier_map{$types_data_map{$param_type}};
    }
    $ret; 
}

sub emit_tracker_and_hijack_stubs {
    my (
        $data_file,                  $data_file_line,
        $kernel_function,            $syscall_id,
        $syscall_name,               $syscall_params,
        $syscall_track_mux_init_ref, $syscall_track_mux_cleanup_ref
      )
      = @_;
    my $is_mux_func ;

    # check if this 'syscall id' actually belongs to a muxed function.
    my $mux_func_id = ( $syscall_id >> 16 ) & 0xffff;
    $syscall_id  = $syscall_id & 0xffff;
    if ( $mux_func_id > 0 ) {
        $is_mux_func = 1;
        print "    syscall $syscall_id - muxes function $mux_func_id\n";

        # make the init code for this mux function
        if ( !defined( %$syscall_track_mux_init_ref->{$syscall_id} ) ) {
            %$syscall_track_mux_init_ref->{$syscall_id} =
              "\t/* init for mux syscall '$syscall_id'. */\n";
        }
        %$syscall_track_mux_init_ref->{$syscall_id} .=
          "\tinit_mux_function(${syscall_id}, ${mux_func_id}, "
          . "(func_ptr)track_before_syscall_${syscall_id}_${mux_func_id},"
          . " (func_ptr)track_after_syscall_"
          . "${syscall_id}_${mux_func_id});\n";

        # make the cleanup code for this mux function
        if ( !defined( %$syscall_track_mux_cleanup_ref->{$syscall_id} ) ) {
            %$syscall_track_mux_cleanup_ref->{$syscall_id} =
              "\t/* cleanup for mux syscall '$syscall_id'. */\n";
        }
        %$syscall_track_mux_cleanup_ref->{$syscall_id} .=
          "\tcleanup_mux_function(${syscall_id}, ${mux_func_id});\n";
    }
    else{
	
        $is_mux_func = 0;
    }

    my (@syscall_params_strings) = split ( ',', $syscall_params );

    my (@syscall_params_breakdown);

    # break down the parameters list.
    &break_down_params_list( $data_file, $data_file_line,
        \@syscall_params_strings, \@syscall_params_breakdown, );

    &screen_print(@syscall_params_breakdown);

    my $params_logging_spec=prepare_params_logging_spec
	($data_file, $data_file_line,@syscall_params_breakdown);

    # prepare code for generating function parameters variables,
    # and freeing them.
    my $syscall_params_filter_make = "";
    my $syscall_params_filter_free = "";

    # prepare code for filling function struct parameters variables.
    my $syscall_params_struct_filter_assign = "";

    # prepare code for generating function parameters variable setting calls.
    my $syscall_params_filter_vars_set = prepare_syscall_params_filter_vars_set
	($data_file, $data_file_line, @syscall_params_breakdown);

    # prepare code for generating syscall retval variable setting call.
    my $syscall_params_filter_retval_var_set =
	"\t\tset_var_callback(VT_RETVAL, global_retval_callback, "
	    . "(void*)syscall_rc);\n";

    my ($syscall_copied_vars_def, $syscall_copied_vars_copy, 
     $syscall_copied_vars_goto_free, $syscall_copied_vars_free)
	= prepare_syscall_copied_vars
	    ($data_file, $data_file_line, $syscall_id,
	     @syscall_params_breakdown);

    # void handling
    # we expect void to come in one way only: alone.
    # the only legitimate void is when the syscall gets
    # no parameters at all, and void represents that.


    my ( $param_output_string, $param_names_string );
    my ( $before_syscall_params_recv, $after_syscall_params_recv );
    my ($after_retval_func_param);
    my ($after_retval_sys_ret);
    my ($before_retval_variable);


    #check if the first record's type is void.
    #void is not legitimate any other way
    if ($syscall_params_breakdown[0]->{'pre'}->{'type'} eq "void"){

	#check if the discovered void type is alone,
	#void is not legitimate any other way
	
	( scalar(@syscall_params_breakdown) > 1 )
	    and die "a parameter of a function with more than "
		. "one null variable was void,"
		    . "in '$data_file' line $data_file_line";
	($param_output_string, $param_names_string,
	 $param_names_for_logging_string)=('','','');
        $before_retval_variable     = "&fake_syscall_retval";
        $after_retval_sys_ret       = "orig_syscall_retval";
        $after_retval_func_param    = "int syscall_rc";
        $after_syscall_params_recv  = $after_retval_func_param;
        $before_syscall_params_recv = "int* syscall_rc";
    }else{
	# generate parameters definition string.
	($param_output_string, $param_names_string, 
	 $param_names_for_logging_string) =
	     prepare_parameters_definition_strings
		 ($data_file, $data_file_line, @syscall_params_breakdown);
	# if we are logging any parameters, add a ',' before them 
	if ($param_names_for_logging_string ne '') {
	    $param_names_for_logging_string =
		", " . $param_names_for_logging_string; 
	}
        $after_retval_func_param   = "int syscall_rc,";
        $before_retval_variable    = ", &fake_syscall_retval";
        $after_retval_sys_ret      = "orig_syscall_retval,";
        $after_syscall_params_recv =
	    $after_retval_func_param . " " . $param_output_string;
        $before_syscall_params_recv = $param_output_string . "int* syscall_rc";
	# get rid of the trailing ', '
	$after_syscall_params_recv        =~ s/, $//;
    }

    # generate the syscall hijack stub function for this syscall,
    # unless its not a syscall, but rather a muxed function.
    if ( $is_mux_func == 0 ) {
        my $hijack_params_recv;
        if ( $param_output_string eq '' ) {
            $hijack_params_recv = "void";
        }
        else {
            $hijack_params_recv = $param_output_string;
	    # get rid of the trailing ', '
	    $hijack_params_recv=~ s/, $//;
        }
        my %mapping = (
            'SYSCALL_ID'                 => $syscall_id,
            'SYSCALL_NAME'               => $syscall_name,
            'AFTER_RETVAL_FUNC_PARAM'    => $after_retval_func_param,
            'AFTER_SYSCALL_PARAMS_RECV'  => $after_syscall_params_recv,
            'BEFORE_RETVAL_VARIABLE'     => $before_retval_variable,
            'AFTER_RETVAL_SYS_RET'       => $after_retval_sys_ret,
            'SYSCALL_PARAMS_RECV'        => $hijack_params_recv,
            'BEFORE_SYSCALL_PARAMS_RECV' => $before_syscall_params_recv,
            'SYSCALL_PARAMS_NAMES'       => $param_names_string,
        );


	my $str = &call_replace_macros
	    (\%mapping, {},@syscall_hijack_stub_tmpl);
	print (HIJACK_AUTOGEN $str. "\n");
    }
    my (@tracker_stub_tmpl);
    @tracker_stub_tmpl =
      ( ( $is_mux_func == 0 )
        ? @syscall_tracker_stub_tmpl
        : @syscall_tracker_mux_stub_tmpl );

    my     %mapping_no_chomp = 
	(
	 'MUX_FUNC_ID'                => $mux_func_id,
	 'SYSCALL_ID'                 => $syscall_id,
	 'SYSCALL_NAME'               => $syscall_name,
	 'SYSCALL_PARAMS_RECV'        => $param_output_string,
	 'SYSCALL_PARAMS_NAMES'       => $param_names_string,
	 'AFTER_SYSCALL_PARAMS_RECV'  => $after_syscall_params_recv,
	 'BEFORE_SYSCALL_PARAMS_RECV' => $before_syscall_params_recv,
	 'AFTER_RETVAL_SYS_RET'       => $after_retval_sys_ret,
	 'PARAMS_LOGGING_SPEC'        => $params_logging_spec,
	 'PARAM_NAMES_FOR_LOGGING'    => $param_names_for_logging_string,
	);

    my %mapping_with_chomp = (
        'SYSCALL_COPIED_VARS_DEF'             => $syscall_copied_vars_def,
        'SYSCALL_COPIED_VARS_COPY'            => $syscall_copied_vars_copy,
        'SYSCALL_COPIED_VARS_GOTO_FREE'       => 
			      $syscall_copied_vars_goto_free,
        'SYSCALL_PARAMS_FILTER_MAKE'          => $syscall_params_filter_make,
        'SYSCALL_PARAMS_FILTER_FREE'          => $syscall_params_filter_free,
        'SYSCALL_PARAMS_STRUCT_FILTER_ASSIGN' =>
          $syscall_params_struct_filter_assign,
        'SYSCALL_PARAMS_FILTER_VARS_SET' => $syscall_params_filter_vars_set,
        'SYSCALL_PARAMS_FILTER_RETVAL_VAR_SET' =>
          $syscall_params_filter_retval_var_set,
        'SYSCALL_COPIED_VARS_FREE' => $syscall_copied_vars_free,
    );

    # generate a track function for the tracker module.

    
    my $str = &call_replace_macros
	(\%mapping_no_chomp, \%mapping_with_chomp, @tracker_stub_tmpl);
    print (TRACKER_AUTOGEN $str . "\n");

    # generate a track function for the tracker module.

    my (@tracker_stub_after_tmpl);
    @tracker_stub_after_tmpl =
      ( ( $is_mux_func == 0 )
        ? @syscall_tracker_stub_after_tmpl
        : @syscall_tracker_mux_stub_after_tmpl );

    $str = &call_replace_macros
	(\%mapping_no_chomp, \%mapping_with_chomp,
	 @tracker_stub_after_tmpl);
	    
    print (TRACKER_AUTOGEN $str . "\n");
}

sub break_down_params_list {
    my ( $data_file, $data_file_line, $syscall_params_strings_ref,
        $syscall_params_breakdown_ref )
      = @_;

    my ($syscall_param_string);

    my $dont_try_to_match_void=0;

    foreach $syscall_param_string (@$syscall_params_strings_ref) {

	my ($param_matchable, $param_full_type, $param_name, $param_isptr,
	 $param_isarray, $param_array_len, $param_type) =
	     parse_param_string ( $syscall_param_string,
				  $dont_try_to_match_void,
				  $data_file, $data_file_line );
	my $struct_to_put = {
			  'matchable' => $param_matchable,
			  'type'      => $param_type,
			  'name'      => $param_name,
			  'isptr'     => $param_isptr,
			  'isarray'   => $param_isarray,
			  'array_len' => $param_array_len,
			 };
	
	# avoid passing unknown types to the logging function by
	# not encoding them in the param format spec
	
	my $item;
	if ( $param_isptr == 1 ) {
	    $item = "p";
	}
	elsif ( $param_type eq 'pointer') {
	    $item = "p";
	}
	elsif ( $param_isarray == 1 ) {
	    
	    # TODO - actually print the array's contents.
	    $item = "p";
	}
	elsif ( $param_type =~ /^(const_)?struct_/ ) {
	    $item = "S";
	} 
	else {
	    $item = &get_param_type_specifier($param_type); 
	}
	defined($item) or die
	  "for $param_matchable $param_full_type,"
	    . " item does not"
	      . " get defined in '$data_file' line "
		. "$data_file_line";

	# push @$syscall_param_fmt_spec_ref, $item;
	$struct_to_put->{'fmt_spec'} = $item;
	
        push @$syscall_params_breakdown_ref, { 'pre' => $struct_to_put };
    }
}

# handle struct-in-struct.
sub expand_struct_in_struct {
    my ( $data_file, $data_file_line, $struct_param_breakdown_ref,
        $struct_param_expanded_ref, )  = @_;
    my $pre_num_struct_fields = scalar(@$struct_param_breakdown_ref);

    for (my $i = 0 ; $i < $pre_num_struct_fields ; $i++ ) {
        my $record     = $struct_param_breakdown_ref->[$i]->{'pre'};
        my $out_record = {};
        if ( $record->{'matchable'} eq 'matchable' ) {
            if ( $record->{'type'} =~ /^(const_)?struct_(.*)$/ ) {
                my ($param_struct_type)   = $2;
                my ($param_struct_params) = $structs_fields{$param_struct_type};
                my (@param_struct_params_strings) =
                  split ( ',', $param_struct_params );
                my @struct_params_breakdown = ();
                &break_down_params_list( $data_file, $data_file_line,
                    \@param_struct_params_strings,
                    \@struct_params_breakdown );

                my ($param_name);
                push @$struct_param_expanded_ref,

                  # pushd the hash formed in the following manner:
                  # take struct_params_breakdown, and for each element of it
                  # transfer the struct which resides in filed 'pre' to
                  # an independent field of the element.
                  # plus, add to each element under field 'name',
                  # the "name-of-the-record.name-of-the-element"
                  (
                    map {
                        my $r   = $_->{'pre'};
                        my $ret = {
                            (
                                map { $_ => $r->{$_} } (
                                    "matchable", "type",
                                    "isptr",     "isarray",
                                    "array_len", "fmt_spec"
                                )
                            ),
                            'name' =>
                              ( $record->{'name'} . '.' . $r->{'name'} ),
                        };
                        $ret;
                      } @struct_params_breakdown
                  );
            }
            else {    # not a struct
                      # Put a fresh copy of the record
                push @$struct_param_expanded_ref, { %{$record} };
            }
        }
        else {
	    # unmatchable
	    # Put a fresh copy of the record
	    #TODO: check who deals with it later on. 
            push @$struct_param_expanded_ref, { %{$record} };
        }
    }
}


sub emit_structs_data_array() {
    my ( $structs_fields_ref, $structs_types_indexes_ref ) = @_;
    my ( $struct_type, $struct_type_index );
    my (@struct_types) = keys %$structs_fields_ref;
    my ($num_types)    = $#struct_types + 1;
    my ($data_init)           = "";
    my ($struct_data_cleanup) = "";

    my $fill_field_text = sub {
        my $field_name  = shift;
        my $field_value = shift;
        return "\tstruct_type_data[$struct_type_index].$field_name"
          . " = $field_value;\n";
    };

    # prepare code to init each entry in the 'struct_type_data' array.
    foreach $struct_type (@struct_types) {

        # make the struct data init code
        $struct_type_index = $$structs_types_indexes_ref{$struct_type};
        $data_init .= "\toffsets_vec = "
          . "sct_make_struct_offsets_vec_${struct_type}(&num_fields,"
          . " &struct_size);\n";
        $data_init .= "\ttypes_vec = sct_make_struct_types_"
          . "vec_${struct_type}(&num_fields);\n";
        foreach my $field (qw(offsets_vec types_vec num_fields struct_size)) {
            $data_init .= $fill_field_text->( $field, $field );
        }
        $data_init .= $fill_field_text->( "logger",
            "sct_log_struct_fields_${struct_type}" );

        # make the struct data cleanup code 
        # offsets_vec
        $struct_data_cleanup .=
          "\tkfree(struct_type_data[$struct_type_index].offsets_vec);\n";
        $struct_data_cleanup .=
          "\tkfree(struct_type_data[$struct_type_index].types_vec);\n";
        foreach my $field (
            qw(offsets_vec types_vec num_fields
            struct_size logger)
          )
        {
            $struct_data_cleanup .= $fill_field_text->( $field, "0" );
        }
    }

    my %mapping = (
		   'STRUCT_TYPES_NUM'           => $num_types,
		   'STRUCT_DATA_INIT'           => $data_init,
		   'STRUCT_DATA_CLEANUP'        => $struct_data_cleanup,
		  );

    my $str = &call_replace_macros
	(\%mapping, {}, @syscall_tracker_struct_data_init_tmpl); 
    
    print TRACKER_AUTOGEN $str . "\n";

}

sub emit_stubs_arrays {
    my ( $num_syscalls, $syscall_ids_ref ) = @_;
    my ($syscall_id,$i);

    print( "\n", "number syscall ids - ", $#$syscall_ids_ref, "\n" );
    print HIJACK_AUTOGEN "/* pointers to stub functions - to be used */\n";
    print HIJACK_AUTOGEN "/* when hijacking system calls.            */\n";
    print HIJACK_AUTOGEN "func_ptr stub_functions[] = {\n";
    print TRACKER_AUTOGEN
      "/* pointers to 'before' stub functions - to be used */\n";
    print TRACKER_AUTOGEN
      "/* when hijacking system calls.                     */\n";
    print TRACKER_AUTOGEN "func_ptr before_track_functions[] = {\n";

    for ( 0 .. $num_syscalls ) {
        $syscall_id = $_;

        # check if we have a stub function for this syscall.
        for ( $i = 0 ; $i <= $#$syscall_ids_ref ; $i++ ) {
            if ( $syscall_ids_ref->[$i] == $syscall_id ) {
                last;
            }
        }
	# TODO: check if it is insecure to use a loop var after the loop
        if ( $i <= $#$syscall_ids_ref ) {
            print HIJACK_AUTOGEN "\t(func_ptr)stub_syscall_${_},\n";
            print TRACKER_AUTOGEN "\t(func_ptr)track_before_syscall_${_},\n";
        }
        else {
            print HIJACK_AUTOGEN "\t(func_ptr)NULL,\n";
            print TRACKER_AUTOGEN "\t(func_ptr)NULL,\n";
        }
    }
    print HIJACK_AUTOGEN "};\n";
    print TRACKER_AUTOGEN "};\n";

    print TRACKER_AUTOGEN "\n";
    print TRACKER_AUTOGEN
      "/* pointers to 'after' stub functions - to be used */\n";
    print TRACKER_AUTOGEN
      "/* when hijacking system calls.                     */\n";
    print TRACKER_AUTOGEN "func_ptr after_track_functions[] = {\n";
    for ( 0 .. $num_syscalls ) {
        $syscall_id = $_;

        # check if we have a stub function for this syscall.
        for ( $i = 0 ; $i <= $#$syscall_ids_ref ; $i++ ) {
            if ( $syscall_ids_ref->[$i] == $syscall_id ) {
                last;
            }
        }
	# TODO: check if it is insecure (loop var outside loop)
        if ( $i <= $#$syscall_ids_ref ) {
            print TRACKER_AUTOGEN "\t(func_ptr)track_after_syscall_${_},\n";
        }
        else {
            print TRACKER_AUTOGEN "\t(func_ptr)NULL,\n";
        }
    }
    print TRACKER_AUTOGEN "};\n";
}

# match to $1=(un)?matchable $3=full_type $4=name, or to void
# if the void is allowed to be match, and is matched,
# then $1,$2,$4 are empty for compatibility, $3 is 'void'.
# also, check if the param is an array or a ptr.

sub parse_param_string{
    my ($param_string, $dont_try_to_match_void, $data_file, $data_file_line)
	= @_;
    
    (( $param_string !~
       /^\s*((un)?matchable)\s+(\S+)\s+(\S+)\s*$/ )
     && ( ($dont_try_to_match_void) ||
	  ( $param_string !~ /^\s*()()(void)()/ )))
	&& die "Invalid parameters syntax in "
	    . "'$data_file' line $data_file_line";
    my $param_matchable = $1;
    my $param_full_type = $3;
    my $param_name      = $4;
    my $param_isptr     = 0;
    my $param_isarray   = 0;
    my $param_array_len = -1;
    my $param_type;
    # strip the param_type of array, ptr
    if ( $param_full_type =~ /^ptr_(.+)$/ ) {
	$param_isptr = 1;
	$param_type  = $1;
    }
    elsif ( $param_full_type =~ /^array\[(\d*)\]_(.*)$/ ) {
	$param_isarray = 1;
	# if the length of the array is not noted, take it to be -1
	if ( $1 ne '' ) {
	    $param_array_len = $1;
	}
	$param_type = $2;
    }
    else {
	$param_type = $param_full_type;
    }
    return ($param_matchable, $param_full_type, $param_name, $param_isptr,
	    $param_isarray, $param_array_len, $param_type);

}

# append the param_output_type in the proper format to param_output_string
sub append_param_output_type{
    my ($param_type, $param_isptr, $param_isarray,
	$param_name, $param_array_len,
	$data_file, $data_file_line) = @_;
    my $param_output_type;

    # special handling for 'struct' types.
    if ( $param_type =~ /^((const_)?struct_)(.*)$/ ) {
	my $param_const_struct  = $1;
	my $param_struct_type   = $3;
	if ( !defined( $structs_fields{$param_struct_type} ) ) {
	    die "Invalid parameter struct type '$param_type' "
		. "in '$data_file' line $data_file_line";
	}
	
	# replace trailing '_' by a space
	$param_const_struct =~ s/_/ /g;
	$param_output_type =
	    $param_const_struct
		. $param_struct_type . "*";
    }else{		
	$param_output_type = $types_map{$param_type};
    }
    ( !defined($param_output_type) || ( $param_output_type eq '' ) )
	&& die "Invalid parameter type '$param_type' "
	    . "in '$data_file' line $data_file_line";
    my $append = $param_output_type;
    if ( $param_isptr == 1 ) {
	$append .= '*';
    }
    $append .=
	( defined($param_name) ? " $param_name" : "" );
    if ($param_isarray) {
	$append .= '[';
	if ( $param_array_len >= 1 ) {
	    $append .= $param_array_len;
	}
	$append .= ']';
    }
    $append .= ", ";
    return $append;
}

# prepare an fmt_spec for this struct.
sub prepare_struct_fmt_spec{
    my ($param_type, $param_name, $param_matchable,  $struct_type,
	$data_file, $data_file_line)=@_;
    my $record_fmt_spec;
    if ( $param_matchable eq 'matchable' ) {
	my ($struct_params)         = $structs_fields{$struct_type};
	my (@struct_params_strings) = split ( ',', $struct_params );
	
	my @struct_params_breakdown = ();
	
	&break_down_params_list
	    ( $data_file, $data_file_line,
	      \@struct_params_strings, \@struct_params_breakdown, );
	
	my @struct_params_expanded = ();
	
	&expand_struct_in_struct
	    ( $data_file, $data_file_line,
	      \@struct_params_breakdown, \@struct_params_expanded, );
	#note: $i is not passed to the function
	# not in use - struct params are now logged via callbacks,
	#   not directly.
	#$syscall_param_fmt_spec[$i] =
	#    "S{" . join(' ', @struct_param_fmt_spec) . "}";
	#$syscall_param_fmt_spec[$i] = "S";
	$record_fmt_spec = "S";
    }
    else {
	#note: $i is not passed to the function
	# we go by the implicit assumption here that any unmatchable 
	# param is a pointer, and that's why it's unmatchable. It is 
	# conceivable that we will one day have a
	# non pointer unmatchable param,
	# in which case this code loses. 
	# TODO: handle unmatchables more intelligently, not sure how.
	#$syscall_param_fmt_spec[$i] = "p";
	$record_fmt_spec = "p";
    }
    return $record_fmt_spec;
}



    # prepare code for param logging spec, including logging of
    # struct members.

sub prepare_params_logging_spec{
    my($data_file, $data_file_line,@syscall_params_breakdown)=@_;
    my $params_logging_spec;
    for ( my $i = 0 ; $i < scalar(@syscall_params_breakdown) ; $i++ ) {
        my $record = $syscall_params_breakdown[$i]->{'pre'};	
        # if it's a struct - prepare a log spec for this struct.
        if ( $record->{'type'} =~ /^(const_)?struct_(.*)$/ ) {
	    my $struct_type = $2;
	    $record->{'fmt_spec'} = prepare_struct_fmt_spec
		($record->{'type'}, $record->{'name'},
		 $record->{'matchable'}, $struct_type,
		 $data_file, $data_file_line);
        }
    }
    # handle undefined fmt_spec: turn it into "",
    # and then join the whole thing
    #TODO: make sure everythignget defined.
    $params_logging_spec = join (
        ' ',
        map {
            (
                defined( $_->{'pre'}->{'fmt_spec'} )
                ? $_->{'pre'}->{'fmt_spec'}
                : "" )
          } @syscall_params_breakdown
    );
    print "params_spec: '", $params_logging_spec, "'\n";
    return $params_logging_spec;
}

# prepare code for generating function parameters variable setting calls.
sub prepare_syscall_params_filter_vars_set{
    my ($data_file, $data_file_line, @syscall_params_breakdown)=@_;
    my $syscall_params_filter_vars_set = "";
    for ( my $i = 0 ; $i <= $#syscall_params_breakdown ; $i++ ) {
        my $record = $syscall_params_breakdown[$i]->{'pre'};
        my $param_type      = $record->{'type'};
        my $param_name      = $record->{'name'};
        my $param_matchable = $record->{'matchable'};
        my $param_datatype  = $types_data_map{$param_type};
	my $param_makeval;
        if ( defined($param_datatype) ) {
            $param_makeval = $types_makeval_map{$param_datatype};
        }
        if ( $param_matchable eq 'matchable' ) {
            if ( $param_type =~ /^(const_)?struct_/ ) {
                $syscall_params_filter_vars_set .=
		    "\t\tset_var_indexed_field_callback(VT_PARAM, ${i}, "
			. "global_struct_field_param_callback, DT_INT, "
			    . "(void*)&${param_name}_struct_data);\n";
            }
            elsif ( $param_type =~ /^(const_)?string$/ ) {
                $syscall_params_filter_vars_set .=
		    "\t\tset_var_indexed_callback(VT_PARAM, ${i}, "
			. "global_param_callback, ${param_datatype}, "
			    . "(void*)${param_name});\n";
            }
            elsif ( $param_type =~ /^(const_)?buffer$/ ) {
		
                # find the name of the parameter which has this buffer's size
                my $buf_size_param_name = undef;
                for ( my $j = 0 ; $j <= $#syscall_params_breakdown ; $j++ ) {
                    my $record = $syscall_params_breakdown[$j]->{'pre'};
                    if ( $record->{'type'} eq 'buffer_size' ) {
                        $buf_size_param_name = $record->{'name'};
                    }
                }
                if ( !defined($buf_size_param_name) ) {
                    die "no buffer_size provided for buffer '${param_name}' "
			. "in '$data_file' line $data_file_line";
                }
                $syscall_params_filter_vars_set .=
		    "\t\tset_var_indexed_callback2(VT_PARAM, ${i},"
			. " global_param_callback, ${param_datatype}, "
			    . "(void*)${param_name}, "
				. "(void*)${buf_size_param_name});\n";
            }
            elsif ( defined($param_makeval) && $param_makeval ne '' ) {
                $syscall_params_filter_vars_set .=
		    "\t\tset_var_indexed_callback(VT_PARAM, ${i}, "
			. "global_param_callback, ${param_datatype}, "
			    . "(void*)&${param_name});\n";
            }
        }
    }
    return $syscall_params_filter_vars_set;
}

sub prepare_parameters_definition_strings{
    my      ($data_file, $data_file_line, @syscall_params_breakdown)=@_;
    my $param_output_string            = '';
    my $param_names_string             = '';
    my $param_names_for_logging_string = '';
    for ( my $i = 0 ; $i < scalar(@syscall_params_breakdown) ; $i++ ) {
	my $record = $syscall_params_breakdown[$i]->{'pre'};
	my $param_type      = $record->{'type'};
	
	( $param_type eq "void" )
	    and die "a parameter of a function with more than "
		. "one null variable was void,"
		    . "in '$data_file' line $data_file_line";
	
	my $param_name      = $record->{'name'};
	my $param_isptr     = $record->{'isptr'};
	my $param_isarray   = $record->{'isarray'};
	my $param_array_len = $record->{'array_len'};
	
	$param_output_string .= append_param_output_type
	    ($param_type, $param_isptr, $param_isarray,
	     $param_name, $param_array_len,
	     $data_file, $data_file_line);
	
	my $param_type_specifier = &get_param_type_specifier($param_type);
	if ( defined($param_type_specifier) 
	     || ( $param_type =~ /^(const_)?struct_(.*)$/ ) 
	   )
	{
	    if ( $param_type =~ /^(const_)?struct_(.*)$/ ) {
		$struct_type = $2;
		# handle passing the members of the struct itself.
		my ($struct_params) = $structs_fields{$struct_type};
		my (@struct_params_strings) =
		    split ( ',', $struct_params );
		$param_names_for_logging_string .= "${param_name}, "
		    . "(void*)&${param_name}_logger_data, ";
	    }
	    elsif ( $param_type =~ /^(const_)?(string|buffer)$/ ) {
		
		# the ordinal number of the parameter for 
		# strings and buffers
		# 'param spec' is 'S' or 'B', which means
		# the logging code will call a callback to get it,
		# and said callback
		# will need to know what's the parameter's 
		# ordinal number, in order to
		# return this parameter and none other.
		# (in short, because we have to pass its index,
		# we cannot pass it directly)
		$param_names_for_logging_string .= "$i, ";
	    }
	    else {
		$param_names_for_logging_string .=
		    ( defined($param_name) ? "$param_name, " : "" );
	    }
	}
	if ( defined($param_name) ) {
	    $param_names_string .= "$param_name, ";
	}
    }
    
    
    # chop the last comma+space added
    $param_names_string             =~ s/\s*, $//;
    $param_names_for_logging_string =~ s/\s*, $//;
    return ($param_output_string, $param_names_string,
	    $param_names_for_logging_string);
}

sub prepare_syscall_copied_vars{
    my ($data_file, $data_file_line, $syscall_id, 
	@syscall_params_breakdown)=@_;

    # prepare code for defining variables for copying strings and structs
    # from user-space.
    my $syscall_copied_vars_def = "";
    
    # prepare code for copying strings and structs from user-space.
    my $syscall_copied_vars_copy = "";
    
    # default action to take if we hit an error is to just return,
    # since we might do not necessarily have a a label to 'goto'.
    my $syscall_copied_vars_goto_free = "\t\t\treturn retval;\n";

    # prepare code for generating process state variable setting calls.
    my $syscall_copied_vars_free = "";

    for ( my $i = 0 ; $i <= $#syscall_params_breakdown ; $i++ ) {
        my $record = $syscall_params_breakdown[$i]->{'pre'};
        my $param_type      = $record->{'type'};
        my $param_name      = $record->{'name'};
        my $param_matchable = $record->{'matchable'};
        if (   $param_matchable eq 'matchable'
	       && $param_type =~ /^((const_)?)struct_(.*)$/ )
        {
            my $param_struct_type      = $3;
	    
	    $syscall_copied_vars_free = "yes";
	    
            # prepare code for defining variables for 
            # copying strings and structs
            # from user-space.
            $syscall_copied_vars_def .=
		"\tstruct $param_struct_type kern_${param_name};\n"
		    . "\tstruct {\n"
			. "\t    char* data;\n"
			    . "\t    int struct_type;\n"
				. "\t} ${param_name}_struct_data;\n";

            $syscall_copied_vars_def .=
		"\tstruct struct_logger_data ${param_name}_logger_data;\n";
	    
            $syscall_copied_vars_copy .= "\tif (${param_name}) {\n".
		"\t\tif (copy_from_user(&kern_${param_name},"
		    . " ${param_name}, sizeof(kern_${param_name}))) {\n".
			"\t\t\tretval = -EFAULT;\n".
			    "\t\t\tgoto tracker_${syscall_id}_failed;\n".
				"\t\t}\n".
				    "\t\t${param_name}_struct_data.data ="
					. " (char*)&kern_${param_name};\n";
	    $syscall_copied_vars_copy .=
		"\t\t${param_name}_struct_data.struct_type ="
		    . " $structs_types_indexes{$param_struct_type};\n".
			"\t\t${param_name}_logger_data.param_index =" .
			    " ${i};\n".
				"\t\t${param_name}_logger_data.struct_type =";
	    $syscall_copied_vars_copy .= 
		" $structs_types_indexes{$param_struct_type};\n".
		    "\t}\n";
	    
            # make sure to generate code to free them
            # later if we run into errors
            $syscall_copied_vars_goto_free =
		"\t\t\tgoto tracker_${syscall_id}_failed;\n";
	    # if there were structs, insert the failer
	    # label and retval checking code.
	    #insert it only once, for any number of structs.
	    $syscall_copied_vars_free = "  tracker_${syscall_id}_failed:\n";
        }
    }

    return ($syscall_copied_vars_def, $syscall_copied_vars_copy, 
     $syscall_copied_vars_goto_free, $syscall_copied_vars_free);
}

sub screen_print{
    my (@syscall_params_breakdown)=@_;
    print "    params num: ", scalar(@syscall_params_breakdown), "\n";
    for ( my $i = 0 ; $i < scalar(@syscall_params_breakdown) ; $i++ ) {
        my $record = $syscall_params_breakdown[$i]->{'pre'};
        print "     param $i: ";
        print ""
          . ( ( $record->{'matchable'} eq 'matchable' ) ? "matchable " : "" )
          . "type '"
          . $record->{'type'} . "'"
          . "  name '"
          . ( defined( $record->{'name'} ) ? $record->{'name'} : "" ) . "'\n";
    }
    return 0;
}
