/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 guy keren, choo@actcom.co.il	            */
/* License: GNU Lesser General Public License                       */

/*
 *$Id: encode_decode.c,v 1.6 2002/06/15 15:16:13 mulix Exp $
 */
/*
 * Rule engine - primitive types encoding/decoding functions.
 */

#include "mem_manager_public.h"
#include "archdep.h"

/* encode the given ptr with the given size     */
/* at the given pos, as long as the buffer is   */
/* not too small.                               */
static int
do_encode(unsigned char* buf,
	  int buflen,
	  unsigned char** pos,
	  void* ptr, 
	  unsigned char size)
{
    assert(ptr && size); 
    
    if (!ptr || !size)
	return -1;
    
    if (*pos - buf <= buflen - size) {
	memcpy(*pos, ptr, size); 
	(*pos) += size;
	return 1;
    }
    return -1;
}

/* decode the given ptr at the given pos,           */
/* as long as we are not past end of buf.           */
int
do_decode(unsigned char* buf,
	  int buflen,
	  unsigned char** pos,
	  void* ptr,
	  unsigned char size)
{
    assert(ptr && size);

    if (!ptr || !size)
	return -1;
    
    if (*pos - buf <= buflen - size) {
	memcpy(ptr, *pos, size);
	(*pos) += size;
	return 1;
    }
    return -1;
}



/* encode the given unsigned char at the given pos, */
/* as long as the buffer is not too small.          */
int
encode_uchar(unsigned char* buf,
	    int buflen,
	    unsigned char** pos,
	    unsigned char c)
{
    return do_encode(buf, buflen, pos, (void*)&c, sizeof(c));
}

/* decode the given unsigned char at the given pos, */
/* as long as we are not past end of buf.           */
int
decode_uchar(unsigned char* buf,
	    int buflen,
	    unsigned char** pos,
	    unsigned char* pc)
{
    return do_decode(buf, buflen, pos, pc, sizeof(*pc));
}

/* encode the given signed char at the given pos, */
/* as long as the buffer is not too small.          */
int
encode_char(unsigned char* buf,
	    int buflen,
	    unsigned char** pos,
	    char c)
{
    return do_encode(buf, buflen, pos, (void*)&c, sizeof(c));
}

/* decode the given signed char at the given pos, */
/* as long as we are not past end of buf.           */
int
decode_char(unsigned char* buf,
	    int buflen,
	    unsigned char** pos,
	    char* pc)
{
    return do_decode(buf, buflen, pos, pc, sizeof(*pc));
}

/* encode the given int at the given pos,  */
/* as long as the buffer is not too small. */
int
encode_int(unsigned char* buf,
	   int buflen,
	   unsigned char** pos,
	   int i)
{
    return do_encode(buf, buflen, pos, (void*)&i, sizeof(i));
}

/* decode the given int at the given pos, */
/* as long as we are not past end of buf. */
int
decode_int(unsigned char* buf,
	   int buflen,
	   unsigned char** pos,
	   int* pi)
{
    return do_decode(buf, buflen, pos, pi, sizeof(*pi));
}

/* encode the given unsigned int at the given pos, */
/* as long as the buffer is not too small.         */
int
encode_uint(unsigned char* buf,
	   int buflen,
	   unsigned char** pos,
	   unsigned int i)
{
    return do_encode(buf, buflen, pos, (void*)&i, sizeof(i));
}

/* decode the given unsigned int at the given pos, */
/* as long as we are not past end of buf.          */
int
decode_uint(unsigned char* buf,
	   int buflen,
	   unsigned char** pos,
	   unsigned int* pi)
{
    return do_decode(buf, buflen, pos, pi, sizeof(*pi));
}

/* encode the given long at the given pos, */
/* as long as the buffer is not too small. */
int
encode_long(unsigned char* buf,
	    int buflen,
	    unsigned char** pos,
	    long l)
{
    return do_encode(buf, buflen, pos, (void*)&l, sizeof(l));
}

/* decode the given long at the given pos, */
/* as long aswe are not past end of buf.   */
int
decode_long(unsigned char* buf,
	    int buflen,
	    unsigned char** pos,
	    long* pl)
{
    return do_decode(buf, buflen, pos, pl, sizeof(*pl));
}

/* encode the given unsigned long at the given pos, */
/* as long as the buffer is not too small.          */
int
encode_ulong(unsigned char* buf,
	    int buflen,
	    unsigned char** pos,
	    unsigned long l)
{
    return do_encode(buf, buflen, pos, (void*)&l, sizeof(l));
}

/* decode the given unsigned long at the given pos, */
/* as long aswe are not past end of buf.            */
int
decode_ulong(unsigned char* buf,
	    int buflen,
	    unsigned char** pos,
	    unsigned long* pl)
{
    return do_decode(buf, buflen, pos, pl, sizeof(*pl));
}

/* encode the given short at the given pos, */
/* as long as the buffer is not too small.  */
int
encode_short(unsigned char* buf,
	     int buflen,
	     unsigned char** pos,
	     short s)
{
    return do_encode(buf, buflen, pos, (void*)&s, sizeof(s));
}

/* decode the given short at the given pos, */
/* as long as we are not past end of buf.   */
int
decode_short(unsigned char* buf,
	     int buflen,
	     unsigned char** pos,
	     short* ps)
{
    return do_decode(buf, buflen, pos, ps, sizeof(*ps));
}

/* encode the given unsigned short at the given pos, */
/* as long as the buffer is not too small.           */
int
encode_ushort(unsigned char* buf,
	     int buflen,
	     unsigned char** pos,
	     unsigned short s)
{
    return do_encode(buf, buflen, pos, (void*)&s, sizeof(s));
}

/* decode the given unsigned short at the given pos, */
/* as long as we are not past end of buf.            */
int
decode_ushort(unsigned char* buf,
	     int buflen,
	     unsigned char** pos,
	     unsigned short* ps)
{
    return do_decode(buf, buflen, pos, ps, sizeof(*ps));
}

/* encode the given float at the given pos, */
/* as long as the buffer is not too small.  */
int
encode_float(unsigned char* buf,
	     int buflen,
	     unsigned char** pos,
	     float f)
{
    return do_encode(buf, buflen, pos, (void*)&f, sizeof(f));
}

/* decode the given float at the given pos, */
/* as long as we are nto past end of buf.   */
int
decode_float(unsigned char* buf,
	     int buflen,
	     unsigned char** pos,
	     float* pf)
{
    return do_decode(buf, buflen, pos, pf, sizeof(*pf));
}

/* encode the given string at the given pos, */
/* as long as the buffer is not too small.   */
int
encode_string(unsigned char* buf,
	      int buflen,
	      unsigned char** pos,
	      const char* str)
{
    /* handle NULL strings */
    int len = (str ? strlen(str) : 0);
    
    /* encode the string's len, and then the string - */
    /* withut the string's delimiting '\0' character. */
    if (*pos - buf <= buflen - (sizeof(int) + len)) {
	memcpy(*pos, &len, sizeof(int));
	(*pos) += sizeof(int);
	if (str) {
	    memcpy(*pos, str, len);
	    (*pos) += len;
	}
	return 1;
    }
    
    return -1;
}

/* decode the given string at the given pos, */
/* as long as we are not past end of buf.    */
int
decode_string(unsigned char* buf,
	      int buflen,
	      unsigned char** pos,
	      char** pstr)
{
    int rc = 0;
    int len;
    
    /* decode the string's len, and then the string -    */
    /* without the string's delimiting '\0' character.   */
    /* finally, append a '\0' char to the end of string. */
    rc = decode_int(buf, buflen, pos, &len);
    if (rc > 0) {
	/* handle NULL strings */
	if (len == 0) {
	    *pstr = NULL;
	    return 1;
	}
	if (*pos - buf <= buflen - len) {
	    (*pstr) = mem_alloc(len+1);
	    if (*pstr) {
		memcpy(*pstr, *pos, len);
		(*pos) += len;
		*(*pstr + len) = '\0';
		rc = 1;
	    }
	    else {
		rc = 0;
	    }
	}
	else {
	    rc = -1;
	}
    }
    
    return rc;
} 

/* encode the given buffer 'in' at the given pos, */
/* as long as the destination buffer 'out' is not too small. */
int
encode_buffer(unsigned char* out,
	      int buflen,
	      unsigned char** pos,
	      const void* in, 
	      unsigned int size)
{
    /* handle NULL buffers */
    int len = (in ? size : 0);
    
    /* encode the buffer's len, and then the buffer */ 
    if (*pos - out <= buflen - (sizeof(int) + len)) {
	memcpy(*pos, &len, sizeof(int));
	(*pos) += sizeof(int);
	if (out) {
	    memcpy(*pos, in, len);
	    (*pos) += len;
	}
	return 1;
    }
    return -1;
}

/* decode the given buffer '*pdst' at the given pos, */
/* as long as we are not past end of the buf 'src'.  */
int
decode_buffer(unsigned char* src,
	      int buflen,
	      unsigned char** pos,
	      void** pdst, 
	      unsigned int* psize)
{
    int rc = 0;
    int len;
    
    /* decode the buffer's len, and then the buffer */ 
    rc = decode_int(src, buflen, pos, &len);
    if (rc <= 0)
	    return rc; 

    /* handle NULL strings */
    if (len == 0) {
	    *pdst = NULL;
	    *psize = 0; 
	    return 1;
    }
    if (*pos - src > buflen - len)
	    return -1; 

    (*pdst) = mem_alloc(len);
    if (*pdst) {
	    memcpy(*pdst, *pos, len);
	    *psize = len; 
	    (*pos) += len;
	    rc = 1;
    }
    else {
	    rc = 0;
    }
    return rc; 
}
