/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 guy keren, choo@actcom.co.il	            */
/* License: GNU Lesser General Public License                       */

/*
 * $Id: filter_eval.c,v 1.19 2002/06/15 15:16:13 mulix Exp $
 */
/*
 * Rule engine - implementation of filter node evaluation.
 */

#include "mem_manager_public.h"
#include "rule_engine_private.h"
#include "filter_vars_public.h"

#include "archdep.h"
/*
 * forward declarations.
 */

/* you can call do_sct_eval_filter_node when f and result are not null */
sct_exp_error_t do_sct_eval_filter_node(struct tracker_filter_node* f,
					struct tracker_filter_node* result);

/*
 * Private macros.
 */

/* typed result macros. */
#define	RESULT_DATA_TYPE(res)	((res)->u.constant.type)
#define	RESULT_BOOL(res)	((res)->u.constant.u.b)
#define	RESULT_CHAR(res)	((res)->u.constant.u.c)
#define	RESULT_SHORT(res)	((res)->u.constant.u.s)
#define	RESULT_USHORT(res)	((res)->u.constant.u.us)
#define	RESULT_INT(res)		((res)->u.constant.u.i)
#define	RESULT_UINT(res)	((res)->u.constant.u.ui)
#define	RESULT_LONG(res)	((res)->u.constant.u.l)
#define	RESULT_ULONG(res)	((res)->u.constant.u.ul)
#define	RESULT_FLOAT(res)	((res)->u.constant.u.f)
#define	RESULT_STRING(res)	((res)->u.constant.u.str)
#define RESULT_BUFFER(res)      ((res)->u.constant.u.buf.data)
#define RESULT_BUFFER_SIZE(res) ((res)->u.constant.u.buf.size)

/*
 * Private functions.
 */

/* number of bytes each data type occupies.   */
/* must be in sycn with the sct_data_type_t enum. */
static int data_type_sizes[] = {
    1,			/* DT_BOOL   */
    sizeof(char),		/* DT_CHAR   */
    sizeof(int),		/* DT_INT    */
    sizeof(unsigned int),	/* DT_UINT   */
    sizeof(short),		/* DT_SHORT  */
    sizeof(unsigned short),	/* DT_USHORT */
    sizeof(long),		/* DT_LONG   */
    sizeof(unsigned long),	/* DT_ULONG  */
    sizeof(float),		/* DT_FLOAT  */
    sizeof(char*),		/* DT_STRING */
    sizeof(void*)               /* DT_BUFFER */ 
};

/* compare two data types. return 0 if they are equal, -1  */
/* if the first takes more space, 1 if the second takes    */
/* more space.                                             */
int
typecmp(sct_data_type_t t1, sct_data_type_t t2)
{
    int ts1 = 0;
    int ts2 = 0;

    if (t1 >= 0 && t1 < DT_LAST_TYPE) {
	ts1 = data_type_sizes[t1];
    }
    if (t2 >= 0 && t2 < DT_LAST_TYPE) {
	ts2 = data_type_sizes[t2];
    }

    return (ts1 < ts2 ? -1 : (ts1 == ts2 ? 0 : 1));
}

/* cast a constant value into a boolean value, similarly   */
/* to how this is doen in C. Exception for a string value. */
int
cast_bool(struct tracker_filter_node* val)
{
    int result = TF_FALSE;

    switch(CONST_DATA_TYPE(val)) {
    case DT_BOOL:
	result = CONST_BOOL(val);
	break;
    case DT_CHAR:
	result = CONST_CHAR(val) ? TF_TRUE : TF_FALSE;
	break;
    case DT_SHORT:
	result = CONST_SHORT(val) ? TF_TRUE : TF_FALSE;
	break;
    case DT_USHORT:
	result = CONST_USHORT(val) ? TF_TRUE : TF_FALSE;
	break;
    case DT_INT:
	result = CONST_INT(val) ? TF_TRUE : TF_FALSE;
	break;
    case DT_UINT:
	result = CONST_UINT(val) ? TF_TRUE : TF_FALSE;
	break;
    case DT_LONG:
	result = CONST_LONG(val) ? TF_TRUE : TF_FALSE;
	break;
    case DT_ULONG:
	result = CONST_ULONG(val) ? TF_TRUE : TF_FALSE;
	break;
    case DT_FLOAT:
	result = CONST_FLOAT(val) ? TF_TRUE : TF_FALSE;
	break;
    case DT_STRING:
	result = CONST_STRING(val) ? TF_TRUE : TF_FALSE;
	break;
    case DT_BUFFER:
        result = CONST_BUFFER(val) ? TF_TRUE : TF_FALSE; 
	break; 
    case DT_MULTI:
	/* TODO - how is a multi cast to a bool? */
	result = 0;
	break;
    default:
	/* shouldn't happen. */
	result = 0;
	break;
    }

    return result;
}

/* cast a constant value into a char value, similarly      */
/* to how this is doen in C. Exception for a string value. */
char
cast_char(struct tracker_filter_node* val)
{
    char result = 0;

    switch(CONST_DATA_TYPE(val)) {
    case DT_BOOL:
	result = (char)CONST_BOOL(val);
	break;
    case DT_CHAR:
	result = CONST_CHAR(val);
	break;
    case DT_SHORT:
	result = (char)CONST_SHORT(val);
	break;
    case DT_USHORT:
	result = (char)CONST_USHORT(val);
	break;
    case DT_INT:
	result = (char)CONST_INT(val);
	break;
    case DT_UINT:
	result = (char)CONST_UINT(val);
	break;
    case DT_LONG:
	result = (char)CONST_LONG(val);
	break;
    case DT_ULONG:
	result = (char)CONST_ULONG(val);
	break;
    case DT_FLOAT:
	result = (char)CONST_FLOAT(val);
	break;
    case DT_STRING:
	/* TODO - how is a string cast to a char? */
	result = 0;
	break;
    case DT_BUFFER:
	/* TODO - how is a buffer cast to a char? */
	result = 0;
	break;
    case DT_MULTI:
	/* TODO - how is a multi cast to a char? */
	result = 0;
	break;
    default:
	/* shouldn't happen. */
	result = '\0';
	break;
    }

    return result;
}

/* cast a constant value into a short value, similarly     */
/* to how this is doen in C. Exception for a string value. */
short
cast_short(struct tracker_filter_node* val)
{
    short result = 0;

    switch(CONST_DATA_TYPE(val)) {
    case DT_BOOL:
	result = CONST_BOOL(val);
	break;
    case DT_CHAR:
	result = CONST_CHAR(val);
	break;
    case DT_SHORT:
	result = CONST_SHORT(val);
	break;
    case DT_USHORT:
	result = (short)CONST_USHORT(val);
	break;
    case DT_INT:
	result = (short)CONST_INT(val);
	break;
    case DT_UINT:
	result = (short)CONST_UINT(val);
	break;
    case DT_LONG:
	result = (short)CONST_LONG(val);
	break;
    case DT_ULONG:
	result = (short)CONST_ULONG(val);
	break;
    case DT_FLOAT:
	result = (short)CONST_FLOAT(val);
	break;
    case DT_STRING:
	/* TODO - write an atoi function (for kernel mode) */
	/*        and use it here.                         */
	result = 0;
	break;
    case DT_BUFFER:
	result = 0;
	break; 
    case DT_MULTI:
	/* TODO - how is a multi cast to a short? */
	result = 0;
	break;
    default:
	/* shouldn't happen. */
	result = 0;
	break;
    }

    return result;
}

/* cast a constant value into an unsigned short value, similarly */
/* to how this is doen in C. Exception for a string value.       */
unsigned short
cast_ushort(struct tracker_filter_node* val)
{
    short result = 0;

    switch(CONST_DATA_TYPE(val)) {
    case DT_BOOL:
	result = CONST_BOOL(val);
	break;
    case DT_CHAR:
	result = CONST_CHAR(val);
	break;
    case DT_SHORT:
	result = (unsigned short)CONST_SHORT(val);
	break;
    case DT_USHORT:
	result = CONST_USHORT(val);
	break;
    case DT_INT:
	result = (unsigned short)CONST_INT(val);
	break;
    case DT_UINT:
	result = (unsigned short)CONST_UINT(val);
	break;
    case DT_LONG:
	result = (unsigned short)CONST_LONG(val);
	break;
    case DT_ULONG:
	result = (unsigned short)CONST_ULONG(val);
	break;
    case DT_FLOAT:
	result = (unsigned short)CONST_FLOAT(val);
	break;
    case DT_STRING:
	/* TODO - write an atoi function (for kernel mode) */
	/*        and use it here.                         */
	result = 0;
	break;
    case DT_BUFFER:
	result = 0;
	break; 
    case DT_MULTI:
	/* TODO - how is a multi cast to a ushort? */
	result = 0;
	break;
    default:
	/* shouldn't happen. */
	result = 0;
	break;
    }

    return result;
}

/* cast a constant value into an int value, similarly      */
/* to how this is doen in C. Exception for a string value. */
int
cast_int(struct tracker_filter_node* val)
{
    int result = 0;

    switch(CONST_DATA_TYPE(val)) {
    case DT_BOOL:
	result = CONST_BOOL(val);
	break;
    case DT_CHAR:
	result = CONST_CHAR(val);
	break;
    case DT_SHORT:
	result = CONST_SHORT(val);
	break;
    case DT_USHORT:
	result = (int)CONST_USHORT(val);
	break;
    case DT_INT:
	result = CONST_INT(val);
	break;
    case DT_UINT:
	result = (int)CONST_UINT(val);
	break;
    case DT_LONG:
	result = (int)CONST_LONG(val);
	break;
    case DT_ULONG:
	result = (int)CONST_ULONG(val);
	break;
    case DT_FLOAT:
	result = (int)CONST_FLOAT(val);
	break;
    case DT_STRING:
	/* TODO - write an atoi function (for kernel mode) */
	/*        and use it here.                         */
	result = 0;
	break;
    case DT_BUFFER:
	result = 0;
	break; 
    case DT_MULTI:
	/* TODO - how is a multi cast to a int? */
	result = 0;
	break;
    default:
	/* shouldn't happen. */
	result = 0;
	break;
    }

    return result;
}

/* cast a constant value into an unsigned int value, similarly */
/* to how this is doen in C. Exception for a string value.     */
unsigned int
cast_uint(struct tracker_filter_node* val)
{
    int result = 0;

    switch(CONST_DATA_TYPE(val)) {
    case DT_BOOL:
	result = CONST_BOOL(val);
	break;
    case DT_CHAR:
	result = CONST_CHAR(val);
	break;
    case DT_SHORT:
	result = (unsigned int)CONST_SHORT(val);
	break;
    case DT_USHORT:
	result = (unsigned int)CONST_USHORT(val);
	break;
    case DT_INT:
	result = (unsigned int)CONST_INT(val);
	break;
    case DT_UINT:
	result = (unsigned int)CONST_UINT(val);
	break;
    case DT_LONG:
	result = (unsigned int)CONST_LONG(val);
	break;
    case DT_ULONG:
	result = (unsigned int)CONST_ULONG(val);
	break;
    case DT_FLOAT:
	result = (unsigned int)CONST_FLOAT(val);
	break;
    case DT_STRING:
	/* TODO - write an atoi function (for kernel mode) */
	/*        and use it here.                         */
	result = 0;
	break;
    case DT_BUFFER:
	result = 0;
	break; 
    case DT_MULTI:
	/* TODO - how is a multi cast to a uint? */
	result = 0;
	break;
    default:
	/* shouldn't happen. */
	result = 0;
	break;
    }

    return result;
}

/* cast a constant value into a long value, similarly      */
/* to how this is doen in C. Exception for a string value. */
long
cast_long(struct tracker_filter_node* val)
{
    long result = 0;

    switch(CONST_DATA_TYPE(val)) {
    case DT_BOOL:
	result = CONST_BOOL(val);
	break;
    case DT_CHAR:
	result = CONST_CHAR(val);
	break;
    case DT_SHORT:
	result = CONST_SHORT(val);
	break;
    case DT_USHORT:
	result = (long)CONST_USHORT(val);
	break;
    case DT_INT:
	result = CONST_INT(val);
	break;
    case DT_UINT:
	result = (long)CONST_UINT(val);
	break;
    case DT_LONG:
	result = CONST_LONG(val);
	break;
    case DT_ULONG:
	result = (long)CONST_ULONG(val);
	break;
    case DT_FLOAT:
	result = (long)CONST_FLOAT(val);
	break;
    case DT_STRING:
	/* TODO - write an atoi function (for kernel mode) */
	/*        and use it here.                         */
	result = 0;
	break;
    case DT_BUFFER:
	result = 0;
	break; 
    case DT_MULTI:
	/* TODO - how is a multi cast to a long? */
	result = 0;
	break;
    default:
	/* shouldn't happen. */
	result = 0;
	break;
    }

    return result;
}

/* cast a constant value into an unsigned long value, similarly */
/* to how this is doen in C. Exception for a string value.      */
unsigned long
cast_ulong(struct tracker_filter_node* val)
{
    long result = 0;

    switch(CONST_DATA_TYPE(val)) {
    case DT_BOOL:
	result = CONST_BOOL(val);
	break;
    case DT_CHAR:
	result = CONST_CHAR(val);
	break;
    case DT_SHORT:
	result = (unsigned long)CONST_SHORT(val);
	break;
    case DT_USHORT:
	result = (unsigned long)CONST_USHORT(val);
	break;
    case DT_INT:
	result = (unsigned long)CONST_INT(val);
	break;
    case DT_UINT:
	result = (unsigned long)CONST_UINT(val);
	break;
    case DT_LONG:
	result = (unsigned long)CONST_LONG(val);
	break;
    case DT_ULONG:
	result = CONST_ULONG(val);
	break;
    case DT_FLOAT:
	result = (unsigned long)CONST_FLOAT(val);
	break;
    case DT_STRING:
	/* TODO - write an atoi function (for kernel mode) */
	/*        and use it here.                         */
	result = 0;
	break;
    case DT_BUFFER:
	result = 0;
	break; 
    case DT_MULTI:
	/* TODO - how is a multi cast to a ulong? */
	result = 0;
	break;
    default:
	/* shouldn't happen. */
	result = 0;
	break;
    }

    return result;
}

/* cast a constant value into a string value, similarly    */
/* to how this is doen in C. Strings are strings. numbers  */
/* are being stringified.                                  */
/* the caller should free the resulting string.            */
char*
cast_string(struct tracker_filter_node* val)
{
    char* result = NULL;

    switch(CONST_DATA_TYPE(val)) {
    case DT_BOOL:
	result = mem_alloc(1+1);
	if (!result)
	    break;
	result[0] = (CONST_BOOL(val) == 1 ? '1' : '0');
	result[1] = '\0';
	break;
    case DT_CHAR:
	result = mem_alloc(1+1);
	if (!result)
	    break;
	result[0] = CONST_CHAR(val);
	result[1] = '\0';
	break;
    case DT_SHORT:
	result = mem_alloc(5+1);
	if (!result)
	    break;
	sprintf(result, "%hd", CONST_SHORT(val));
	break;
    case DT_USHORT:
	result = mem_alloc(5+1);
	if (!result)
	    break;
	sprintf(result, "%hu", CONST_USHORT(val));
	break;
    case DT_INT:
	result = mem_alloc(20+1);
	if (!result)
	    break;
	sprintf(result, "%d", CONST_INT(val));
	break;
    case DT_UINT:
	result = mem_alloc(20+1);
	if (!result)
	    break;
	sprintf(result, "%u", CONST_UINT(val));
	break;
    case DT_LONG:
	result = mem_alloc(20+1);
	if (!result)
	    break;
	sprintf(result, "%ld", CONST_LONG(val));
	break;
    case DT_ULONG:
	result = mem_alloc(20+1);
	if (!result)
	    break;
	sprintf(result, "%lu", CONST_ULONG(val));
	break;
    case DT_FLOAT:
	result = mem_alloc(50+1);
	if (!result)
	    break;
	sprintf(result, "%f", CONST_FLOAT(val));
	break;
    case DT_STRING:
	if (!CONST_STRING(val))
	    break; 
	result = mem_alloc(strlen(CONST_STRING(val))+1);
	if (!result)
	    break;
	strcpy(result, CONST_STRING(val));
	break;
    case DT_BUFFER:
	if (!CONST_BUFFER(val))
	    break; 
	result = sct_buffer_to_string(CONST_BUFFER(val), 
				      BUFFER_SIZE(val)); 
	break; 
    case DT_MULTI:
	/* TODO - how is a multi cast to a string? */
	result = NULL;
	break;
    default:
	/* shouldn't happen. */
	result = NULL; 
	break;
    }

    return result;
}

void cast_buffer(struct tracker_filter_node* val, void** buf, 
		 unsigned int* size)
{
    void* result; 

    *buf = NULL; 
    *size = 0; 

    /* only handle the BUFFER case explicitly, the rest is handled like we handle strings */ 
    switch(CONST_DATA_TYPE(val)) {
    case DT_BOOL:
    case DT_CHAR:
    case DT_SHORT:
    case DT_USHORT:
    case DT_INT:
    case DT_UINT:
    case DT_LONG:
    case DT_ULONG:
    case DT_FLOAT:
    case DT_STRING:
    case DT_MULTI:
    default: 
    {
	    char* str = cast_string(val); 
	    *buf = str; 
	    *size = (str ? strlen(str) : 0); 
	    return; 
    }
    }
    
    if (!CONST_BUFFER(val))
	    return; 
    result = mem_alloc(BUFFER_SIZE(val)); 
    if (!result)
	    return; 
    memcpy(result, CONST_BUFFER(val), BUFFER_SIZE(val)); 
    *buf = result; 
    *size = BUFFER_SIZE(val); 
}

sct_exp_error_t
filter_node_cmp(sct_filter_node_t* val1, 
		sct_filter_node_t* val2, 
		int* cmp)
{
    sct_exp_error_t rc = ERR_NONE;
    sct_data_type_t t1 = val1->data_type;
    sct_data_type_t t2 = val2->data_type;
    
    if (t1 == DT_STRING && t2 == DT_STRING) {
	const char* str1 = CONST_STRING(val1); 
	const char* str2 = CONST_STRING(val2); 
	if (str1 && str2) 
	    *cmp = strcmp(str1, str2);
	else if (str1)
	    *cmp = -1;
	else if (str2)
	    *cmp = 1;
	else 
	    *cmp = 0;
    }
    else if (t1 == DT_STRING || t2 == DT_STRING) {
	/* one is a string and the other is not - type mismatch. */
	rc = ERR_TYPE_MISMATCH;
    } 
    else if (t1 == DT_BUFFER && t2 == DT_BUFFER) {
	void* b1 = CONST_BUFFER(val1);     
	void* b2 = CONST_BUFFER(val2); 
	unsigned int b1sz = BUFFER_SIZE(val1); 
	unsigned int b2sz = BUFFER_SIZE(val2); 
	if (b1 && b2) {
	    int c = memcmp(b1, b2, sct_min(b1sz, b2sz));
	    if (c || (b1sz == b2sz)) {/* not equal in the first n bytes */
		*cmp = c;        /* or both of size n */ 
	    } else { /* equal in first n bytes, but different sizes */
		*cmp = (b1sz < b2sz ? -1 : 1);
	    }
	} else if (b1) {
	    *cmp = -1; 
	} else if (b2) {
	    *cmp = 1; 
	} else {
	    /* if we get here, they're both NULL */ 
	    *cmp = 0; 
	}
    }
    else if (t1 == DT_BUFFER || t2 == DT_BUFFER) {
	    rc = ERR_TYPE_MISMATCH; 
    }
    else if (t1 == DT_FLOAT && t2 == DT_FLOAT) {
        float f1 = CONST_FLOAT(val1); 
	float f2 = CONST_FLOAT(val2); 
	*cmp = (f1 < f2 ? -1 : ((f1 == f2) ? 0 : 1));
    } 
    else if (t1 == DT_FLOAT || t2 == DT_FLOAT) {
	rc = ERR_TYPE_MISMATCH;
    } 
    else if (t1 == DT_USHORT || t1 == DT_UINT || t1 == DT_ULONG ||
	     t2 == DT_USHORT || t2 == DT_UINT || t2 == DT_ULONG) {
	unsigned long ul1 = cast_ulong(val1);
	unsigned long ul2 = cast_ulong(val2);
	/* DONT use sct_sign for unsigned types */
	*cmp = (ul1 < ul2 ? -1 : ((ul1 == ul2) ? 0 : 1));
    }
    else {
	long l1 = cast_long(val1);
	long l2 = cast_long(val2);
	*cmp = (l1 < l2 ? -1 : ((l1 == l2) ? 0 : 1));
    } 

    return rc;
}

/* check if the given string matches the givven pattern. */
/* currently implements a very trivial strstr matching.  */
/* TODO - make real pattern matching.                    */
int
str_pattern_match(const char* str, const char* pat)
{
    if (str && pat && strstr(str, pat))
	return 1;
    else
	return 0;
}

int 
buf_pattern_match(const void* v1, unsigned int v1sz, 
		  const void* pat, unsigned int patsz)
{

	unsigned char* haystack_start = (unsigned char*)v1; 
	unsigned char* haystack_end = (unsigned char*)v1 + v1sz; /* one beyond the end of the buffer */
	void* res;
     
	if (!v1 || !pat || !v1sz || !patsz)
		return 0; 
     
	/* check if the first char of patern is in haystack */
	while ((res = memchr(haystack_start, *(unsigned char*)pat, 
			     haystack_end - haystack_start)) != NULL) {
		/* if it is, check if the rest are as well */
		/* make sure we are not accessing beyond the end of haystack */
		if ((unsigned char*)res + patsz >= haystack_end)
			return 0;
		
		if (memcmp(res, pat, patsz) == 0)
			return 1; 
		
		haystack_start = ((unsigned char*)res)+1;
		if (haystack_start == haystack_end)
		    return 0; 
	}
	return 0; 
}

/* evaluate the given constant node.		  */
sct_exp_error_t
sct_eval_filter_node_const(struct tracker_filter_node* f,
		       struct tracker_filter_node* result)
{
    sct_exp_error_t rc = ERR_NONE;

    result->type = NT_CONST;
    result->data_type = f->data_type;
    RESULT_DATA_TYPE(result) = f->data_type;
    switch (result->data_type) {
    case DT_BOOL:
	RESULT_BOOL(result) = CONST_BOOL(f);
	break;
    case DT_CHAR:
	RESULT_CHAR(result) = CONST_CHAR(f);
	break;
    case DT_INT:
	RESULT_INT(result) = CONST_INT(f);
	break;
    case DT_UINT:
	RESULT_UINT(result) = CONST_UINT(f);
	break;
    case DT_SHORT:
	RESULT_SHORT(result) = CONST_SHORT(f);
	break;
    case DT_USHORT:
	RESULT_USHORT(result) = CONST_USHORT(f);
	break;
    case DT_LONG:
	RESULT_LONG(result) = CONST_LONG(f);
	break;
    case DT_ULONG:
	RESULT_ULONG(result) = CONST_ULONG(f);
	break;
    case DT_FLOAT:
	RESULT_FLOAT(result) = CONST_FLOAT(f);
	break;
    case DT_STRING:
    {
	const char* str = CONST_STRING(f);
	if (str) {
	    RESULT_STRING(result) = mem_alloc(strlen(str)+1);
	    if (!RESULT_STRING(result))
		break;
	    strcpy(RESULT_STRING(result), str);
	}
	else {
	    RESULT_STRING(result) = NULL;
	}
    }
    break;
    case DT_BUFFER: 
    {
	void* src = CONST_BUFFER(f); 
	if (src) {
	    RESULT_BUFFER(result) = mem_alloc(BUFFER_SIZE(f)); 
	    if (!RESULT_BUFFER(result))
		break; 
	    RESULT_BUFFER_SIZE(result) = BUFFER_SIZE(f); 
	    memcpy(RESULT_BUFFER(result), CONST_BUFFER(f), BUFFER_SIZE(f)); 
	} else {
            RESULT_BUFFER(result) = NULL; 
	    RESULT_BUFFER_SIZE(result) = 0; 
	}
    }
    break; 
    default:
	    /* shouldn't happen. */
	rc = ERR_INVALID_DATA_TYPE;
	break;
    }

    return rc;
}

/* evaluate the given variable node.		  */
sct_exp_error_t
sct_eval_filter_node_var(struct tracker_filter_node* f,
			 struct tracker_filter_node* result)
{
    sct_exp_error_t rc = ERR_NONE;
    sct_filter_node_t* var_val = NULL;
    sct_filter_node_t* field = NULL;
    int free_field = 0;
    
    result->type = NT_CONST;
    result->data_type = f->data_type;
    RESULT_DATA_TYPE(result) = f->data_type;

    switch (f->u.var.type) {
    case VT_PARAM:
	if (f->u.var.field_index >= 0) {
	    field = get_var_indexed_field(f->u.var.type,
					  f->u.var.var_index,
					  f->u.var.field_index,
					  f->u.var.struct_type);
	    if (field)
		free_field = 1;
	}
	if (!field) {
	    var_val = get_var_indexed(f->u.var.type, f->u.var.var_index);
	}
	break;
    default:
	var_val = get_var(f->u.var.type);
	break;
    }

    if (!var_val && !field) {
	rc = ERR_INVALID_VAR;
	goto return_sct_eval_filter_node_var;
    }

    if (f->u.var.field_index < 0) {
	rc = sct_eval_filter_node_const(var_val, result);
	goto return_sct_eval_filter_node_var;
    }

    /* field index is 0 or bigger, so the type must be NT_VECTOR */
    if ( (var_val && var_val->type != NT_VECTOR) && !field) {
	rc = ERR_TYPE_MISMATCH;
	goto return_sct_eval_filter_node_var;
    }

    if (!field) {
        field = sct_get_filter_node_at(var_val, f->u.var.field_index);
    }
    if (field) {
	rc = sct_eval_filter_node_const(field, result);
    } else {
	rc = ERR_TYPE_MISMATCH;
    }
    
 return_sct_eval_filter_node_var:
    if (free_field) {
	sct_free_filter_node(field);
    }
    return rc;
}


/*
 * Evaluating operators. All 'values' are constants.
 */

sct_exp_error_t
eval_op_equal(struct tracker_filter_node* f_op,
	      struct tracker_filter_node* val1,
	      struct tracker_filter_node* val2,
	      struct tracker_filter_node* result)
{
    int cmp;
    sct_exp_error_t rc;

    result->type = NT_CONST;
    RESULT_DATA_TYPE(result) = DT_BOOL;

    rc = filter_node_cmp(val1, val2, &cmp);
    if (rc == ERR_NONE){
	if (cmp == 0)
	    RESULT_BOOL(result) = TF_TRUE;
	else 
	    RESULT_BOOL(result) = TF_FALSE;
    }
    return rc;
}

sct_exp_error_t
eval_op_less(struct tracker_filter_node* f_op,
	     struct tracker_filter_node* val1,
	     struct tracker_filter_node* val2,
	     struct tracker_filter_node* result)
{
    int cmp;
    sct_exp_error_t rc;
    
    result->type = NT_CONST;
    RESULT_DATA_TYPE(result) = DT_BOOL;

    rc = filter_node_cmp(val1, val2, &cmp);
    if (rc == ERR_NONE){
	if (cmp == -1)
	    RESULT_BOOL(result) = TF_TRUE;
	else 
	    RESULT_BOOL(result) = TF_FALSE;
    }
    return rc;
}

sct_exp_error_t
eval_op_greater(struct tracker_filter_node* f_op,
		struct tracker_filter_node* val1,
		struct tracker_filter_node* val2,
		struct tracker_filter_node* result)
{
    int cmp;
    sct_exp_error_t rc;
    
    result->type = NT_CONST;
    RESULT_DATA_TYPE(result) = DT_BOOL;

    rc = filter_node_cmp(val1, val2, &cmp);
    if (rc == ERR_NONE){
	if (cmp == 1)
	    RESULT_BOOL(result) = TF_TRUE;
	else 
	    RESULT_BOOL(result) = TF_FALSE;
    }
    return rc;
}

sct_exp_error_t
eval_op_less_equal(struct tracker_filter_node* f_op,
		   struct tracker_filter_node* val1,
		   struct tracker_filter_node* val2,
		   struct tracker_filter_node* result)
{
    /* less_equal is the same as !(greater) */
    sct_exp_error_t rc = eval_op_greater(f_op, val1, val2, result);

    if (rc == ERR_NONE) {
	int eq_res = RESULT_BOOL(result);
	RESULT_BOOL(result) = (eq_res ? TF_FALSE : TF_TRUE);
    }

    return rc;
}

sct_exp_error_t
eval_op_greater_equal(struct tracker_filter_node* f_op,
		      struct tracker_filter_node* val1,
		      struct tracker_filter_node* val2,
		      struct tracker_filter_node* result)
{
    /* greater_equal is the same as !(less) */
    sct_exp_error_t rc = eval_op_less(f_op, val1, val2, result);

    if (rc == ERR_NONE) {
	int eq_res = RESULT_BOOL(result);
	RESULT_BOOL(result) = (eq_res ? TF_FALSE : TF_TRUE);
    }

    return rc;
}

sct_exp_error_t
eval_op_not_equal(struct tracker_filter_node* f_op,
		  struct tracker_filter_node* val1,
		  struct tracker_filter_node* val2,
		  struct tracker_filter_node* result)
{
    /* not_equal is the same as !(equal) */
    sct_exp_error_t rc = eval_op_equal(f_op, val1, val2, result);

    if (rc == ERR_NONE) {
	int eq_res = RESULT_BOOL(result);
	RESULT_BOOL(result) = (eq_res ? TF_FALSE : TF_TRUE);
    }

    return rc;
}

static void 
swap_filter_node(sct_filter_node_t** f1, sct_filter_node_t** f2, 
		 sct_data_type_t* dt1, sct_data_type_t* dt2)
{
    /* swap the values and types. */
    sct_data_type_t t_tmp = *dt1;
    struct tracker_filter_node* val_tmp = *f1;
    *f1 = *f2;
    *f2 = val_tmp;
    *dt1 = *dt2;
    *dt2 = t_tmp;
}

sct_exp_error_t
eval_op_plus(struct tracker_filter_node* f_op,
	     struct tracker_filter_node* val1,
	     struct tracker_filter_node* val2,
	     struct tracker_filter_node* result)
{
    sct_exp_error_t rc = ERR_NONE;
    sct_data_type_t t1 = val1->data_type;
    sct_data_type_t t2 = val2->data_type;

    result->type = NT_CONST;
	
    /* TODO - implement string concatenation here? */
    if (t1 == DT_STRING || t2 == DT_STRING) {
	rc = ERR_TYPE_MISMATCH;
    }
    else if (t1 == DT_BOOL || t2 == DT_BOOL) {
	/* cannot use '+' over bool input. */
	rc = ERR_TYPE_MISMATCH;
    }
    else {
	/* make sure the larger data type is in 'val1'. */
	int t2_is_larger = (typecmp(t1, t2) < 0);
	if (t2_is_larger) 
	    swap_filter_node(&val1, &val2, &t1, &t2);
	
	switch (t1) {
	case DT_LONG:
	    RESULT_LONG(result) =
		CONST_LONG(val1) + cast_long(val2);
	    result->data_type = DT_LONG;
	    RESULT_DATA_TYPE(result) = DT_LONG;
	    break;
	case DT_ULONG:
	    RESULT_ULONG(result) =
		CONST_ULONG(val1) + cast_ulong(val2);
	    result->data_type = DT_ULONG;
	    RESULT_DATA_TYPE(result) = DT_ULONG;
	    break;
	case DT_INT:
	    RESULT_INT(result) =
		CONST_INT(val1) + cast_int(val2);
	    result->data_type = DT_INT;
	    RESULT_DATA_TYPE(result) = DT_INT;
	    break;
	case DT_UINT:
	    RESULT_UINT(result) =
		CONST_UINT(val1) + cast_uint(val2);
	    result->data_type = DT_UINT;
	    RESULT_DATA_TYPE(result) = DT_UINT;
	    break;
	case DT_SHORT:
	    RESULT_SHORT(result) =
		CONST_SHORT(val1) + cast_short(val2);
	    result->data_type = DT_SHORT;
	    RESULT_DATA_TYPE(result) = DT_SHORT;
	    break;
	case DT_USHORT:
	    RESULT_USHORT(result) =
		CONST_USHORT(val1) + cast_ushort(val2);
	    result->data_type = DT_USHORT;
	    RESULT_DATA_TYPE(result) = DT_USHORT;
	    break;
	case DT_CHAR:
	    RESULT_CHAR(result) =
		CONST_CHAR(val1) + cast_char(val2);
	    result->data_type = DT_CHAR;
	    RESULT_DATA_TYPE(result) = DT_CHAR;
	    break;
	default:
	    /* shouldn't happen. */
	    rc = ERR_TYPE_MISMATCH;
	    break;
	}
    }

    return rc;
}

sct_exp_error_t
eval_op_minus(struct tracker_filter_node* f_op,
	      struct tracker_filter_node* val1,
	      struct tracker_filter_node* val2,
	      struct tracker_filter_node* result)
{
    sct_exp_error_t rc = ERR_NONE;
    sct_data_type_t t1 = val1->data_type;
    sct_data_type_t t2 = val2->data_type;

    result->type = NT_CONST;
	
    /* TODO - implement string de-concatenation here? */
    if (t1 == DT_STRING || t2 == DT_STRING || 
	t1 == DT_BUFFER || t2 == DT_BUFFER) {
	rc = ERR_TYPE_MISMATCH;
    }
    else if (t1 == DT_BOOL || t2 == DT_BOOL) {
	/* cannot use '-' over bool input. */
	rc = ERR_TYPE_MISMATCH;
    }
    else {
	/* make sure the larger data type is in 'val1'. */
	int t2_is_larger = (typecmp(t1, t2) < 0);
	if (t2_is_larger) 
	    swap_filter_node(&val1, &val2, &t1, &t2);

	switch (t1) {
	case DT_LONG:
	    RESULT_LONG(result) =
		CONST_LONG(val1) - cast_long(val2);
	    result->data_type = DT_LONG;
	    RESULT_DATA_TYPE(result) = DT_LONG;
	    break;
	case DT_ULONG:
	    RESULT_ULONG(result) =
		CONST_ULONG(val1) - cast_ulong(val2);
	    result->data_type = DT_ULONG;
	    RESULT_DATA_TYPE(result) = DT_ULONG;
	    break;
	case DT_INT:
	    RESULT_INT(result) =
		CONST_INT(val1) - cast_int(val2);
	    result->data_type = DT_INT;
	    RESULT_DATA_TYPE(result) = DT_INT;
	    break;
	case DT_UINT:
	    RESULT_UINT(result) =
		CONST_UINT(val1) - cast_uint(val2);
	    result->data_type = DT_UINT;
	    RESULT_DATA_TYPE(result) = DT_UINT;
	    break;
	case DT_SHORT:
	    RESULT_SHORT(result) =
		CONST_SHORT(val1) - cast_short(val2);
	    result->data_type = DT_SHORT;
	    RESULT_DATA_TYPE(result) = DT_SHORT;
	    break;
	case DT_USHORT:
	    RESULT_USHORT(result) =
		CONST_USHORT(val1) - cast_ushort(val2);
	    result->data_type = DT_USHORT;
	    RESULT_DATA_TYPE(result) = DT_USHORT;
	    break;
	case DT_CHAR:
	    RESULT_CHAR(result) =
		CONST_CHAR(val1) - cast_char(val2);
	    result->data_type = DT_CHAR;
	    RESULT_DATA_TYPE(result) = DT_CHAR;
	    break;
	default:
	    /* shouldn't happen. */
	    rc = ERR_TYPE_MISMATCH;
	    break;
	}
    }

    return rc;
}

sct_exp_error_t
eval_op_shift_left(struct tracker_filter_node* f_op,
		   struct tracker_filter_node* val1,
		   struct tracker_filter_node* val2,
		   struct tracker_filter_node* result)
{
    sct_exp_error_t rc = ERR_NONE;
    sct_data_type_t t1 = val1->data_type;
    sct_data_type_t t2 = val2->data_type;

    /* we don't support shift operations for some data types. */
    if (t1 == DT_STRING || 
	t1 == DT_BOOL || 
	t2 == DT_STRING || 
	t2 == DT_BOOL ||
	t1 == DT_BUFFER || 
	t2 == DT_BUFFER) {
	rc = ERR_TYPE_MISMATCH;
    }
    else {
	long ammount = cast_long(val2);
	result->type = NT_CONST;
	result->data_type = val1->data_type;
	RESULT_DATA_TYPE(result) = val1->data_type;

	switch (result->data_type) {
	case DT_CHAR:
	    RESULT_CHAR(result) =
		CONST_CHAR(val1) << ammount;
	    break;
	case DT_INT:
	    RESULT_INT(result) =
		CONST_INT(val1) << ammount;
	    break;
	case DT_UINT:
	    RESULT_UINT(result) =
		CONST_UINT(val1) << ammount;
	    break;
	case DT_SHORT:
	    RESULT_SHORT(result) =
		CONST_SHORT(val1) << ammount;
	    break;
	case DT_USHORT:
	    RESULT_USHORT(result) =
		CONST_USHORT(val1) << ammount;
	    break;
	case DT_LONG:
	    RESULT_LONG(result) =
		CONST_LONG(val1) << ammount;
	    break;
	case DT_ULONG:
	    RESULT_ULONG(result) =
		CONST_ULONG(val1) << ammount;
	    break;
	default:
	    /* shouldn't happen. */
	    rc = ERR_INVALID_DATA_TYPE;
	    break;
	}
    }

    return rc;
}

sct_exp_error_t
eval_op_shift_right(struct tracker_filter_node* f_op,
		    struct tracker_filter_node* val1,
		    struct tracker_filter_node* val2,
		    struct tracker_filter_node* result)
{
    sct_exp_error_t rc = ERR_NONE;
    sct_data_type_t t1 = val1->data_type;
    sct_data_type_t t2 = val2->data_type;

    /* we don't support shift operations for some data types. */
    if (t1 == DT_STRING || t1 == DT_BOOL || 
	t2 == DT_STRING || t2 == DT_BOOL ||
	t1 == DT_BUFFER || t2 == DT_BUFFER) {
	rc = ERR_TYPE_MISMATCH;
    }
    else {
	long ammount = cast_long(val2);
	result->type = NT_CONST;
	result->data_type = val1->data_type;
	RESULT_DATA_TYPE(result) = val1->data_type;

	switch (result->data_type) {
	case DT_CHAR:
	    RESULT_CHAR(result) =
		CONST_CHAR(val1) >> ammount;
	    break;
	case DT_INT:
	    RESULT_INT(result) =
		CONST_INT(val1) >> ammount;
	    break;
	case DT_UINT:
	    RESULT_UINT(result) =
		CONST_UINT(val1) >> ammount;
	    break;
	case DT_SHORT:
	    RESULT_SHORT(result) =
		CONST_SHORT(val1) >> ammount;
	    break;
	case DT_USHORT:
	    RESULT_USHORT(result) =
		CONST_USHORT(val1) >> ammount;
	    break;
	case DT_LONG:
	    RESULT_LONG(result) =
		CONST_LONG(val1) >> ammount;
	    break;
	case DT_ULONG:
	    RESULT_ULONG(result) =
		CONST_ULONG(val1) >> ammount;
	    break;
	default:
	    /* shouldn't happen. */
	    rc = ERR_INVALID_DATA_TYPE;
	    break;
	}
    }

    return rc;
}

sct_exp_error_t
eval_op_bit_and(struct tracker_filter_node* f_op,
		struct tracker_filter_node* val1,
		struct tracker_filter_node* val2,
		struct tracker_filter_node* result)
{
    sct_exp_error_t rc = ERR_NONE;
    sct_data_type_t t1 = val1->data_type;
    sct_data_type_t t2 = val2->data_type;

    /* we don't support bit operations for some data types. */
    if (t1 == DT_STRING || t1 == DT_BOOL || t1 == DT_FLOAT || t1 == DT_BUFFER) {
	rc = ERR_TYPE_MISMATCH;
    }
    if (t2 == DT_STRING || t2 == DT_BOOL || t2 == DT_FLOAT || t2 == DT_BUFFER) {
	rc = ERR_TYPE_MISMATCH;
    }
    else {
	/* make sure the larger data type is in 'val1'. */
	int t2_is_larger = (typecmp(t1, t2) < 0);
	if (t2_is_larger) 
	    swap_filter_node(&val1, &val2, &t1, &t2);

	result->type = NT_CONST;
	result->data_type = t1;
	RESULT_DATA_TYPE(result) = t1;

	switch (t1) {
	case DT_LONG:
	    RESULT_LONG(result) =
		CONST_LONG(val1) & cast_long(val2);
	    break;
	case DT_ULONG:
	    RESULT_ULONG(result) =
		CONST_ULONG(val1) & cast_ulong(val2);
	    break;
	case DT_INT:
	    RESULT_INT(result) =
		CONST_INT(val1) & cast_int(val2);
	    break;
	case DT_UINT:
	    RESULT_UINT(result) =
		CONST_UINT(val1) & cast_uint(val2);
	    break;
	case DT_SHORT:
	    RESULT_SHORT(result) =
		CONST_SHORT(val1) & cast_short(val2);
	    break;
	case DT_USHORT:
	    RESULT_USHORT(result) =
		CONST_USHORT(val1) & cast_ushort(val2);
	    break;
	case DT_CHAR:
	    RESULT_CHAR(result) =
		CONST_CHAR(val1) & cast_char(val2);
	    break;
	default:
	    /* shouldn't happen. */
	    rc = ERR_TYPE_MISMATCH;
	    break;
	}
    }

    return rc;
}

sct_exp_error_t
eval_op_bit_or(struct tracker_filter_node* f_op,
	       struct tracker_filter_node* val1,
	       struct tracker_filter_node* val2,
	       struct tracker_filter_node* result)
{
    sct_exp_error_t rc = ERR_NONE;
    sct_data_type_t t1 = val1->data_type;
    sct_data_type_t t2 = val2->data_type;

    /* we don't support bit operations for some data types. */
    if (t1 == DT_STRING || t1 == DT_BOOL || t1 == DT_FLOAT || t1 == DT_BUFFER) {
	rc = ERR_TYPE_MISMATCH;
    }
    if (t2 == DT_STRING || t2 == DT_BOOL || t2 == DT_FLOAT || t1 == DT_BUFFER) {
	rc = ERR_TYPE_MISMATCH;
    }
    else {
	/* make sure the larger data type is in 'val1'. */
	int t2_is_larger = (typecmp(t1, t2) < 0);
	if (t2_is_larger) 
	    swap_filter_node(&val1, &val2, &t1, &t2);

	result->type = NT_CONST;
	result->data_type = t1;
	RESULT_DATA_TYPE(result) = t1;

	switch (t1) {
	case DT_LONG:
	    RESULT_LONG(result) =
		CONST_LONG(val1) | cast_long(val2);
	    break;
	case DT_ULONG:
	    RESULT_ULONG(result) =
		CONST_ULONG(val1) | cast_ulong(val2);
	    break;
	case DT_INT:
	    RESULT_INT(result) =
		CONST_INT(val1) | cast_int(val2);
	    break;
	case DT_UINT:
	    RESULT_UINT(result) =
		CONST_UINT(val1) | cast_uint(val2);
	    break;
	case DT_SHORT:
	    RESULT_SHORT(result) =
		CONST_SHORT(val1) | cast_short(val2);
	    break;
	case DT_USHORT:
	    RESULT_USHORT(result) =
		CONST_USHORT(val1) | cast_ushort(val2);
	    break;
	case DT_CHAR:
	    RESULT_CHAR(result) =
		CONST_CHAR(val1) | cast_char(val2);
	    break;
	default:
	    /* shouldn't happen. */
	    rc = ERR_TYPE_MISMATCH;
	    break;
	}
    }

    return rc;
}

sct_exp_error_t
eval_op_bit_xor(struct tracker_filter_node* f_op,
		struct tracker_filter_node* val1,
		struct tracker_filter_node* val2,
		struct tracker_filter_node* result)
{
    sct_exp_error_t rc = ERR_NONE;
    sct_data_type_t t1 = val1->data_type;
    sct_data_type_t t2 = val2->data_type;

    /* we don't support bit operations for some data types. */
    if (t1 == DT_STRING || t1 == DT_BOOL || t1 == DT_FLOAT || t1 == DT_BUFFER) {
	rc = ERR_TYPE_MISMATCH;
    }
    if (t2 == DT_STRING || t2 == DT_BOOL || t2 == DT_FLOAT || t1 == DT_BUFFER) {
	rc = ERR_TYPE_MISMATCH;
    }
    else {
	/* make sure the larger data type is in 'val1'. */
	int t2_is_larger = (typecmp(t1, t2) < 0);
	if (t2_is_larger) 
	    swap_filter_node(&val1, &val2, &t1, &t2);

	result->type = NT_CONST;
	result->data_type = t1;
	RESULT_DATA_TYPE(result) = t1;

	switch (t1) {
	case DT_LONG:
	    RESULT_LONG(result) =
		CONST_LONG(val1) ^ cast_long(val2);
	    break;
	case DT_ULONG:
	    RESULT_ULONG(result) =
		CONST_ULONG(val1) ^ cast_ulong(val2);
	    break;
	case DT_INT:
	    RESULT_INT(result) =
		CONST_INT(val1) ^ cast_int(val2);
	    break;
	case DT_UINT:
	    RESULT_UINT(result) =
		CONST_UINT(val1) ^ cast_uint(val2);
	    break;
	case DT_SHORT:
	    RESULT_SHORT(result) =
		CONST_SHORT(val1) ^ cast_short(val2);
	    break;
	case DT_USHORT:
	    RESULT_USHORT(result) =
		CONST_USHORT(val1) ^ cast_ushort(val2);
	    break;
	case DT_CHAR:
	    RESULT_CHAR(result) =
		CONST_CHAR(val1) ^ cast_char(val2);
	    break;
	default:
	    /* shouldn't happen. */
	    rc = ERR_TYPE_MISMATCH;
	    break;
	}
    }

    return rc;
}

sct_exp_error_t
eval_op_bit_not(struct tracker_filter_node* f_op,
		struct tracker_filter_node* val,
		struct tracker_filter_node* result)
{
    sct_exp_error_t rc = ERR_NONE;
    sct_data_type_t t = val->data_type;

    /* we don't support shift operations for some data types. */
    if (t == DT_STRING || t == DT_BOOL || t == DT_FLOAT || t == DT_BUFFER) {
	rc = ERR_TYPE_MISMATCH;
    }
    else {
	result->type = NT_CONST;
	result->data_type = val->data_type;
	RESULT_DATA_TYPE(result) = val->data_type;

	switch (result->data_type) {
	case DT_CHAR:
	    RESULT_CHAR(result) = ~CONST_CHAR(val);
	    break;
	case DT_INT:
	    RESULT_INT(result) = ~CONST_INT(val);
	    break;
	case DT_UINT:
	    RESULT_UINT(result) = ~CONST_UINT(val);
	    break;
	case DT_SHORT:
	    RESULT_SHORT(result) = ~CONST_SHORT(val);
	    break;
	case DT_USHORT:
	    RESULT_USHORT(result) = ~CONST_USHORT(val);
	    break;
	case DT_LONG:
	    RESULT_LONG(result) = ~CONST_LONG(val);
	    break;
	case DT_ULONG:
	    RESULT_ULONG(result) = ~CONST_ULONG(val);
	    break;
	default:
	    /* shouldn't happen. */
	    rc = ERR_INVALID_DATA_TYPE;
	    break;
	}
    }

    return rc;
}

sct_exp_error_t
eval_op_logic_and(struct tracker_filter_node* f_op,
		  struct tracker_filter_node* val1,
		  struct tracker_filter_node* val2,
		  struct tracker_filter_node* result)
{
    int val1_bool = cast_bool(val1);
    int val2_bool = cast_bool(val2);

    result->type = NT_CONST;
    RESULT_DATA_TYPE(result) = DT_BOOL;
    RESULT_BOOL(result) = (val1_bool && val2_bool ? TF_TRUE : TF_FALSE);

    return ERR_NONE;
}

sct_exp_error_t
eval_op_logic_or(struct tracker_filter_node* f_op,
		 struct tracker_filter_node* val1,
		 struct tracker_filter_node* val2,
		 struct tracker_filter_node* result)
{
    int val1_bool = cast_bool(val1);
    int val2_bool = cast_bool(val2);

    result->type = NT_CONST;
    RESULT_DATA_TYPE(result) = DT_BOOL;
    RESULT_BOOL(result) = (val1_bool || val2_bool ? TF_TRUE : TF_FALSE);

    return ERR_NONE;
}

sct_exp_error_t
eval_logic_not(struct tracker_filter_node* f_op,
	       struct tracker_filter_node* val,
	       struct tracker_filter_node* result)
{
    int val_bool = cast_bool(val);

    result->type = NT_CONST;
    RESULT_DATA_TYPE(result) = DT_BOOL;
    RESULT_BOOL(result) = (val_bool ? TF_FALSE : TF_TRUE);

    return ERR_NONE;
}

sct_exp_error_t
eval_pattern_match(struct tracker_filter_node* f_op,
	           struct tracker_filter_node* val1,
	           struct tracker_filter_node* val2,
	           struct tracker_filter_node* result)
{
    char* str = NULL, *spat = NULL; 
    void* buf = NULL, *bpat = NULL; 
    unsigned int bufsz, bpatsz; 
    int res = 0; 

    result->type = NT_CONST;
    RESULT_DATA_TYPE(result) = DT_BOOL;

    if (CONST_DATA_TYPE(val1) == DT_BUFFER) {
	    cast_buffer(val2, &bpat, &bpatsz); 
	    if (!bpat)
		    return ERR_INVALID_VAR; 
	    cast_buffer(val1, &buf, &bufsz); 
	    if (!buf) {
		    mem_free(bpat); 
		    return ERR_INVALID_VAR; 
	    }
#if 0	 
	    /* if val2 (the pattern) is a string (not a buffer), we might want to */ 
	    /* cast it into a buffer *without* its terminating NULL string, to avoid  */ 
	    /* matching for the NULL ptr as well in buf */ 
	    if (CONST_DATA_TYPE(val2) != DT_BUFFER) {
		    if (((unsigned char*)bpat)[bpatsz-1] == '\0') {
			    PPRINT("truncating NULL from string pattern matching "
				   "to a buffer"); 
			    --bpatsz; 
		    }
	    }
#endif /* 0 */ 
	    res = buf_pattern_match(buf, bufsz, bpat, bpatsz); 
	    mem_free(buf);
	    mem_free(bpat);
    } else { /* treat as strings */ 
	    str = cast_string(val1);
	    if (!str)
		    return ERR_INVALID_VAR; 
	    spat = cast_string(val2);
	    if (!spat) {
		    mem_free(str); 
		    return ERR_INVALID_VAR; 
	    }
	    res = str_pattern_match(str, spat); 
	    mem_free(str); 
	    mem_free(spat); 
    }
    
    RESULT_BOOL(result) = res; 
    return ERR_NONE;
}

sct_exp_error_t
eval_in_match(struct tracker_filter_node* f_op,
	       struct tracker_filter_node* val1,
	       struct tracker_filter_node* val2,
	       struct tracker_filter_node* result)
{
    int index;
    sct_filter_node_t* f1;
    sct_exp_error_t rc = ERR_INVALID_NODE;
    sct_exp_error_t rc1 = ERR_NONE;

    if (val2->type != NT_VECTOR) { /* can happen ? */
	rc = ERR_INVALID_DATA_TYPE;
    } else
	for (index = 0; index < val2->u.vec.len; index++) {
	    f1 = val2->u.vec.nodes[index];
	    if (f1 == NULL)
		continue;
	    rc1 = eval_op_equal(f_op, val1, f1, result);
	    if (rc1 == ERR_NONE && RESULT_BOOL(result)) {
	        rc = rc1;
	        break;
	    } else {
	        rc = rc1;
	    }
	}
    return rc;
}

/* calculate the operator found in the given node */
/* on the two results.				  */
sct_exp_error_t
sct_eval_filter_node_op(struct tracker_filter_node* f_op,
		    struct tracker_filter_node* val1,
		    struct tracker_filter_node* val2,
		    struct tracker_filter_node* result)
{
    sct_exp_error_t rc = ERR_NONE;

    switch(f_op->u.compound.op) {
    case OT_EQUAL:
	rc = eval_op_equal(f_op, val1, val2, result);
	break;
    case OT_LESS:
	rc = eval_op_less(f_op, val1, val2, result);
	break;
    case OT_GREATER:
	rc = eval_op_greater(f_op, val1, val2, result);
	break;
    case OT_LESS_EQUAL:
	rc = eval_op_less_equal(f_op, val1, val2, result);
	break;
    case OT_GREATER_EQUAL:
	rc = eval_op_greater_equal(f_op, val1, val2, result);
	break;
    case OT_NOT_EQUAL:
	rc = eval_op_not_equal(f_op, val1, val2, result);
	break;
    case OT_PLUS:
	rc = eval_op_plus(f_op, val1, val2, result);
	break;
    case OT_MINUS:
	rc = eval_op_minus(f_op, val1, val2, result);
	break;
    case OT_BIT_SHIFT_LEFT:
	rc = eval_op_shift_left(f_op, val1, val2, result);
	break;
    case OT_BIT_SHIFT_RIGHT:
	rc = eval_op_shift_right(f_op, val1, val2, result);
	break;
    case OT_BIT_AND:
	rc = eval_op_bit_and(f_op, val1, val2, result);
	break;
    case OT_BIT_OR:
	rc = eval_op_bit_or(f_op, val1, val2, result);
	break;
    case OT_BIT_XOR:
	rc = eval_op_bit_xor(f_op, val1, val2, result);
	break;
    case OT_BIT_NOT:
	rc = eval_op_bit_not(f_op, val1, result);
	break;
    case OT_LOGIC_AND:
	rc = eval_op_logic_and(f_op, val1, val2, result);
	break;
    case OT_LOGIC_OR:
	rc = eval_op_logic_or(f_op, val1, val2, result);
	break;
    case OT_LOGIC_NOT:
	rc = eval_logic_not(f_op, val1, result);
	break;
    case OT_PATTERN_MATCH:
	rc = eval_pattern_match(f_op, val1, val2, result);
	break;
    case OT_IN_MATCH:
	rc = eval_in_match(f_op, val1, val2, result);
	break;
    default:
	/* shouldn't happen. */
	rc = ERR_INVALID_OPERATOR;
	break;
    }

    return rc;
}

/* evaluate the given compound node recursively. */
/* first recursively calculate each sub-node,    */
/* and then calculate its operator.              */
/* TODO - add short-circuiting support.          */
sct_exp_error_t
sct_eval_filter_node_compound(struct tracker_filter_node* f,
			  struct tracker_filter_node* result)
{
    struct tracker_filter_node* result1 = sct_make_bool_val(TF_FALSE);
    struct tracker_filter_node* result2 = sct_make_bool_val(TF_FALSE);
    struct tracker_filter_node* f1;
    struct tracker_filter_node* f2;
    sct_exp_error_t rc1 = ERR_NONE;
    sct_exp_error_t rc2 = ERR_NONE;
    sct_exp_error_t rc = ERR_NONE;
    oper_type_t op;

    f1 = f->u.compound.node1;
    if (!f1) {
	rc = ERR_INVALID_NODE; /* a compound node *must* have at least one leaf */
	goto eval_done;
    }

    /* evaluate that leaf */
    rc1 = sct_eval_filter_node(f1, result1);
    if (rc1 != ERR_NONE) {
	rc = rc1;
	goto eval_done;
    }

    /* the second sub-node is used only for binary operators. if we are 
       evaluating a unary operator, we dont care if f2 is null. */
    op = f->u.compound.op; 
    f2 = f->u.compound.node2;
    if (f2 == NULL && op != OT_LOGIC_NOT && op != OT_BIT_NOT) {
	rc = ERR_INVALID_NODE;
	goto eval_done;
    }

    /* are we evaluating a short circuiting operator? (logical or or logical and)) */
    if (op == OT_LOGIC_OR || op == OT_LOGIC_AND) {
	/* we might be able to short citcuit - 
	   for or, if the lhs is true, we return true for the whole expression
	   for and, if the lhs if false, we return false for the whole expression  */
	result->type = NT_CONST;
	RESULT_DATA_TYPE(result) = DT_BOOL;
	if (op == OT_LOGIC_OR &&  result1->type == NT_CONST && 
	    cast_bool(result1) == TF_TRUE){
	    RESULT_BOOL(result) = TF_TRUE;
	    rc  = rc1;
	    goto eval_done;
	} else if (op == OT_LOGIC_AND &&  result1->type == NT_CONST && 
		   cast_bool(result1) == TF_FALSE) {
	    RESULT_BOOL(result) = TF_FALSE;
	    rc  = rc1;
	    goto eval_done;
	}
    }
	
    /* evaluate the second node, if we have a second node */
    if (f2) {
	rc2 = sct_eval_filter_node(f2, result2);
	if (rc2 != ERR_NONE) {
	    rc = rc2;
	    goto eval_done;
	}
    }

    /* no error so far, try to evaluate the op */
    rc = sct_eval_filter_node_op(f, result1, result2, result);

 eval_done:
    sct_free_filter_node(result1);
    sct_free_filter_node(result2);
    return rc;
}

sct_exp_error_t
sct_eval_filter_node_vector(struct tracker_filter_node* f,
			  struct tracker_filter_node* result)
{
    struct tracker_filter_node* result1 = NULL;
    int len = f->u.vec.len;
    sct_exp_error_t rc = ERR_NONE;
    result->type = NT_VECTOR;
    result->data_type = DT_MULTI;
    result->u.vec.len = len;
    result->u.vec.nodes = mem_alloc(sizeof(sct_filter_node_t*) * len);
    if (!result->u.vec.nodes) {
	rc = ERR_INVALID_NODE;
    } else {
        int i;
        for (i=0; i < len; i++) {
	    if (f->u.vec.nodes[i] == NULL) {
		result->u.vec.nodes[i] = NULL;
	    } else {
		result1 = sct_make_bool_val(TF_FALSE);
		rc = sct_eval_filter_node(f->u.vec.nodes[i], result1);
		if (rc != ERR_NONE) {
		    sct_free_filter_node(result1);
		    break;
		} else {
		    result->u.vec.nodes[i] = result1;
		}
	    }
	}
    }
    return rc;
}

/* evaluate the given filter node. set the result */
/* in the given result node.			  */
sct_exp_error_t
do_sct_eval_filter_node(struct tracker_filter_node* f,
		    struct tracker_filter_node* result)
{
    assert(f && result);

    switch (f->type) {
    case NT_VAR:
	return sct_eval_filter_node_var(f, result);
    case NT_CONST:
	return sct_eval_filter_node_const(f, result);
    case NT_COMPOUND:
	return sct_eval_filter_node_compound(f, result);
    case NT_VECTOR:
	return sct_eval_filter_node_vector(f, result);
    default:
	/* shouldn't happen. */
	return ERR_INVALID_NODE;
    }
}

/* evaluate the given filter node. */
/* since this is a generic function, it returns a */
/* filter node, containing a constant and the     */
/* proper result.                                 */
sct_exp_error_t
sct_eval_filter_node(sct_filter_node_t* f, sct_filter_node_t* result)
{
    /* assume failure. */
    sct_exp_error_t rc = ERR_INVALID_NODE;

    if (f && result) {
	/* free result contents, to make sure we dont override any memory */
	sct_free_filter_node_contents(result);
	
	rc = do_sct_eval_filter_node(f, result);
    }

    return rc;
}
