/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 guy keren, choo@actcom.co.il	            */
/* License: GNU Lesser General Public License                       */

/*
 *$Id: filter_in_out.c,v 1.15 2002/06/15 15:16:13 mulix Exp $
 */
/*
 * Rule engine - filter input/output and serialization routines.
 */

#include "mem_manager_public.h"
#include "encode_decode_public.h"
#include "rule_engine_private.h"


/* operators symbols, used for printing. */
const char* operators_symbols[] = {
  "==", /* OT_EQUAL */
  "<",  /* OT_LESS */
  ">",  /* OT_GREATER */
  "<=", /* OT_LESS_EQUAL */
  ">=", /* OT_GREATER_EQUAL */
  "!=", /* OT_NOT_EQUAL */
  "+",  /* OT_PLUS */
  "-",  /* OT_MINUS */
  "<<", /* OT_BIT_SHIFT_LEFT */
  ">>", /* OT_BIT_SHIFT_RIGHT */
  "&",  /* OT_BIT_AND */
  "|",  /* OT_BIT_OR */
  "^",  /* OT_BIT_XOR */
  "~",  /* OT_BIT_NOT */
  "&&", /* OT_LOGIC_AND */
  "||", /* OT_LOGIC_OR */
  "!",  /* OT_LOGIC_NOT */
  "~=", /* OT_PATTERN_MATCH */
  "in", /* OT_IN_MATCH */
};

/*
 * Private functions.
 */

/* print spaces in the ammount given. */
void
print_indent(unsigned int len)
{
    char buf[200];
    unsigned int bufsiz = sizeof(buf) - 1;

    len = (len > bufsiz) ? bufsiz : len;

    memset(buf, ' ', len);
    buf[len] = '\0';

    PPRINT(buf);
}

/* get a var as string */ 
const char* sct_get_var_type_as_str(sct_var_type_t var)
{
    switch (var) {
    case VT_UNDEF: return "undef"; 
    case VT_SYSCALL_ID: return "syscall id";
    case VT_PROC_UID: return "uid"; 
    case VT_PROC_EUID: return "uid"; 
    case VT_PROC_SUID: return "suid"; 
    case VT_PROC_GID: return "gid"; 
    case VT_PROC_EGID: return "egid"; 
    case VT_PROC_SGID: return "sgid"; 
    case VT_PROC_PID: return "pid"; 
    case VT_PROC_COMM: return "command"; 
    case VT_RETVAL: return "return value"; 
    case VT_PROC_PPID: return "parent pid"; 
    case VT_PROC_PCOMM: return "parent command"; 
    case VT_PROC_ANCESTORS: return "parent tree"; 
    case VT_PARAM: return "syscall parameter"; 
    case VT_LAST_TYPE: return "last type"; 
    default: return "unknown"; 
    }	
}

/* print a var node. */
void
sct_print_filter_node_var(struct tracker_filter_node* f, unsigned int indent)
{
    print_indent(indent);
    PPRINT("var '%s'", sct_get_var_type_as_str(f->u.var.type));
    PPRINT(" index '%d'", f->u.var.var_index + 1);
    PPRINT(" field '%d'", f->u.var.field_index);
    PPRINT(" struct type '%d'", f->u.var.struct_type);
    PPRINT("\n");
}

/* print a const node. */
void
sct_print_filter_node_const(struct tracker_filter_node* f, unsigned int indent)
{
    print_indent(indent);
    switch (CONST_DATA_TYPE(f)) {
    case DT_BOOL:
	PPRINT("bool: %d\n", CONST_BOOL(f));
	break;
    case DT_CHAR:
	PPRINT("char: %c (%d)\n",
	       CONST_CHAR(f), (int)CONST_CHAR(f));
	break;
    case DT_INT:
	PPRINT("int: %d\n", CONST_INT(f));
	break;
    case DT_UINT:
	PPRINT("unsigned int: %u\n", CONST_UINT(f));
	break;
    case DT_SHORT:
	PPRINT("short: %hd\n", CONST_SHORT(f));
	break;
    case DT_USHORT:
	PPRINT("unsigned short: %hu\n", CONST_USHORT(f));
	break;
    case DT_LONG:
	PPRINT("long: %ld\n", CONST_LONG(f));
	break;
    case DT_ULONG:
	PPRINT("unsigned long: %lu\n", CONST_ULONG(f));
	break;
    case DT_FLOAT:
	PPRINT("float: %f\n", CONST_FLOAT(f));
	break;
    case DT_STRING:
    {
	const char* str = CONST_STRING(f);
	PPRINT("str: %s\n", (str ? str : "(nil)"));
	break; 
    }
    case DT_BUFFER:
    {
	char* out = sct_buffer_to_string(CONST_BUFFER(f), BUFFER_SIZE(f)); 
	PPRINT("buffer [size %d]: %s\n", BUFFER_SIZE(f), 
	       (out ? out : "(nil)"));
	if (out)
		mem_free(out); 
	break; 
    }
    default:
	/* shouldn't happen. */
	PPRINT("(unknown type %d)\n", CONST_DATA_TYPE(f));
	break;
    }
}

/* print a compound node. */
void
sct_print_filter_node_compound(struct tracker_filter_node* f, unsigned int indent)
{
    const char* op = "(error)";

    if (f->u.compound.op >= 0 && f->u.compound.op < OT_LAST_OPERATOR)
	op = operators_symbols[f->u.compound.op];
    print_indent(indent);
    PPRINT("operator: '%s' (%d)\n", op, f->u.compound.op);
    print_indent(indent);
    PPRINT("f1:\n");
    if (f->u.compound.node1) {
	do_sct_print_filter_node(f->u.compound.node1, indent+INDENT_LEVEL);
    }
    else {
	print_indent(indent);
	PPRINT("Empty\n");
    }
    print_indent(indent);
    PPRINT("f2:\n");
    if (f->u.compound.node2) {
	do_sct_print_filter_node(f->u.compound.node2, indent+INDENT_LEVEL);
    }
    else {
	print_indent(indent);
	PPRINT("Empty\n");
    }
}

/* recursively print the given filter node, beginning */
/* with the given indentation level.                  */
void
do_sct_print_filter_node(struct tracker_filter_node* f, unsigned int indent)
{
    switch(f->type) {
    case NT_VAR:
	sct_print_filter_node_var(f, indent);
	break;
    case NT_CONST:
	sct_print_filter_node_const(f, indent);
	break;
    case NT_COMPOUND:
	sct_print_filter_node_compound(f, indent);
	break;
    case NT_VECTOR:
	sct_print_filter_node_vec(f, indent);
	break;
    default:
	/* shouldn't happen. */
	print_indent(indent);
	PPRINT("Error: invalid node type '%d'\n", f->type);
	break;
    }
}


/* print the given filter node (and all its sub-nodes) to output. */
/* output is stdout if in user-mode, or using printk if in kernel */
/* mode. the output is being formatted to reflext the hierarchy   */
/* of the expression stored in the tree - good for debugging.     */
void
sct_print_filter_node(sct_filter_node_t* f)
{
    unsigned int indent = 3;
    if (f) {
	PPRINT("  Filter node data:\n");
	do_sct_print_filter_node(f, indent);
	PPRINT("\n");
    }
    else {
	PPRINT("Empty filter node\n");
    }
}


/*
 * Filter serialization/de-serialization functions.
 */


/*
 * Private functions.
 */

/* serialize a 'var' filter node. */
int
sct_serialize_filter_node_var(sct_filter_node_t* f,
			  unsigned char* buf,
			  int buflen,
			  unsigned char** pos)
{
    int rc = encode_char(buf, buflen, pos, NT_VAR);
    if (rc > 0) {
	rc = encode_char(buf, buflen, pos, f->data_type);
    }
    if (rc > 0) {
	rc = encode_char(buf, buflen, pos, f->u.var.type);
    }
    if (rc > 0) {
	rc = encode_int(buf, buflen, pos, f->u.var.struct_type);
    }
    if (rc > 0) {
	rc = encode_int(buf, buflen, pos, f->u.var.var_index);
    }
    if (rc > 0) {
	rc = encode_int(buf, buflen, pos, f->u.var.field_index);
    }

    return rc;
}

/* de-serialize a 'var' filter node. */
int
sct_deserialize_filter_node_var(sct_filter_node_t** pf,
			    unsigned char* buf,
			    int buflen,
			    unsigned char** pos)
{
    char var_data_type;
    char var_type;
    int struct_type;
    int var_index;
    int var_field;
    int rc = decode_char(buf, buflen, pos, &var_data_type);

    if (rc > 0) {
	rc = decode_char(buf, buflen, pos, &var_type);
    }
    if (rc > 0) {
	rc = decode_int(buf, buflen, pos, &struct_type);
    }
    if (rc > 0) {
	rc = decode_int(buf, buflen, pos, &var_index);
    }
    if (rc > 0) {
	rc = decode_int(buf, buflen, pos, &var_field);
    }
    if (rc > 0) {
	*pf = sct_make_var_indexed_field_val(var_type,
					 var_index,
					 var_field,
					 var_data_type,
					 struct_type);
	if (!*pf) {
	    rc = 0;
	}
    }

    return rc;
}

/* serialize a 'const' filter node. */
int
sct_serialize_filter_node_const(sct_filter_node_t* f,
			    unsigned char* buf,
			    int buflen,
			    unsigned char** pos)
{
    int rc = encode_char(buf, buflen, pos, NT_CONST);
    if (rc > 0) {
	rc = encode_char(buf, buflen, pos, f->data_type);
    }
    if (rc > 0) {
	rc = encode_char(buf, buflen, pos, f->u.constant.type);
    }
    if (rc > 0) {
	switch (f->u.constant.type) {
	case DT_BOOL:
	    rc = encode_char(buf,
			     buflen, pos, CONST_BOOL(f));
	    break;
	case DT_CHAR:
	    rc = encode_char(buf,
			     buflen, pos, CONST_CHAR(f));
	    break;
	case DT_SHORT:
	    rc = encode_short(buf,
			      buflen, pos, CONST_SHORT(f));
	    break;
	case DT_USHORT:
	    rc = encode_ushort(buf,
			       buflen, pos, CONST_USHORT(f));
	    break;
	case DT_INT:
	    rc = encode_int(buf,
			    buflen, pos, CONST_INT(f));
	    break;
	case DT_UINT:
	    rc = encode_uint(buf,
			     buflen, pos, CONST_UINT(f));
	    break;
	case DT_LONG:
	    rc = encode_long(buf,
			     buflen, pos, CONST_LONG(f));
	    break;
	case DT_ULONG:
	    rc = encode_ulong(buf,
			      buflen, pos, CONST_ULONG(f));
	    break;
	case DT_FLOAT:
	    rc = encode_float(buf,
			      buflen, pos, CONST_FLOAT(f));
	    break;
	case DT_STRING:
	    rc = encode_string(buf,
			       buflen, pos, CONST_STRING(f));
	    break; 
	case DT_BUFFER:
	    rc = encode_buffer(buf, buflen, pos, CONST_BUFFER(f), 
			       BUFFER_SIZE(f)); 
	    break;
	default:
	    /* shouldn't happen. */
	    rc = 0;
	    break;
	}
    }

    return rc;
}

/* de-serialize a 'const' filter node. */
int
sct_deserialize_filter_node_const(sct_filter_node_t** pf,
				  unsigned char* buf,
				  int buflen,
				  unsigned char** pos)
{
    char data_type;
    char constant_type;
    int rc = decode_char(buf, buflen, pos, &data_type);

    if (rc > 0) {
	rc = decode_char(buf, buflen, pos, &constant_type);
    }
    if (rc > 0) {
	switch (constant_type) {
	case DT_BOOL:
	{
	    char c;
	    rc = decode_char(buf,
			     buflen, pos, &c);
	    if (rc > 0) {
		*pf = sct_make_bool_val(c);
		if (!*pf) {
		    rc = 0;
		}
	    }
	}
	break;
	case DT_CHAR:
	{
	    char c;
	    rc = decode_char(buf,
			     buflen, pos, &c);
	    if (rc > 0) {
		*pf = sct_make_char_val(c);
		if (!*pf) {
		    rc = 0;
		}
	    }
	}
	break;
	case DT_SHORT:
	{
	    short s;
	    rc = decode_short(buf,
			      buflen, pos, &s);
	    if (rc > 0) {
		*pf = sct_make_short_val(s);
		if (!*pf) {
		    rc = 0;
		}
	    }
	}
	break;
	case DT_USHORT:
	{
	    unsigned short us;
	    rc = decode_ushort(buf,
			       buflen, pos, &us);
	    if (rc > 0) {
		*pf = sct_make_ushort_val(us);
		if (!*pf) {
		    rc = 0;
		}
	    }
	}
	break;
	case DT_INT:
	{
	    int i;
	    rc = decode_int(buf,
			    buflen, pos, &i);
	    if (rc > 0) {
		*pf = sct_make_int_val(i);
		if (!*pf) {
		    rc = 0;
		}
	    }
	}
	break;
	case DT_UINT:
	{
	    unsigned int ui;
	    rc = decode_uint(buf,
			     buflen, pos, &ui);
	    if (rc > 0) {
		*pf = sct_make_uint_val(ui);
		if (!*pf) {
		    rc = 0;
		}
	    }
	}
	break;
	case DT_LONG:
	{
	    long l;
	    rc = decode_long(buf,
			     buflen, pos, &l);
	    if (rc > 0) {
		*pf = sct_make_long_val(l);
		if (!*pf) {
		    rc = 0;
		}
	    }
	}
	break;
	case DT_ULONG:
	{
	    unsigned long ul;
	    rc = decode_ulong(buf,
			      buflen, pos, &ul);
	    if (rc > 0) {
		*pf = sct_make_ulong_val(ul);
		if (!*pf) {
		    rc = 0;
		}
	    }
	}
	break;
	case DT_FLOAT:
	{
	    float f;
	    rc = decode_float(buf,
			      buflen, pos, &f);
	    if (rc > 0) {
		*pf = sct_make_float_val(f);
		if (!*pf) {
		    rc = 0;
		}
	    }
	}
	break;
	case DT_STRING:
	{
	    char* str;
	    rc = decode_string(buf,
			       buflen, pos, &str);
	    if (rc > 0) {
		*pf = sct_make_string_val_without_copy(str);
		if (!*pf) {
		    rc = 0;
		    mem_free(str); 
		}
	    }
	}
	break;
	case DT_BUFFER:
	{
	    void* in; 
	    unsigned int size; 
	    rc = decode_buffer(buf, buflen, pos, &in, &size); 
	    if (rc > 0) {
		*pf = sct_make_buffer_val_without_copy(in, size); 
		if (!*pf) {
		    rc = 0; 
		    mem_free(in); 
		}
	    }
	}
	default:
	    /* shouldn't happen. */
	    rc = 0;
	    break;
	}
    }

    return rc;
}

/* serialize a 'compound' filter node, and its sub-nodes. */
int
sct_serialize_filter_node_compound(sct_filter_node_t* f,
			       unsigned char* buf,
			       int buflen,
			       unsigned char** pos)
{
    int rc = encode_char(buf, buflen, pos, NT_COMPOUND);
    if (rc > 0) {
	rc = encode_char(buf, buflen, pos, f->data_type);
    }
    if (rc > 0) {
	rc = encode_char(buf, buflen, pos, f->u.compound.op);
    }
    if (rc > 0) {
	rc = do_sct_serialize_filter_node(f->u.compound.node1,
				      buf, buflen, pos);
    }
    if (rc > 0) {
	rc = do_sct_serialize_filter_node(f->u.compound.node2,
				      buf, buflen, pos);
    }

    return rc;
}

/* de-serialize a 'compound' filter node, and its sub-nodes. */
int
sct_deserialize_filter_node_compound(sct_filter_node_t** pf,
			         unsigned char* buf,
			         int buflen,
			         unsigned char** pos)
{
    char data_type;
    char op;
    sct_filter_node_t* f1 = NULL;
    sct_filter_node_t* f2 = NULL;
    int rc = decode_char(buf, buflen, pos, &data_type);

    if (rc > 0) {
	rc = decode_char(buf, buflen, pos, &op);
    }
    if (rc > 0) {
	rc = do_sct_deserialize_filter_node(&f1, buf, buflen, pos);
    }
    if (rc > 0) {
	rc = do_sct_deserialize_filter_node(&f2, buf, buflen, pos);
    }
    if (rc > 0) {
	*pf = sct_make_compound_node_with_type(op, f1, f2, data_type);
	if (!*pf) {
	    rc = 0;
	}
    }

    /* free the sub-nodes, if we failed somewhere along the path. */
    if (rc < 1) {
	if (f1)
	    sct_free_filter_node(f1);
	if (f2)
	    sct_free_filter_node(f2);
    }

    return rc;
}

/* serialize a filter node (and all its sub-nodes) into a */
/* buffer, for export purposes. returns the size of the   */
/* serialized data, 0 if there was an error, -1 if buffer */
/* is too small. serialization begins at 'pos' in 'buf'.  */
int
do_sct_serialize_filter_node(sct_filter_node_t* f,
		         unsigned char* buf,
		         int buflen,
		         unsigned char** pos)
{
    int rc = 0;

    if (f) {
	switch (f->type) {
	case NT_VAR:
	    rc = sct_serialize_filter_node_var(f, buf, buflen, pos);
	    break;
	case NT_CONST:
	    rc = sct_serialize_filter_node_const(f, buf, buflen, pos);
	    break;
	case NT_COMPOUND:
	    rc = sct_serialize_filter_node_compound(f, buf, buflen, pos);
	    break;
	case NT_VECTOR:
	    rc = sct_serialize_filter_node_vector(f, buf, buflen, pos);
	    break;
	default:
	    /* shouldn't happen. */
	    rc = 0;
	    break;
	}
    }
    else {
	rc = encode_char(buf, buflen, pos, NT_NIL);
    }

    return rc;
}

/* de-serialize a filter node (and all its sub-nodes)     */
/* from a buffer, for import purposes. returns the new    */
/* allocated filter node (the top of the tree), or NULL   */
/* if an error occured. de-serialization begins as 'pos'  */
/* in 'buf'.                                              */
int
do_sct_deserialize_filter_node(sct_filter_node_t** pf,
			   unsigned char* buf,
			   int buflen,
			   unsigned char** pos)
{
    int rc = 0;
    char nt;

    rc = decode_char(buf, buflen, pos, &nt);
    if (rc > 0) {
	switch (nt) {
	case NT_VAR:
	    rc = sct_deserialize_filter_node_var(pf, buf, buflen, pos);
	    break;
	case NT_CONST:
	    rc = sct_deserialize_filter_node_const(pf, buf, buflen, pos);
	    break;
	case NT_COMPOUND:
	    rc = sct_deserialize_filter_node_compound(pf, buf, buflen, pos);
	    break;
	case NT_VECTOR:
	    rc = sct_deserialize_filter_node_vector(pf, buf, buflen, pos);
	    break;
	case NT_NIL:
	    /* a NULL filter_node. */
	    *pf = NULL;
	    rc = 1;
	    break;
	default:
	    /* shouldn't happen. */
	    rc = 0;
	    break;
	}
    }

    return rc;
}

/*
 * Public interface implementation.
 */

/* serialize a filter node (and all its sub-nodes) into a */
/* buffer, for export purposes. returns the size of the   */
/* serialized data, 0 if there was an error, -1 if buffer */
/* is too small.                                          */
int
sct_serialize_filter_node(sct_filter_node_t* f, unsigned char* buf, int buflen)
{
    unsigned char* pos = buf;
    int rc = do_sct_serialize_filter_node(f, buf, buflen, &pos);

    return (rc > 0 ? pos - buf : rc);
}

/* de-serialize a filter node (and all its sub-nodes)     */
/* from a buffer, for import purposes. returns the new    */
/* allocated filter node (the top of the tree), or NULL   */
/* if an error occured.                                   */
sct_filter_node_t*
sct_deserialize_filter_node(unsigned char* buf, int buflen)
{
    unsigned char* pos = buf;
    sct_filter_node_t* f = NULL;
    int rc = 0;

    if (buflen > 0) {
	rc = do_sct_deserialize_filter_node(&f, buf, buflen, &pos);
    }

    return (rc > 0 ? f : NULL);
}

