/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 guy keren, choo@actcom.co.il	            */
/* License: GNU Lesser General Public License                       */

/*
 *$Id: filter_node_build.c,v 1.15 2002/06/15 15:16:13 mulix Exp $
 */
/*
 * Rule engine - implementation of the filtering nodes building.
 */

#include "rule_engine_private.h"
#include "mem_manager_public.h"

#include "archdep.h"
/*
 * Private functions.
 */


/* build a compound node. */
struct tracker_filter_node*
sct_make_compound_node_with_type(oper_type_t op,
		   	     struct tracker_filter_node* node1,
		   	     struct tracker_filter_node* node2,
		   	     sct_data_type_t data_type)
{
    struct tracker_filter_node* node = NULL;
    if (op_allowed(op, node1, node2)) {
	node = (struct tracker_filter_node*)
	    mem_alloc(sizeof(struct tracker_filter_node));
	if (node) {
	    node->type = NT_COMPOUND;
	    node->data_type = data_type;
	    node->u.compound.op = op;
	    node->u.compound.node1 = node1;
	    node->u.compound.node2 = node2;
	}
    }

    return node;
}

/* free the 'var' part of a filter node. don't free the node itself. */
static void
free_var_of_filter_node(struct tracker_filter_node* f)
{
    /* nothing to do, since a var node contains only basic data types. */
}

/* free the 'const' part of a filter node. don't free the node itself. */
static void
free_const_of_filter_node(struct tracker_filter_node* f)
{
    /* free string, in case of a string type node. */
    if (CONST_DATA_TYPE(f) == DT_STRING) {
	if (CONST_STRING(f)) {
	    mem_free(CONST_STRING(f));
	    CONST_STRING(f) = NULL;
	}
    } else if (CONST_DATA_TYPE(f) == DT_BUFFER) {
        if (CONST_BUFFER(f)) {
	    mem_free(CONST_BUFFER(f)); 
	    CONST_BUFFER(f) = NULL; 
	    BUFFER_SIZE(f) = 0; 
	}
    }
}

/* free the 'compound' part of a filter node. don't free the node itself. */
static void
free_compound_of_filter_node(struct tracker_filter_node* f)
{
    if (f->u.compound.node1) {
	sct_free_filter_node(f->u.compound.node1);
	f->u.compound.node1 = NULL;
    }
    if (f->u.compound.node2) {
	sct_free_filter_node(f->u.compound.node2);
	f->u.compound.node2 = NULL;
    }
}


/*
 * Implementation of public filter building functions.
 */

/*
 * functions that are used to build a filtering expression.
 */

/* create primitive types - int, string, variable, indexed variable */
/* (e.g. parameter).                                                */

/* create a node with the given parameter and data type. we use a void*
   ugliness for polymorphic behaviour. return the filter node or NULL
   if we are OOM */
sct_filter_node_t*
do_sct_make_val(void* v, size_t size, sct_data_type_t type)
{
    struct tracker_filter_node* node =
	(struct tracker_filter_node*)
	mem_alloc(sizeof(struct tracker_filter_node));
    if (node) {
	/* better safe than sorry */
	memset(node, 0, sizeof(struct tracker_filter_node));
	node->type = NT_CONST;
	node->data_type = type;
	node->u.constant.type = type;
	
	/* make sure we are not overwriting memory */
	assert(size <= sizeof(node->u.constant.u));

	/* this *should* be valid, since all union members should occupy 
	   the same physical memory. should be verified in the c standard */ 
	memcpy(&node->u.constant.u, v, size);
    }
    
    return node;
}

sct_filter_node_t*
sct_make_bool_val(int num)
{
    return do_sct_make_val((void*)&num, sizeof(num), DT_BOOL);
}

sct_filter_node_t*
sct_make_int_val(int num)
{
    return do_sct_make_val((void*)&num, sizeof(num), DT_INT);
}

sct_filter_node_t*
sct_make_uint_val(unsigned int num)
{
    return do_sct_make_val((void*)&num, sizeof(num), DT_UINT);
}

sct_filter_node_t*
sct_make_char_val(char c)
{
    return do_sct_make_val((void*)&c, sizeof(c), DT_CHAR);
}

sct_filter_node_t*
sct_make_short_val(short num)
{
    return do_sct_make_val((void*)&num, sizeof(num), DT_SHORT);
}

sct_filter_node_t*
sct_make_ushort_val(unsigned short num)
{
    return do_sct_make_val((void*)&num, sizeof(num), DT_USHORT);
}

sct_filter_node_t*
sct_make_long_val(long num)
{
    return do_sct_make_val((void*)&num, sizeof(num), DT_LONG);
}

sct_filter_node_t*
sct_make_ulong_val(unsigned long num)
{
    return do_sct_make_val((void*)&num, sizeof(num), DT_ULONG);
}

sct_filter_node_t*
sct_make_float_val(float num)
{
    return do_sct_make_val((void*)&num, sizeof(num), DT_FLOAT);
}

sct_filter_node_t*
sct_make_string_val_without_copy(char* str)
{
    struct tracker_filter_node* node =
	(struct tracker_filter_node*)
	mem_alloc(sizeof(struct tracker_filter_node));
    if (node) {
	node->type = NT_CONST;
	node->data_type = DT_STRING;
	node->u.constant.type = DT_STRING;
	node->u.constant.u.str = str;
    }

    return node;
}

sct_filter_node_t*
sct_make_string_val(const char* str)
{
    struct tracker_filter_node* node =
	(struct tracker_filter_node*)
	mem_alloc(sizeof(struct tracker_filter_node));
    if (node) {
	node->type = NT_CONST;
	node->data_type = DT_STRING;
	node->u.constant.type = DT_STRING;
	if (str) {
	    node->u.constant.u.str =
		(char*)mem_alloc(strlen(str)+1);
	    if (node->u.constant.u.str)
		strcpy(node->u.constant.u.str, str);
	    else {
		mem_free(node);
		node = NULL;
	    }
	}
	else {
	    node->u.constant.u.str = NULL;
	}
    }
    return node;
}

sct_filter_node_t*
sct_make_buffer_val_without_copy(void* buf, unsigned int size)
{
    struct tracker_filter_node* node =
	(struct tracker_filter_node*)
	mem_alloc(sizeof(struct tracker_filter_node));
    if (node) {
	node->type = NT_CONST;
	node->data_type = DT_BUFFER; 
	node->u.constant.type = DT_BUFFER; 
	node->u.constant.u.buf.data = buf; 
	node->u.constant.u.buf.size = size; 
    }
    return node;
}

sct_filter_node_t*
sct_make_buffer_val(const void* buf, unsigned int size)
{
    struct tracker_filter_node* node =
	(struct tracker_filter_node*)
	mem_alloc(sizeof(struct tracker_filter_node));
    if (node) {
	node->type = NT_CONST;
	node->data_type = DT_BUFFER; 
	node->u.constant.type = DT_BUFFER; 
	if (buf) {
	    node->u.constant.u.buf.data = mem_alloc(size);
	    if (node->u.constant.u.buf.data) {
		memcpy(node->u.constant.u.buf.data, buf, size);
		node->u.constant.u.buf.size = size; 
	    } else {
		mem_free(node);
		node = NULL;
	    }
	} else {
	    node->u.constant.u.buf.data = NULL;
	    node->u.constant.u.buf.size = 0; 
	}
    }

    return node;
}


static sct_filter_node_t*
sct_make_generic_var(sct_var_type_t v, sct_data_type_t dt, 
		     int field_index, int var_index, int struct_type)
{
    struct tracker_filter_node* node =
	(struct tracker_filter_node*)
	mem_alloc(sizeof(struct tracker_filter_node));
    if (node) {
	node->type = NT_VAR;
	node->data_type = dt;
	node->u.var.type = v;
	node->u.var.struct_type = struct_type;
	node->u.var.var_index = var_index;
	node->u.var.field_index = field_index;
    }

    return node;
}
	     

sct_filter_node_t*
sct_make_var_val(sct_var_type_t v, sct_data_type_t var_data_type)
{
    return sct_make_generic_var(v, var_data_type, -1, 0, -1);
}

sct_filter_node_t*
sct_make_var_field_val(sct_var_type_t v, int field_index, 
		       sct_data_type_t var_data_type, int struct_type)
{
    return sct_make_generic_var(v, var_data_type, field_index, 
				0, struct_type);
}

sct_filter_node_t*
sct_make_var_indexed_val(sct_var_type_t v, int var_index, sct_data_type_t var_data_type)
{
    return sct_make_generic_var(v, var_data_type, -1, var_index, -1);
}

sct_filter_node_t*
sct_make_var_indexed_field_val(sct_var_type_t v,
			   int var_index,
			   int field_index,
			   sct_data_type_t var_data_type,
			   int struct_type)
{
    return sct_make_generic_var(v, var_data_type, field_index, 
				var_index, struct_type);
}

static sct_data_type_t
choose_larger_type(sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    sct_data_type_t d1 = f1 ? f1->data_type : DT_LAST_TYPE;
    sct_data_type_t d2 = f2 ? f2->data_type : DT_LAST_TYPE;
    int cmp;

    if (d1 == DT_LAST_TYPE)
	return d2;

    if (d2 == DT_LAST_TYPE)
	return d1;

    cmp = typecmp(d1, d2);
    return (cmp <= 0 ? d1 : d2);
}

static sct_filter_node_t* 
do_sct_make_op(sct_filter_node_t* f1, sct_filter_node_t* f2, oper_type_t op, 
	   sct_data_type_t res_dt)
{
    sct_filter_node_t* f = NULL;

    if (f1 && f2) {
	f = sct_make_compound_node_with_type(op, f1, f2, res_dt);
    } 

    /* if the making a compound node failed, 
       or f1 or f2 is NULL, free f1 and f2 */
    if (!f) {
	sct_free_filter_node(f1);
	sct_free_filter_node(f2);
    }
    
    return f;
}

static sct_filter_node_t* 
do_sct_make_unary_op(sct_filter_node_t* f1, oper_type_t op, sct_data_type_t dt)
{
    sct_filter_node_t* f = NULL;
    
    if (f1) {
	f = sct_make_compound_node_with_type(op, f1, NULL, dt);
    }

    if (!f)
	sct_free_filter_node(f1);

    return f;
    
}


/* combine expressions using operators.                          */
/* checks for matching (e.g. can't comprate and int and an str). */
sct_filter_node_t*
sct_make_equal(sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    return do_sct_make_op(f1, f2, OT_EQUAL, DT_BOOL);
}

sct_filter_node_t*
sct_make_less_then(sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    return do_sct_make_op(f1, f2, OT_LESS, DT_BOOL);
}

sct_filter_node_t*
sct_make_less_or_equal(sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    return do_sct_make_op(f1, f2, OT_LESS_EQUAL, DT_BOOL);
}

sct_filter_node_t*
sct_make_greater_then(sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    return do_sct_make_op(f1, f2, OT_GREATER, DT_BOOL);
}

sct_filter_node_t*
sct_make_greater_or_equal(sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    return do_sct_make_op(f1, f2, OT_GREATER_EQUAL, DT_BOOL);
}

sct_filter_node_t*
sct_make_not_equal(sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    return do_sct_make_op(f1, f2, OT_NOT_EQUAL, DT_BOOL);
}

sct_filter_node_t*
sct_make_logical_and(sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    return do_sct_make_op(f1, f2, OT_LOGIC_AND, DT_BOOL);
}

sct_filter_node_t*
sct_make_logical_or(sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    return do_sct_make_op(f1, f2, OT_LOGIC_OR, DT_BOOL);
}

sct_filter_node_t*
sct_make_logical_not(sct_filter_node_t* f1)
{
    return do_sct_make_unary_op(f1, OT_LOGIC_NOT, DT_BOOL);
}

sct_filter_node_t*
sct_make_bit_and(sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    return do_sct_make_op(f1, f2, OT_BIT_AND, 
		      choose_larger_type(f1, f2));
}

sct_filter_node_t*
sct_make_bit_or(sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    return do_sct_make_op(f1, f2, OT_BIT_OR,
		      choose_larger_type(f1, f2));
}

sct_filter_node_t*
sct_make_bit_xor(sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    return do_sct_make_op(f1, f2, OT_BIT_XOR,
		      choose_larger_type(f1, f2));
}

sct_filter_node_t*
sct_make_bit_not(sct_filter_node_t* f1)
{
    return do_sct_make_unary_op(f1, OT_BIT_NOT, 
			    f1 ? f1->data_type : DT_LAST_TYPE);
}

sct_filter_node_t*
sct_make_bit_left_shift(sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    return do_sct_make_op(f1, f2, OT_BIT_SHIFT_LEFT, 
		      f1? f1->data_type: DT_LAST_TYPE);
}

sct_filter_node_t*
sct_make_bit_right_shift(sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    return do_sct_make_op(f1, f2, OT_BIT_SHIFT_RIGHT, 
		      f1? f1->data_type: DT_LAST_TYPE);
}

sct_filter_node_t*
sct_make_math_plus(sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    return do_sct_make_op(f1, f2, OT_PLUS, 
		      choose_larger_type(f1, f2));
}

sct_filter_node_t*
sct_make_math_minus(sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    return do_sct_make_op(f1, f2, OT_MINUS, 
		      choose_larger_type(f1, f2));
}

sct_filter_node_t*
sct_make_pattern_match(sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    /* pattern match only makes sense for strings */
    return do_sct_make_op(f1, f2, OT_PATTERN_MATCH, DT_BOOL);
}

sct_filter_node_t*
sct_make_in_match(sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    return do_sct_make_op(f1, f2, OT_IN_MATCH, DT_BOOL);
}

/* free the given filter node contents, as well as all sub-nodes and */
/* heap memory it contains, directly and indirectly, 	 	     */
/* but leave the node memory alloced		 	             */
void
sct_free_filter_node_contents(sct_filter_node_t* f)
{
    if (f) {
	/* free contents, based on node type. */
	switch (f->type) {
	case NT_VAR:
	    free_var_of_filter_node(f);
	    break;
	case NT_CONST:
	    free_const_of_filter_node(f);
	    break;
	case NT_COMPOUND:
	    free_compound_of_filter_node(f);
	    break;
	case NT_VECTOR:
	    free_vec_of_filter_node(f);
	    break;
	default:
	    /* shouldn't happen. */
	    break;
	}
    }
}

/* free the given filter node, as well as all sub-nodes and */
/* heap memory it contains, directly and indirectly.        */
void
sct_free_filter_node(sct_filter_node_t* f)
{
    if (f) {
	sct_free_filter_node_contents(f);
	mem_free(f);
    }
}

