/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU Lesser General Public License                       */

/*
 * $Id: filter_type_check.c,v 1.6 2002/06/15 15:16:13 mulix Exp $
 */

/*
 * filter_type_checking: check that an operation is allowed between
 * two data types. 
 */

#include "rule_engine_private.h"
#include "mem_manager_public.h"

#include "archdep.h"

static int 
check_logical_ops(sct_data_type_t d1, sct_data_type_t d2)
{
    /* logical ops are valid for all types */
    return (d1 != DT_LAST_TYPE && d1 != DT_MULTI && 
	    d2 != DT_LAST_TYPE && d2 != DT_MULTI);
}

static int 
check_math_ops(sct_data_type_t d1, sct_data_type_t d2)
{
    return (d1 != DT_LAST_TYPE && d1 != DT_STRING && 
	    d1 != DT_FLOAT && d1 != DT_MULTI && 
	    d1 != DT_BUFFER && 
	    d2 != DT_LAST_TYPE && d2 != DT_STRING && 
	    d2 != DT_FLOAT && d1 != DT_MULTI &&
	    d2 != DT_BUFFER);
}

static int 
check_bit_ops(sct_data_type_t d1, sct_data_type_t d2)
{
    /* exactly the same check as math ops, for now */
    return check_math_ops(d1, d2);
}

static int
check_logic_not(sct_data_type_t d1, sct_data_type_t d2)
{
    /* logical ops are valid for all types */
    return (d1 != DT_LAST_TYPE && d1 != DT_MULTI);
}

static int
check_bit_not(sct_data_type_t d1, sct_data_type_t d2)
{
    return (d1 != DT_LAST_TYPE && d1 != DT_MULTI && d1 != DT_STRING);
}

static int 
check_pattern_match(sct_data_type_t d1, sct_data_type_t d2)
{
    return ((d1 == DT_STRING || d1 == DT_BUFFER) && 
	    (d2 == DT_STRING || d2 == DT_BUFFER));
}

static int
check_in_match(sct_data_type_t d1, sct_data_type_t d2)
{
    return (d1 != DT_LAST_TYPE && d2 == DT_MULTI);
}

typedef int (*type_check_func)(sct_data_type_t d1, sct_data_type_t d2);

static type_check_func check_funcs[] = 
{
    check_logical_ops,   /* OT_EQUAL */
    check_logical_ops,   /* OT_LESS */
    check_logical_ops,   /* OT_GREATER */
    check_logical_ops,   /* OT_LESS_EQUAL */
    check_logical_ops,   /* OT_GREATER_EQUAL */
    check_logical_ops,   /* OT_NOT_EQUAL */
    check_math_ops,      /* OT_PLUS */
    check_math_ops,      /* OT_MINUS */
    check_bit_ops,       /* OT_BIT_SHIFT_LEFT */
    check_bit_ops,       /* OT_BIT_SHIFT_RIGHT */
    check_bit_ops,       /* OT_BIT_AND */
    check_bit_ops,       /* OT_BIT_OR */
    check_bit_ops,       /* OT_BIT_XOR */
    check_bit_not,       /* OT_BIT_NOT */
    check_logical_ops,   /* OT_LOGIC_AND */ 
    check_logical_ops,   /* OT_LOGIC_OR */
    check_logic_not,     /* OT_LOGIC_NOT */
    check_pattern_match, /* OT_PATTERN_MATCH */
    check_in_match,	 /* OT_IN_MATCH */
    NULL,                /* OT_LAST_OPERATOR */
};

/* check that an operation 'op' is allowed between the two 
   filter nodes f1 and f2, based on their types. returns 
   1 if allowed, 0 if not */
int 
op_allowed(oper_type_t op, sct_filter_node_t* f1, sct_filter_node_t* f2)
{
    sct_data_type_t d1, d2;
    type_check_func check_fnc = NULL;

    if (!f1)
	return 0;

    d1 = f1->data_type; 
    d2 = f2 ? f2->data_type : DT_LAST_TYPE;

    assert(op < (sizeof(check_funcs)/sizeof(check_funcs[0])));
    check_fnc = check_funcs[op];

    if (!check_fnc)
	return 0;

    return check_fnc(d1, d2);
}

