/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 guy keren, choo@actcom.co.il	            */
/* License: GNU Lesser General Public License                       */

/*
 *$Id: filter_vars.c,v 1.14 2002/06/15 15:16:13 mulix Exp $
 */

#include "mem_manager_public.h"
#include "filter_vars_public.h"

/* constants. */
#define MAX_PARAM_NUM 5		/* highest supported parameter. */

typedef enum {
    CB_TYPE_NONE = 0,
    CB_TYPE_VAR,
    CB_TYPE_INDEXED,
    CB_TYPE_INDEXED_FIELD
} callback_type_t;

/* attribute of a var. */
struct var_attr {
    int is_set;
    int val;
};

/* var struct - contains data about each var */
struct var_struct {
    int persistent;
    int free_data; 
    void *value;
    struct var_attr attrs[VAN_LAST_TYPE];
    void *data;
    sct_data_type_t dtype;
    callback_type_t cb_type;
    union {
	callback_func var;
	callback_indexed_func indexed;
	callback_indexed_field_func indexed_field;
    } f;
};

/* list of variables data. */
struct var_struct vars[VT_LAST_TYPE];
struct var_struct params[MAX_PARAM_NUM];
static int var_initialized = 0;

/*
 * Private functions.
 */

/* get the value of the given parameter */
sct_filter_node_t*
get_param(int param_index)
{
    sct_filter_node_t* val = NULL;

    if (var_initialized && (param_index >= 0 && param_index < MAX_PARAM_NUM)) {
	val = params[param_index].value;
	if (val == NULL && params[param_index].cb_type == CB_TYPE_INDEXED &&
				params[param_index].f.indexed != NULL) {
	    val = params[param_index].f.indexed(VT_PARAM,
						param_index,
						params[param_index].dtype,
						params[param_index].data);
	    params[param_index].value = val;
	    /* persistent callback support for params isn't needed for now */
	    params[param_index].f.indexed = NULL; 
	    params[param_index].cb_type = CB_TYPE_NONE; 
	}
    }

    return val;
}

/* get the value of the given field of the given parameter */
sct_filter_node_t*
get_param_field(int param_index, int field_num, int struct_type)
{
    sct_filter_node_t* field = NULL;

    if (var_initialized && (param_index >= 0 && param_index < MAX_PARAM_NUM)) {
	if (params[param_index].cb_type == CB_TYPE_INDEXED_FIELD &&
				params[param_index].f.indexed_field != NULL) {
	    field = params[param_index].f.indexed_field(
					VT_PARAM,
					param_index,
					field_num,
					params[param_index].dtype,
					params[param_index].data, struct_type);
	}
    }

    return field;
}

/* get the value of the given attribute of the given parameter */
/* returns '1' if its set, '0' if its not. the value is set    */
/* into '*p_val'.                                              */
int
get_param_attr(int param_index, sct_var_attr_name_t name, int* p_val)
{
    int is_set = 0;

    if (var_initialized && (param_index >= 0 && param_index < MAX_PARAM_NUM)) {
	if (params[param_index].attrs[name].is_set) {
	    *p_val = params[param_index].attrs[name].val;
	    is_set = 1;
	}
    }

    return is_set;
}


/* set the value of the given parameter to the given value. */
void
set_param(int param_index, sct_filter_node_t* param_val)
{
    /* range checking. */
    if (param_index < 0 || param_index >= MAX_PARAM_NUM)
	return;

    sct_free_filter_node(params[param_index].value);
    params[param_index].value = param_val;
}

/* set the given attribute of the given parameter to the given value. */
void
set_param_attr(int param_index, sct_var_attr_name_t name, int val)
{
    /* range checking. */
    if (param_index < 0 || param_index >= MAX_PARAM_NUM)
	return;

    params[param_index].attrs[name].is_set = 1;
    params[param_index].attrs[name].val = val;
}

/* set the value of the given parameter to the callback function */
void
set_param_callback(int param_index,
		   callback_indexed_func func,
		   sct_data_type_t dtype,
		   void *data)
{
    /* range checking. */
    if (param_index < 0 || param_index >= MAX_PARAM_NUM)
	return;

    params[param_index].cb_type = CB_TYPE_INDEXED;
    params[param_index].f.indexed = func;
    params[param_index].data = data;
    params[param_index].dtype = dtype;
    params[param_index].free_data = 0; 
}

/* set the value of the given 2 parameters to the callback function */
void
set_param_callback2(int param_index,
		   callback_indexed_func func,
		   sct_data_type_t dtype,
		   void *data1, 
		   void *data2)
{
    struct sct_aggregate_data2* d; 

    /* range checking. */
    if (param_index < 0 || param_index >= MAX_PARAM_NUM)
	return;

    if (!(d = mem_alloc(sizeof(*d))))
	return; 
    
    d->v1 = data1; 
    d->v2 = data2; 

    params[param_index].cb_type = CB_TYPE_INDEXED;
    params[param_index].f.indexed = func;
    params[param_index].data = d; 
    params[param_index].dtype = dtype;
    params[param_index].free_data = 1; 
}

/* set the value of the given parameter's field to the callback function */
void
set_param_field_callback(int param_index,
			 callback_indexed_field_func func,
			 sct_data_type_t dtype,
			 void *data)
{
    /* range checking. */
    if (param_index < 0 || param_index >= MAX_PARAM_NUM)
	return;

    params[param_index].cb_type = CB_TYPE_INDEXED_FIELD;
    params[param_index].f.indexed_field = func;
    params[param_index].data = data;
    params[param_index].dtype = dtype;
}

/* free all allocated parameter values. */
void
free_params_list(void)
{
    int i;
    int j;

    for (i=0; i<MAX_PARAM_NUM; i++) {
        sct_free_filter_node(params[i].value);
	params[i].value = NULL;
	for (j=0; j < VAN_LAST_TYPE; j++) {
	    params[i].attrs[j].is_set = 0;
	    params[i].attrs[j].val = 0;
	}
	params[i].cb_type = CB_TYPE_NONE;
	params[i].f.indexed = NULL;

	if (params[i].free_data)
		mem_free(params[i].data); 

	params[i].data = NULL;
	params[i].dtype = DT_LAST_TYPE;
	params[i].free_data = 0; 

    }
}

/* initialized the parameter array. */
void
init_params(void)
{
    int i;
    for (i=0; i<MAX_PARAM_NUM; i++) {
	params[i].value = NULL;
	params[i].cb_type = CB_TYPE_NONE;
	params[i].f.indexed = NULL;
	params[i].data = NULL;
	params[i].dtype = DT_LAST_TYPE;
	params[i].persistent = 0;
	params[i].free_data = 0; 
    }
}

/* initialized the variables array. */
void
init_vars(void)
{
    int i;
    
    init_params();
    for (i=0; i<VT_LAST_TYPE; i++) {
	vars[i].value = NULL;
	vars[i].cb_type = CB_TYPE_NONE;
	vars[i].f.var = NULL;
	vars[i].data = NULL;
	vars[i].dtype = DT_LAST_TYPE;
	vars[i].persistent = 0;
	vars[i].free_data = 0; 
    }
    var_initialized = 1;
}


/*
 * Kernal variables manipulation code.
 */

/* empty variable values array. */
void
reset_all_vars(void)
{
    int i;
    int j;

    if (!var_initialized) {
	init_vars();
    }

    for (i=0; i<VT_LAST_TYPE; i++) {
	switch (i) {
	case VT_PARAM:
	    free_params_list();
	    break;
	default:
	    sct_free_filter_node((sct_filter_node_t*)vars[i].value);
	    vars[i].value = NULL;
	    for (j=0; j < VAN_LAST_TYPE; j++) {
		vars[i].attrs[j].is_set = 0;
		vars[i].attrs[j].val = 0;
	    }
	    if (vars[i].persistent == 0) {
	        vars[i].cb_type = CB_TYPE_NONE;
	        vars[i].f.var = NULL;
		vars[i].data = NULL;
		vars[i].dtype = DT_LAST_TYPE;
	    }
	    break;
	}
    }
}

/* store a given variable's value. */
void set_var(sct_var_type_t type, sct_filter_node_t* val)
{
    if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
	switch (type) {
	case VT_PARAM:
	    /* this is an indexed variable. */
	    break;
	default:
	    vars[type].value = val;
	    break;
	}
    }
}

/* store a given variable's callback function. */
static void _set_var_callback(sct_var_type_t type,
			      int persistent,
			      callback_func func,
			      void* data)
{
    if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
	switch (type) {
	case VT_PARAM:
	    /* this is an indexed variable. */
	    break;
	default:
	    vars[type].cb_type = CB_TYPE_VAR;
	    vars[type].f.var = func;
	    vars[type].data = data;
	    vars[type].persistent = persistent;
	    vars[type].free_data = 0; 
	    break;
	}
    }
}

void set_var_callback(sct_var_type_t type, callback_func func, void* data)
{
    _set_var_callback(type, 0, func, data);
}

void set_persistent_callback(sct_var_type_t type, callback_func func, void* data)
{
    _set_var_callback(type, 1, func, data);
}

void set_var_indexed(sct_var_type_t type, int var_index, sct_filter_node_t* val)
{
    if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
	switch (type) {
	case VT_PARAM:
	    set_param(var_index, val);
	    break;
	default:
	    /* these are not indexed variables. */
	    break;
	}
    }
}

void set_var_indexed_callback(sct_var_type_t type,
			      int var_index,
			      callback_indexed_func func,
			      sct_data_type_t dtype,
			      void *data)
{
    if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
	switch (type) {
	case VT_PARAM:
	    set_param_callback(var_index, func, dtype, data);
	    break;
	default:
	    /* these are not indexed variables. */
	    break;
	}
    }
}

void set_var_indexed_callback2(sct_var_type_t type,
			      int var_index,
			      callback_indexed_func func,
			      sct_data_type_t dtype,
			      void *data1, 
			      void *data2)
{
    if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
	switch (type) {
	case VT_PARAM:
	    set_param_callback2(var_index, func, dtype, data1, data2);
	    break;
	default:
	    /* these are not indexed variables. */
	    break;
	}
    }
}

void set_var_indexed_field_callback(sct_var_type_t type,
				    int var_index,
				    callback_indexed_field_func func,
				    sct_data_type_t dtype,
				    void *data)
{
    if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
	switch (type) {
	case VT_PARAM:
	    set_param_field_callback(var_index, func, dtype, data);
	    break;
	default:
	    /* these are not indexed variables. */
	    break;
	}
    }
}

/* evaluate the value of the given parameter, assuming it was previously */
/* set with set_var/set_var_index.                                       */
sct_filter_node_t* get_var(sct_var_type_t type)
{
    sct_filter_node_t* val = NULL;

    if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
	switch(type) {
	case VT_PARAM:
	    /* VT_PARAM is an indexed var. */
	    break;
	default:
	    val = vars[type].value;
	    if (val == NULL && vars[type].cb_type == CB_TYPE_VAR &&
						vars[type].f.var != NULL) {
		val = vars[type].f.var(type, vars[type].data);
		vars[type].value = val;
		if (vars[type].persistent == 0) {
		    vars[type].cb_type = CB_TYPE_NONE;
		    vars[type].f.var = NULL;
		}
	    }
	    break;
	}
    }

    return val;
}

sct_filter_node_t* get_var_indexed(sct_var_type_t type, int var_index)
{
    sct_filter_node_t* val = NULL;

    if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
	switch(type) {
	case VT_PARAM:
	    val = get_param(var_index);
	    break;
	default:
	    /* these are not indexed vars */
	    break;
	}
    }

    return val;
}

sct_filter_node_t* get_var_indexed_field(sct_var_type_t type,
				     int var_index,
				     int field_num,
				     int struct_type)
{
    sct_filter_node_t* val = NULL;

    if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
	switch(type) {
	case VT_PARAM:
	    val = get_param_field(var_index, field_num, struct_type);
	    break;
	default:
	    /* these are not indexed vars */
	    break;
	}
    }

    return val;
}

/* set a variable's attribute. gets reset when calling 'rest_all_vars()'. */
void
set_var_attr(sct_var_type_t type, sct_var_attr_name_t name, int val)
{
    if (type > 0 && type < VT_LAST_TYPE && var_initialized &&
				name >= 0 && name < VAN_LAST_TYPE) {
	switch (type) {
	case VT_PARAM:
	    /* this is an indexed variable. */
	    break;
	default:
	    vars[type].attrs[name].is_set = 1;
	    vars[type].attrs[name].val = val;
	    break;
	}
    }
}

void
set_var_indexed_attr(sct_var_type_t type, int var_index,
		     sct_var_attr_name_t name, int val)
{
    if (type > 0 && type < VT_LAST_TYPE && var_initialized &&
				name >= 0 && name <= VAN_LAST_TYPE) {
	switch (type) {
	case VT_PARAM:
	    set_param_attr(var_index, name, val);
	    break;
	default:
	    /* these are not indexed variables. */
	    break;
	}
    }
}


/* get a variable's attribute. returns '1' if its defined, '0' if not. */
/* value itself stores in '*p_val'.                                    */
int
get_var_attr(sct_var_type_t type, sct_var_attr_name_t name, int* p_val)
{
    int is_set = 0;

    if (type > 0 && type < VT_LAST_TYPE && var_initialized &&
			name >= 0 && name < VAN_LAST_TYPE && p_val) {
	switch(type) {
	case VT_PARAM:
	    /* VT_PARAM is an indexed var. */
	    break;
	default:
	    if (vars[type].attrs[name].is_set) {
	        *p_val = vars[type].attrs[name].val;
		is_set = 1;
	    }
	    break;
	}
    }

    return is_set;
}

int
get_var_indexed_attr(sct_var_type_t type, int var_index,
		     sct_var_attr_name_t name, int* p_val)
{
    int is_set = 0;

    if (type > 0 && type < VT_LAST_TYPE && var_initialized &&
					name >= 0 && name < VAN_LAST_TYPE) {
	switch(type) {
	case VT_PARAM:
	    is_set = get_param_attr(var_index, name, p_val);
	    break;
	default:
	    /* these are not indexed vars */
	    break;
	}
    }

    return is_set;
}

