/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 guy keren, choo@actcom.co.il	            */
/* License: GNU Lesser General Public License                       */

/*
 *$Id: filter_vector.c,v 1.9 2002/06/15 15:16:13 mulix Exp $
 */
/*
 * Rule engine - implementation of a vector of filter nodes.
 */

#include "rule_engine_private.h"
#include "encode_decode_public.h"
#include "mem_manager_public.h"

#include "archdep.h"
/*
 * Private functions.
 */

/* free the 'vec' part of a filter node. don't free the node itself. */
void
free_vec_of_filter_node(struct tracker_filter_node* f)
{
    assert(f);
	/* nothing to do, since a var node contains only basic data types. */
    if (f->u.vec.nodes) {
	int i;
	for (i=0; i<f->u.vec.len; i++) {
	    if (f->u.vec.nodes[i])
		sct_free_filter_node(f->u.vec.nodes[i]);
	}
	mem_free(f->u.vec.nodes);
	f->u.vec.len = 0;
	f->u.vec.nodes = NULL;
    }
}

/* print a vector node. */
void
sct_print_filter_node_vec(struct tracker_filter_node* f, unsigned int indent)
{
    int i;

    assert(f);

    print_indent(indent);
    PPRINT("vector[%d]:\n", f->u.vec.len);
    print_indent(indent);
    for (i=0; i<f->u.vec.len; i++) {
	PPRINT("f%d:\n", i);
	if (f->u.vec.nodes[i]) {
	    do_sct_print_filter_node(f->u.vec.nodes[i], indent+INDENT_LEVEL);
	}
	else {
	    print_indent(indent);
	    PPRINT("Empty\n");
	}
	print_indent(indent);
    }
}

/*
 * Implementation of public filter building functions.
 */

/* create a vector filter node, with the given size. */
/* returns NULL if there's not enough memory.        */
sct_filter_node_t*
sct_make_filter_vector(int len)
{
    struct tracker_filter_node* node =
		(struct tracker_filter_node*)
			mem_alloc(sizeof(struct tracker_filter_node));
    if (node) {
	node->type = NT_VECTOR;
	node->data_type = DT_MULTI;
	node->u.vec.len = len;
	node->u.vec.nodes = mem_alloc(sizeof(sct_filter_node_t*) * len);
	if (!node->u.vec.nodes) {
	    mem_free(node);
	    node = NULL;
	}
	else {
	    int i;
	    for (i=0; i<len; i++)
		node->u.vec.nodes[i] = NULL;
	}
    }

    return node;
}

/* inserts the given filter node into the given vector node. */
/* returns '1' on success, '0' if 'vec' is not a vector,     */
/* '-1' if index is out of range.                            */
int
sct_insert_filter_node(sct_filter_node_t* vec, int index, sct_filter_node_t* f)
{
    int rc = 1;  /* assume success. */

    assert(vec);
    
    if (vec->type != NT_VECTOR) {
	rc = 0;
	goto sct_insert_filter_node_ret;
    }
    if (index < 0 || index >= vec->u.vec.len) {
	rc = -1;
	goto sct_insert_filter_node_ret;
    }
    if (vec->u.vec.nodes[index]) {
	/* free former value in the vector's cell, if there was any. */
	sct_free_filter_node(vec->u.vec.nodes[index]);
    }
    vec->u.vec.nodes[index] = f;

  sct_insert_filter_node_ret:
    return rc;
}

/* return the filter node in the given index of the given vector.    */
/* returns NULL if index is out of range, or vec is not a DT_VECTOR. */
sct_filter_node_t*
sct_get_filter_node_at(sct_filter_node_t* vec, int index)
{
    sct_filter_node_t* f = NULL;    /* assume failure. */

    assert(vec);

    if (vec->type == NT_VECTOR && index >= 0 && index < vec->u.vec.len) {
	f = vec->u.vec.nodes[index];
    }

    return f;
}

/* return the length of the given vector node. returns '-1'  */
/* if 'vec' is not a vector.				     */
int
sct_get_filter_vector_len(sct_filter_node_t* vec)
{
    assert(vec);

    if (vec->type == NT_VECTOR)
	return vec->u.vec.len;
    else
	return -1;
}

/* serialize a 'vector' filter node. */
int
sct_serialize_filter_node_vector(sct_filter_node_t* f,
			     unsigned char* buf,
			     int buflen,
			     unsigned char** pos)
{
    int vec_len = sct_get_filter_vector_len(f);
    sct_filter_node_t* node;
    int rc = -1;  /* assume failure */

    if (vec_len > -1) {
        rc = encode_char(buf, buflen, pos, NT_VECTOR);
	if (rc > 0) {
            rc = encode_int(buf, buflen, pos, vec_len);
	}

        if (rc > 0) {
	    int i;

	    for (i=0; rc > 0 && i < vec_len; i++) {
		node = sct_get_filter_node_at(f, i);
	        rc = do_sct_serialize_filter_node(node, buf, buflen, pos);
	    }
        }
    }

    return rc;
}

/* de-serialize a 'vector' filter node. */
int
sct_deserialize_filter_node_vector(sct_filter_node_t** pf,
				   unsigned char* buf,
				   int buflen,
				   unsigned char** pos)
{
    int vec_len;
    sct_filter_node_t* vec = NULL;
    sct_filter_node_t* node;
    int rc = decode_int(buf, buflen, pos, &vec_len);

    if (rc > 0) {
	vec = sct_make_filter_vector(vec_len);
	if (vec) {
	    int i;

	    for (i=0; rc > 0 && i < vec_len; i++) {
	        rc = do_sct_deserialize_filter_node(&node, buf, buflen, pos);
		if (rc > 0) {
		    sct_insert_filter_node(vec, i, node);
		}
	    }
	}
	else {
	    rc = 0;
	}
    }

    if (rc > 0) {
	*pf = vec;
    }
    else {
	if (vec)
	    sct_free_filter_node(vec);
    }

    return rc;
}
