/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU Lesser General Public License                       */

/* 
 * $Id: membuf.c,v 1.3 2002/06/15 15:16:13 mulix Exp $
 */

#include "archdep.h"
#include "membuf.h"

struct sct_membuf {
	unsigned char* data; 
	int owns_data;   /* 1 if we malloc'd the data, 0 if it was externally supplied */ 

	unsigned int rh; /* read head */ 
	unsigned int wh; /* write head */ 
	
	size_t size;     /* total data size */ 
	size_t free;     /* free (for writing) space */ 

	membuf_copy_fn tofn;    /* to membuf copy fn */ 
	membuf_copy_fn fromfn;  /* from membuf copy fn */ 
}; 

sct_membuf_t* sct_membuf_create(size_t size, membuf_copy_fn tofn, 
				membuf_copy_fn fromfn)
{
	sct_membuf_t* m = mem_alloc(sizeof(struct sct_membuf)); 
	if (!m)
		return NULL; 

	m->data = mem_alloc(size); 
	if (!m->data) {
		mem_free(m); 
		return NULL; 
	}

	memset(m->data, 0, size); 
	
	m->owns_data = 1; 

	m->rh = m->wh = 0; 

	m->free = m->size = size; 

	m->tofn = tofn; 
	m->fromfn = fromfn; 

	return m; 
}

sct_membuf_t* sct_membuf_create_with_data(unsigned char* data, size_t size, 
					  membuf_copy_fn tofn, membuf_copy_fn fromfn)
{
	sct_membuf_t* m = mem_alloc(sizeof(struct sct_membuf)); 
	if (!m)
		return NULL; 

	m->owns_data = 0; 
	m->data = data; 
	m->size = m->wh = size; 
	m->rh = 0; 
	m->free = 0; 

	m->tofn = tofn; 
	m->fromfn = fromfn; 

	return m; 
}

void sct_membuf_destroy(sct_membuf_t* m)
{
	assert(m); 
	
	if (!m) 
		return; 

	if (m->data) {
		/* only free it if we own it */ 
		if (m->owns_data)
			mem_free(m->data);
		/* but set it to NULL anyway */ 
		m->data = NULL; 
	}
	/* set the fields to 0 to catch usage after free() asap */ 
      	m->rh = m->wh = m->free = m->size = 0;
	mem_free(m); 
}

int
sct_membuf_add_chunk(sct_membuf_t* m, const void* chunk, size_t chunk_size)
{
	unsigned char* whead; 

	assert(m);
	/* allow a NULL chunk if the size is 0, */ 
	/* allow a 0 size if the chunk is NULL */ 
	if (!m || (!chunk && chunk_size) || (chunk && !chunk_size))
		return -1; 

	if (chunk_size > m->free)
		chunk_size = m->free; 

	if (chunk && chunk_size) {
		whead = m->data + m->wh; 
		if (m->tofn) {
			if (m->tofn(whead, chunk, chunk_size))
				/* if an error occurs, return it */ 
				return -1; 
		} else {
			memcpy(whead, chunk, chunk_size); 
		}
		m->wh += chunk_size; 
		m->free -= chunk_size; 
	}
	
	return chunk_size; 
}

int 
sct_membuf_get_chunk(sct_membuf_t* m, void* chunk, size_t chunk_size)
{
	unsigned char* rhead; 

	assert(m); 
	if (!m || (!chunk && chunk_size) || (chunk && !chunk_size))
		return -1; 
	
	if (chunk_size > (m->wh - m->rh))
		chunk_size = (m->wh - m->rh);

	if (chunk && chunk_size) {
		rhead = m->data + m->rh; 
		if (m->fromfn) {
			if (m->fromfn(chunk, rhead, chunk_size))
				/* if an error occures, return it */ 
				return -1; 
		} else {
			memcpy(chunk, rhead, chunk_size); 
		}
		m->rh += chunk_size; 
	}

	return chunk_size; 
}

unsigned char*
sct_membuf_get_data(sct_membuf_t* m)
{
	assert(m); 

	if (!m)
		return NULL; 

	/* might be NULL! */ 
	return m->data; 
}

size_t sct_membuf_get_serialized_size(sct_membuf_t* m)
{
	assert(m); 

	if (!m)
		return 0; 

	return m->wh;
}

size_t sct_membuf_get_size(sct_membuf_t* m)
{
	assert(m); 

	if (!m)
		return 0; 

	return m->size; 
}

/* available for reading */ 
size_t sct_membuf_get_available(sct_membuf_t* m)
{
	assert(m); 

	if (!m)
		return 0; 
	
	return m->wh - m->rh; 
}

size_t sct_membuf_get_free(sct_membuf_t* m)
{
	assert(m); 

	if (!m)
		return 0; 

	return m->free; 
}

void sct_membuf_clear(sct_membuf_t* m)
{
	assert(m); 

	if (!m) 
		return; 

	m->rh = m->wh = 0; 
	m->free = m->size; 
}
