/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU Lesser General Public License                       */

/* 
 * $Id: membuf.h,v 1.3 2002/06/15 15:16:13 mulix Exp $
 */

#ifndef MEMBUF_H_
#define MEMBUF_H_

#include "archdep.h"
#include "rule_engine_public.h"

/* we want to allow membuf users to supply their own functions for copying data  */ 
/* from the membuf / to the membuf. this is crucial, for example, for the kernel,*/ 
/* where data needs to be copied over the user space / kernel space divide       */ 
/* the copy function should return != 0 on error or 0 if succesful               */ 
typedef unsigned long(*membuf_copy_fn)(void* to, const void* from, unsigned long sz);  

/* create a membuf of size 'size         */ 
/* returns the membuf or NULL on failure */ 
sct_membuf_t* sct_membuf_create(size_t size, membuf_copy_fn to_membuf, 
				membuf_copy_fn from_membuf); 

/* create a membuf which holds the preallocated data which is of size 'size' */ 
/* returns the membuf or NULL on failure. Note that the data is not copied,  */
/* and the membuf will NOT release its memory when destructed                */ 
sct_membuf_t* sct_membuf_create_with_data(unsigned char* data, size_t size, 
					  membuf_copy_fn to_membuf, 
					  membuf_copy_fn from_membuf); 

/* destroy the given membuf */ 
void sct_membuf_destroy(sct_membuf_t* m);

/* add the chunk 'chunk' to the membuf 'm'. if m doesn't have enough free     */
/* space, if any parameter is invalid or an error occurs, -1  is returned. If */ 
/* if the addition is succesful, returns the size of the serialized data -    */ 
/* which might be 0.                                                          */ 
int sct_membuf_add_chunk(sct_membuf_t* m, const void* chunk, size_t chunk_size);

/* get a chunk from the membuf. the chunk is copied into 'chunk', which is     */
/* of size 'chunk_size'. If any parameter is invalid, -1 is returned. Returns  */
/* the ammount of bytes in 'chunk', (which can be less than chunk_size)        */
int sct_membuf_get_chunk(sct_membuf_t* m, void* chunk, size_t chunk_size);

/* get a pointer to the data the membuf holds. This returns a pointer to the */
/* entire data buffer, even if parts of it have already been read!           */
unsigned char* sct_membuf_get_data(sct_membuf_t* m);

/* get the size of the serialized data */
size_t sct_membuf_get_serialized_size(sct_membuf_t* m); 

/* get the size this membuf was created with */ 
size_t sct_membuf_get_size(sct_membuf_t* m); 

/* get the ammount of space available for adding a chunk */ 
size_t sct_membuf_get_free(sct_membuf_t* m);

/* get ammount of available data for reading from the membuf */ 
size_t sct_membuf_get_available(sct_membuf_t* m); 

/* "reset" the membuf */ 
void sct_membuf_clear(sct_membuf_t* m); 

#endif /* MEMBUF_H_ */ 
