/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 guy keren, choo@actcom.co.il	            */
/* License: GNU Lesser General Public License                       */

/*
 *$Id: rule_engine.c,v 1.45 2002/06/15 15:16:13 mulix Exp $
 */

#include "mem_manager_public.h"
#include "encode_decode_public.h"
#include "rule_engine_private.h"
#include "filter_vars_public.h"
#include "membuf.h"
#include "archdep.h"

/* a list of tracking rules. */
#define	MAX_NUM_TRACKER_RULES	20

struct rule_list {
    struct tracker_rule* rules[MAX_NUM_TRACKER_RULES];
    int num_rules;
};

/* maximum size of a log record. */
#define MAX_LOG_RECORD	1024

/* types of logging macros, used when logging syscall invocations. */
typedef enum {
    LM_PID = 0,
    LM_UID = 0,
    LM_EUID = 0
} log_macro_t;

/* maximum number of supported logger callbacks. */
#define	MAX_LOGGER_CALLBACKS	5

/* logger callbacks array. */
struct logger_callback logger_callbacks[MAX_LOGGER_CALLBACKS];
static int num_logger_callbacks = 0;


/*
 * Private functions.
 */

/* find the location of a rule with the given ID in the rules list. */
/* returns the rule's location, or '-1' if none is found.           */
int
find_tracking_rule(sct_rule_list_t* rules, int rule_id)
{
    int i;

    if (!rules)
	return -1;

    for (i = 0; i < rules->num_rules; i++) {
	if (rules->rules[i] && rules->rules[i]->rule_id == rule_id)
	    break;
    }

    return (i < rules->num_rules ? i : -1);
}

/* add the given rule to the rules list.                                */
/* returns the rule's location on success, or '-1' if an error occured. */
int
do_sct_add_tracking_rule(sct_rule_list_t* rules,
		     int rule_id,
		     sct_filter_node_t* filter,
		     sct_action_t* action)
{
    int i;
    int j;
    struct tracker_rule* rule;

    /* make sure we have free space. */
    if (!rules || rules->num_rules >= MAX_NUM_TRACKER_RULES) {
	return -1;
    }

    /* allocate a structure for the new rule. */
    rule = mem_alloc(sizeof(struct tracker_rule));
    if (!rule) {
	return -1;
    }

    rule->rule_id = rule_id;
    rule->filter = filter;
    rule->action = action; 

    /* OPTIMIZATION: use something a little saner than this linear search */ 
    /* find the right location for the new rule. */
    for (i = 0; i < rules->num_rules; i++) {
	if (rules->rules[i] && rules->rules[i]->rule_id > rule_id)
	    break;
    }
	
    /* make a free space for the new rule. */
    for (j = rules->num_rules; j > i; j--) {
	rules->rules[j] = rules->rules[j-1];
    }
	
    /* insert the new rule. */
    rules->rules[i] = rule;
    rules->num_rules++;
	
    return (i);
}

/* serialize a 'sct_action_t' action node. */
int
do_sct_serialize_action_node(sct_action_t* a,
			 unsigned char* buf,
			 int buflen,
			 unsigned char** pos)
{
    int rc = 0;

    if (a) {
	char sct_action_type = a->type;
	rc = encode_char(buf, buflen, pos, sct_action_type);
	if (rc > 0) {
	    switch (sct_action_type) {
	    case AT_LOG:
		rc = encode_int(buf,
				buflen,
				pos,
				a->u.log.priority);
		if (rc > 0) {
		    rc = encode_string(buf,
				       buflen,
				       pos,
				       a->u.log.format);
		}
		break;
	    case AT_RULE_ENABLE:
		rc = encode_int(buf,
				buflen, pos, a->u.rule_id);
		break;
	    case AT_RULE_DISABLE:
		rc = encode_int(buf,
				buflen, pos, a->u.rule_id);
		break;
	    case AT_PROCESS_SUSPEND: /* fallthrough */ 
	    case AT_PROCESS_KILL:
		rc = encode_int(buf,
				buflen, pos, a->u.pid);
		break;
	    case AT_SYSCALL_FAIL:
		rc = encode_int(buf,
				buflen, pos, a->u.error_code);
		break;
	    default:
		rc = 0;
		break;
	    }
	}
	if (rc > 0) {
	    /* encode attr set commands. */
	    rc = encode_int(buf, buflen, pos, a->attr_set_cmds_len);
	    if (rc > 0) {
		int i;

		for (i=0; rc > 0 && i < a->attr_set_cmds_len; i++) {
		    struct tracker_action_attr_set_data* atsd =
			&(a->attr_set_cmds[i]);
		    rc = encode_int(buf, buflen, pos, atsd->var_type);
		    if (rc > 0) {
		        rc = encode_int(buf, buflen, pos, atsd->var_index);
		    }
		    if (rc > 0) {
		        rc = encode_int(buf, buflen, pos, atsd->name);
		    }
		    if (rc > 0) {
		        rc = encode_int(buf, buflen, pos, atsd->val);
		    }
		}
	    }
	}
    }
    else {
	rc = encode_char(buf, buflen, pos, AT_NIL);
    }

    return rc;
}

/* de-serialize an 'sct_action_t' action node. */
int
do_sct_deserialize_action_node(sct_action_t** pa,
			       unsigned char* buf,
			       int buflen,
			       unsigned char** pos)
{
    char sct_action_type;
    int rc = decode_char(buf, buflen, pos, &sct_action_type);

    if (rc > 0) {
	switch (sct_action_type) {
	case AT_LOG:
	{
	    int priority;
	    char* format;
	    rc = decode_int(buf,
			    buflen, pos, &priority);
	    if (rc > 0) {
		rc = decode_string(buf,
				   buflen,
				   pos,
				   &format);
	    }
	    if (rc > 0) {
		*pa = sct_make_action_log(priority,
					  format);
		if (format)
		    mem_free(format);

		if (!*pa) {
    	    	    PPRINT(" do_sct_deserialize_action_node: failed sct_make_action_log\n");
		    rc = 0;
		}
	    }
	}
	break;
	case AT_RULE_ENABLE:
	{
	    int rule_id;
	    rc = decode_int(buf,
			    buflen, pos, &rule_id);
	    if (rc > 0) {
		*pa = sct_make_action_enable_rule(
		    rule_id);
		if (!*pa) {
		    rc = 0;
		}
	    }
	}
	break;
	case AT_RULE_DISABLE:
	{
	    int rule_id;
	    rc = decode_int(buf,
			    buflen, pos, &rule_id);
	    if (rc > 0) {
		*pa = sct_make_action_disable_rule(
		    rule_id);
		if (!*pa) {
		    rc = 0;
		}
	    }
	}
	break;
	case AT_PROCESS_SUSPEND:
	{
	    int pid; 
	    rc = decode_int(buf,
			    buflen, pos, &pid);
	    if (rc > 0) {
		*pa = sct_make_action_suspend_process(pid);
		if (!*pa) {
		    rc = 0;
		}
	    }
	}
	break;
	case AT_PROCESS_KILL:
	{
	    int pid; 
	    rc = decode_int(buf,
			    buflen, pos, &pid);
	    if (rc > 0) {
		*pa = sct_make_action_kill_process(pid);
		if (!*pa) {
		    rc = 0;
		}
	    }
	}
	break;
	case AT_SYSCALL_FAIL:
	{
	    int error_code;
	    rc = decode_int(buf,
			    buflen,
			    pos,
			    &error_code);
	    if (rc > 0) {
		*pa = sct_make_action_fail_syscall(
		    error_code);
		if (!*pa) {
		    rc = 0;
		}
	    }
	}
	break;
	case AT_NIL:
	    /* a NULL action_node. */
	    *pa = NULL;
	    rc = 1;
	    break;
	default:
	    /* shouldn't happen. */
	    rc = 0;
	    break;
	}
	if (pa && *pa && rc > 0) {
	    /* decode attr set commands. */
	    rc = decode_int(buf, buflen, pos, &((*pa)->attr_set_cmds_len));
	    if ((*pa)->attr_set_cmds_len > MAX_ATTR_SET_PER_ACTION) {
		/* enforce sane limits, to avoid abuse by user-code */
		(*pa)->attr_set_cmds_len = MAX_ATTR_SET_PER_ACTION;
	    }
	    if (rc > 0) {
		int i;
		int var_type;
		int attr_name;

		for (i=0; rc > 0 && i < (*pa)->attr_set_cmds_len; i++) {
		    struct tracker_action_attr_set_data* atsd =
			&((*pa)->attr_set_cmds[i]);
		    rc = decode_int(buf, buflen, pos, &var_type);
		    atsd->var_type = var_type;
		    if (rc > 0) {
		        rc = decode_int(buf, buflen, pos, &atsd->var_index);
		    }
		    if (rc > 0) {
		        rc = decode_int(buf, buflen, pos, &attr_name);
		        atsd->name = attr_name;
		    }
		    if (rc > 0) {
		        rc = decode_int(buf, buflen, pos, &atsd->val);
		    }
		}
	    }
	    /* if we failed decoding attr commands - free the action. */
	    if (rc <= 0) {
		sct_free_action(*pa);
		*pa = NULL;
	    }
	}
    }

    return rc;
}


/*
 * Functions for actions manipulation.
 */

/* allocate an action object for logging purposes. */
sct_action_t*
sct_make_action_log(int priority, const char* format)
{
    /* OPTIMIZATION: keep static actions and return pointers to them */
    struct tracker_action* action =
	mem_alloc(sizeof(struct tracker_action));
    if (action) {
	action->attr_set_cmds_len = 0;
	action->type = AT_LOG;
	action->u.log.priority = priority;
	if (format) {
	    action->u.log.format =
		(char*)mem_alloc(strlen(format)+1);
	    if (action->u.log.format)
		strcpy(action->u.log.format, format);
	    else {
		mem_free(action);
		action = NULL;
	    }
	}
	else {
	    action->u.log.format = NULL;
	}
    }

    return action;
}

/* allocate an action object for rule enabling purposes. */
sct_action_t*
sct_make_action_enable_rule(int rule_id)
{
    struct tracker_action* action =
	mem_alloc(sizeof(struct tracker_action));
    if (action) {
	action->attr_set_cmds_len = 0;
	action->type = AT_RULE_ENABLE;
	action->u.rule_id = rule_id;
    }

    return action;
}

/* allocate an action object for rule disabling purposes. */
sct_action_t*
sct_make_action_disable_rule(int rule_id)
{
    struct tracker_action* action =
	mem_alloc(sizeof(struct tracker_action));
    if (action) {
	action->attr_set_cmds_len = 0;
	action->type = AT_RULE_DISABLE;
	action->u.rule_id = rule_id;
    }

    return action;
}

/* allocate an action object for process suspending. */
sct_action_t* sct_make_action_suspend_process(int pid)
{
    struct tracker_action* action =
	mem_alloc(sizeof(struct tracker_action));
    if (action) {
	action->attr_set_cmds_len = 0;
	action->type = AT_PROCESS_SUSPEND;
	action->u.pid = pid; 
    }

    return action;
}

/* allocate an action object for process killing. */
sct_action_t* sct_make_action_kill_process(int pid)
{
    struct tracker_action* action =
	mem_alloc(sizeof(struct tracker_action));
    if (action) {
	action->attr_set_cmds_len = 0;
	action->type = AT_PROCESS_KILL;
	action->u.pid = pid; 
    }

    return action;
}

/* allocate an action object for system call failing. */
sct_action_t* sct_make_action_fail_syscall(int error_code)
{
    struct tracker_action* action =
	mem_alloc(sizeof(struct tracker_action));
    if (action) {
	action->attr_set_cmds_len = 0;
	action->type = AT_SYSCALL_FAIL;
	action->u.error_code = error_code;
    }

    return action;
}

/* set action variable attribute set command. */
void set_action_var_attr_set_command(sct_action_t* a,
				     sct_var_type_t var_type,
				     sct_var_attr_name_t name,
				     int val)
{
    if (a && a->attr_set_cmds_len < MAX_ATTR_SET_PER_ACTION) {
	a->attr_set_cmds[a->attr_set_cmds_len].var_type = var_type;
	a->attr_set_cmds[a->attr_set_cmds_len].var_index = -1;
	a->attr_set_cmds[a->attr_set_cmds_len].name = name;
	a->attr_set_cmds[a->attr_set_cmds_len].val = val;
	a->attr_set_cmds_len++;
    }
}

void set_action_var_indexed_attr_set_command(sct_action_t* a,
				 	     sct_var_type_t var_type,
				 	     int var_index,
				 	     sct_var_attr_name_t name,
				 	     int val)
{
    if (a && a->attr_set_cmds_len < MAX_ATTR_SET_PER_ACTION) {
	a->attr_set_cmds[a->attr_set_cmds_len].var_type = var_type;
	a->attr_set_cmds[a->attr_set_cmds_len].var_index = var_index;
	a->attr_set_cmds[a->attr_set_cmds_len].name = name;
	a->attr_set_cmds[a->attr_set_cmds_len].val = val;
	a->attr_set_cmds_len++;
    }
}

void sct_free_action(sct_action_t* a)
{
    if (a) {
	switch(a->type) {
	case AT_LOG:
	    if (a->u.log.format)
		mem_free(a->u.log.format);
	    break;
	default: break; /* do nothing */
	}
	mem_free(a);
    }
}


/*
 * Action serialization/de-serialization functions.
 */

/* serialize an action node into a buffer, for export     */
/* purposes. returns the size of the serialized data, 0   */
/* if there was an error, -1 if buffer is too small.      */
int
sct_serialize_action_node(sct_action_t* a, unsigned char* buf, int buflen)
{
    unsigned char* pos = buf;
    int rc = do_sct_serialize_action_node(a, buf, buflen, &pos);

    return (rc > 0 ? pos - buf : rc);
}

/* de-serialize an action node from a buffer, for import  */
/* purposes. returns the new allocated action node or     */
/* NULL if an error occuerd.                              */
sct_action_t*
sct_deserialize_action_node(unsigned char* buf, int buflen)
{
    unsigned char* pos = buf;
    sct_action_t* a = NULL;
    int rc = 0;

    if (buflen > 0) {
	rc = do_sct_deserialize_action_node(&a, buf, buflen, &pos);
    }

    return (rc > 0 ? a : NULL);
}

#define ACTION_AS_STR_BUFLEN 512

/* get a string representation of this action */
char* sct_get_action_as_string(sct_action_t* a)
{
    char* buf; 
    char* lf; 
    int i;

    if (!a)
	return NULL;

    /* should be large enough for everybody */
    buf = mem_alloc(ACTION_AS_STR_BUFLEN); 
    if (!buf)
	return NULL;

    switch (a->type) {
    case AT_LOG:
	lf = a->u.log.format;
	
	/* be very careful not to overflow our buffer with the */
	/* user supplied log format */
	if (lf && (strlen(lf) > 200))
	    lf = "[truncated]";
		
	/* cant use snprintf, as long as we support 2.2 */
	/* unless we implement it ourselves, of course. */
	sprintf(buf, "log: priority %d, log format %s", 
		a->u.log.priority, lf);
	break;
    case AT_RULE_ENABLE:
	sprintf(buf, "enable_rule, rule id %d", a->u.rule_id);
	break;
    case AT_RULE_DISABLE:
	sprintf(buf, "disable_rule, rule id %d", a->u.rule_id);
	break;
    case AT_PROCESS_SUSPEND:
	sprintf(buf, "suspend process %d", a->u.pid);
	break;
    case AT_PROCESS_KILL:
	sprintf(buf, "kill process %d", a->u.pid);
	break;
    case AT_SYSCALL_FAIL:
	sprintf(buf, "fail syscall, with error code %d", a->u.error_code);
	break;
    default:
	sprintf(buf, "unknown");
	break;
    }
    /* log attr set commands. */
    for (i=0; i < a->attr_set_cmds_len; i++) {
	struct tracker_action_attr_set_data* atsd = &a->attr_set_cmds[i];
	char* pb = buf+strlen(buf);
	if (ACTION_AS_STR_BUFLEN - (pb - buf) > 50) {
	    sprintf(pb, "; attr set: var %d(%d), attr %d = %d",
		    atsd->var_type, atsd->var_index, atsd->name, atsd->val);
	}
	else {
	    sprintf(pb, "; attr set truncated");
	}
    }

    return buf;
}

void sct_get_syscall_id_as_str(int syscall_id, char* buf, size_t sz)
{
	/* ugly hackery, since we dont have snprintf() in all 2.4 kernels */ 
	/* strlen of UINT_MAX == 10 */ 
	const unsigned int minbufsz = 22; 

	assert(buf && sz); 
	
	if (!buf || (sz < minbufsz))
		return; 

	/* if it's a mux syscall */ 
	if (FUNC_PART_OF_SYSCALL_ID(syscall_id)) {
		sprintf(buf, "%d:%d", 
			SYSCALL_PART_OF_SYSCALL_ID(syscall_id), 
			FUNC_PART_OF_SYSCALL_ID(syscall_id)); 
	} else {
		sprintf(buf, "%d", syscall_id); 
	}
}

/* print an action node to a "standard" device */
void sct_print_action_node(sct_action_t* a)
{
    char* desc = sct_get_action_as_string(a);
    PPRINT(" Action: %s\n", desc ? desc : "(nil)");
    mem_free(desc);
}

/*
 * Functions for tracking rules manipulation.
 */

/* serialize a tracking rule into a buffer, for export   */
/* purposes. returns the size of the serialized data, 0  */
/* if there was an error, -1 if the buffer is too small. */
int sct_serialize_tracking_rule(int rule_id, sct_rule_when_t when, int syscall_id,
			    sct_action_t* a, sct_filter_node_t* f,
		            unsigned char* buf, int buflen)
{
    unsigned char* pos_1 = buf;
    int rc = encode_int(buf, buflen, &pos_1, rule_id);
    if (rc > 0) {
	rc = encode_int(buf, buflen, &pos_1, when);
        if (rc > 0) {
	    rc = encode_int(buf, buflen, &pos_1, syscall_id);
	    if (rc > 0) {
	        int action_size =
		    sct_serialize_action_node(a, pos_1, buflen-(pos_1-buf));
	        if (action_size > 0) {
		    unsigned char* pos_2 = pos_1+action_size;
		    int filter_size =
		        sct_serialize_filter_node(f,
					      pos_2,
					      buflen-(pos_2-buf));
		    if (filter_size > 0) {
		        rc = (pos_2 - buf) + filter_size;
		    }
		    else {
		        rc = filter_size;
		    }
	        }
	        else {
		    rc = action_size;
	        }
	    }
        }
    }

    return rc;
}

/* de-serialize a tracking rule from a buffer, for import */
/* purposes. returns the 1 for success, -1 for failure.   */
/* the given action rule and filter node are filled with  */
/* elements geenrated from the buffer, and should be      */
/* freed by the caller.                                   */
int sct_deserialize_tracking_rule(unsigned char* buf, int buflen, 
				  int* p_rule_id, sct_rule_when_t* p_when,
				  int* p_syscall_id,
				  sct_action_t** pa, sct_filter_node_t** pf)
{
    unsigned char* pos_1 = buf;
    int rc = decode_int(buf, buflen, &pos_1, p_rule_id);

    if (rc > 0) {
	rc = decode_int(buf, buflen, &pos_1, (int*)p_when);
        if (rc > 0) {
	    rc = decode_int(buf, buflen, &pos_1, p_syscall_id);
	    if (rc > 0) {
	        unsigned char* pos_2 = pos_1;
	        rc = do_sct_deserialize_action_node(pa,
						    pos_1,
						    buflen-(pos_1-buf),
						    &pos_2);
	        if (rc > 0) {
		    unsigned char* pos_3 = pos_2;
		    rc = do_sct_deserialize_filter_node(pf,
							pos_2,
							buflen-(pos_2-buf),
							&pos_3);
		    if (rc == 0) { 
			sct_free_action(*pa);
			*pa = NULL;
		    }
	        }
	    }
        }
    }

    return rc;
}

/* create an empty tracking rules list. */
sct_rule_list_t*
sct_create_tracking_rule_list()
{
    sct_rule_list_t* rules = mem_alloc(sizeof(struct rule_list));
    if (rules) {
	rules->num_rules = 0;
    }

    return rules;
}

/* delete the given tracking rule list, freeing any rule found in it. */
void
sct_delete_tracking_rule_list(sct_rule_list_t* rules)
{
	if (rules) {
		sct_del_all_tracking_rules_in_list(rules);
		mem_free(rules);
	}
}

/* add a new tracking rule. returns the new rule's id     */
/* returns '-1' if an error occured (including if there's */
/* already a rule with the given rule ID).                 */
int
sct_add_tracking_rule(sct_rule_list_t* rules,
		  int rule_id,
		  sct_filter_node_t* filter,
		  sct_action_t* action)
{
    /* assume failure. */
    int res = -1;

    if (filter && action) {
	/* make sure we don't already have a rule with the same ID. */ 
	int rule_loc = find_tracking_rule(rules, rule_id);
	if (rule_loc >= 0) 
		return -1; 

	/* add the rule, sorted by rule ID. */
	res = do_sct_add_tracking_rule(rules,
				       rule_id,
				       filter,
				       action);
    }
    return res;
}

void do_del_tracking_rule(struct tracker_rule* rule)
{
	if (rule->filter) {
	    sct_free_filter_node(rule->filter);
	}
	if (rule->action) {
	    sct_free_action(rule->action); 
	}
	mem_free(rule);
}

/* delete a tracking rule. */
void
sct_del_tracking_rule(sct_rule_list_t* rules, int rule_id)
{
    int i;

    if (!rules)
	    return; 

    i = find_tracking_rule(rules, rule_id);

    if (i >= 0 && i < rules->num_rules) {
	struct tracker_rule* rule = rules->rules[i];
	int j;

	/* we're removing a rule - close the 'gap' created in the list. */
	for (j = i; j < rules->num_rules-1; j++) {
	    rules->rules[j] = rules->rules[j+1];
	}
	/* set to NULL the last rule slot, since it's now in slot last - 1*/ 
	rules->rules[rules->num_rules-1] = NULL; 
	do_del_tracking_rule(rule); 
	rules->num_rules--;
    }
}

/* deletes all tracking rules in the list. */
void
sct_del_all_tracking_rules_in_list(sct_rule_list_t* rules)
{
    int i;

    if (!rules)
	    return; 

    for (i = 0; i < rules->num_rules; i++) {
	if (rules->rules[i]) {
            do_del_tracking_rule(rules->rules[i]); 
	    rules->rules[i] = NULL;
	}
    }
    rules->num_rules = 0;
}

/* check if there is a rule with the given ID. */
/* returns '1' if there is, '0' otehrwise.     */
int
sct_has_tracking_rule(sct_rule_list_t* rules, int rule_id)
{
    int i = find_tracking_rule(rules, rule_id);

    return (i >= 0 ? 1 : 0);
}

/* returns the number of rules in the given rule list. */
int
sct_get_tracking_rule_count(sct_rule_list_t* rules)
{
    return (rules ? rules->num_rules : 0);
}

/* print out the list of current tracking rules. */
void sct_print_tracking_rules(sct_rule_list_t* rules)
{
    int i;
    if (!rules) {
	    PPRINT("rule list is NULL!\n"); 
	    return; 
    }

    PPRINT("\nActive Tracking Rules List (%d):\n", rules->num_rules);
    for (i=0; i < rules->num_rules; i++) {
	/* skip empty rules */
	if (!rules->rules[i])
	    continue;
	PPRINT(" Rule id: %d\n", rules->rules[i]->rule_id);
	sct_print_action_node(rules->rules[i]->action);
	PPRINT(" Filter:\n");
	sct_print_filter_node(rules->rules[i]->filter);
    }
    PPRINT("\n");
}

/*
 * Functions used by tracking functions to perform filtering.
 */

/* filter an incoming system call.                                          */
/* accepts a system call ID and its (var_args) parameters.                  */
/* returns a tracker_action structure if there was a match, NULL otherwise. */
/* on successfull match, the rule id is set into '*rule_id'.                */
sct_action_t* sct_filter_syscall(sct_rule_list_t* rules, int syscall_id, int* rule_id, ...)
{
    int i;
    sct_action_t* action = NULL;
    sct_filter_node_t* result; 

    if (!rules)
	    return NULL; 

    result = sct_make_bool_val(TF_FALSE);

    for (i=0; i<rules->num_rules; i++) {
	if (rules->rules[i] && rules->rules[i]->filter) {
	    sct_exp_error_t rc =
		sct_eval_filter_node(rules->rules[i]->filter, result);
	    if (rc == ERR_NONE && cast_bool(result)) 
		break;
	}
    }

    if (i < rules->num_rules) {
	action = rules->rules[i]->action;
	if (rule_id)
	    *rule_id = rules->rules[i]->rule_id;
    }

    sct_free_filter_node(result);

    return action;
}

/* find the registered logger calback function, that handles  */
/* logging a given type of spec-char. returns the callback ID */
/* if one was registered, '-1' otherwise.                     */
int
get_logger_callback_index(char spec_char)
{
    int i;

    for (i=0; i < num_logger_callbacks; i++) {
	if (logger_callbacks[i].spec_char == spec_char)
	    break;
    }

    return (i < num_logger_callbacks ? i : -1);
}

/* stringify the given string value, into the given buffer,           */
/* with the given size. returns size of stringified string on succes, */
/* 0 on buffer overflow.                                              */
static int
tracker_stringify_string(const char* string, char* buf, int buflen, int add_quotes)
{
    char* pb = buf;

    /* we want to output '"string"' */
    if (add_quotes && pb - buf < buflen)
        *pb++ = '"';
    if (string) {
        if (strlen(string) < buflen-(pb-buf)) {
	    strcpy(pb, string);
	    pb += strlen(string);
        } else { /* try to output 100 bytes or less f the string if we can */ 
	    int cnt = sct_min(buflen - sizeof("..."), 100); 
	    if (cnt > 0) {
		    strncpy(pb, string, cnt);
		    pb += cnt; 
		    strcpy(pb, "..."); 
		    pb += sizeof("..."); 
	    }
	    
	}
    } else { /* the string was null */
        if (strlen("(nil)") < buflen-(pb-buf)) {
	    strcpy(pb, "(nil)");
	    pb += strlen("(nil)");
        }
    }
    /* emit the trailing '"'. */
    if (add_quotes && pb - buf < buflen) {
	*pb++ = '"';
    }

    return (pb - buf < buflen ? pb - buf : 0);
}

/* stringify the given unsigned numeric value, into the given buffer, */
/* with the given size. returns size of stringified num on succes,    */
/* 0 on buffer overflow.                                              */
static int
tracker_stringify_unum(unsigned long num, char* buf, int buflen)
{
    char sbuf[MAX_LOG_RECORD+1];
    int rc = 0;
    
    if (buflen > 32) {
        rc = sprintf(sbuf, "%lu", num);
        if (rc < buflen) {
	    strcpy(buf, sbuf);
        }
    }

    return (rc > 0 && rc < buflen ? rc : 0);
}

/* stringify the given pointer value, into the given buffer,   */
/* with the given size. returns size of stringified pointer on */
/* succes, 0 on buffer overflow.                               */
static int
tracker_stringify_pointer(void* p, char* buf, int buflen)
{
    char sbuf[MAX_LOG_RECORD+1];
    int rc = 0;
    
    if (buflen > 32) {
        rc = sprintf(sbuf, "%p", p);
        if (rc < buflen) {
	    strcpy(buf, sbuf);
        }
    }

    return (rc > 0 && rc < buflen ? rc : 0);
}

/* stringify the given numeric value, into the given buffer, with the */
/* given size. returns length of stringified num on succes, 0 on      */
/* buffer overflow.                                                   */
static int
tracker_stringify_num(long num, char* buf, int buflen)
{
    char sbuf[MAX_LOG_RECORD+1];
    int rc = 0;
    
    if (buflen > 32) {
        rc = sprintf(sbuf, "%ld", num);
        if (rc < buflen) {
	    strcpy(buf, sbuf);
        }
    }

    return (rc > 0 && rc < buflen ? rc : 0);
}

/* stringify the given variable into the given buffer.                */
/* returns the length of the stringified var, '0' on buffer overflow. */
static int
tracker_stringify_var_node(char* buf, int buflen, sct_filter_node_t *var_val)
{
    int rc = 0;
    int len = 0;
    char* string = NULL;
    char* pb;

    if (var_val && var_val->type == NT_CONST) {
	switch (var_val->data_type) {
	case DT_USHORT:
	    rc = tracker_stringify_unum(CONST_USHORT(var_val), buf, buflen);
	    break;
	case DT_SHORT:
	    rc = tracker_stringify_num(CONST_SHORT(var_val), buf, buflen);
	    break;
	case DT_UINT:
	    rc = tracker_stringify_unum(CONST_UINT(var_val), buf, buflen);
	    break;
	case DT_INT:
	    rc = tracker_stringify_num(CONST_INT(var_val), buf, buflen);
	    break;
	case DT_ULONG:
	    rc = tracker_stringify_unum(CONST_ULONG(var_val), buf, buflen);
	    break;
	case DT_LONG:
	    rc = tracker_stringify_num(CONST_LONG(var_val), buf, buflen);
	    break;
	case DT_STRING:
	    string = CONST_STRING(var_val);
	    rc = tracker_stringify_string(string, buf, buflen, 1);

	    pb = buf + rc;
	    if (pb - buf < buflen)
		*pb = '\0';
	    else
		*(pb - 1) = '\0';
	    break;
	case DT_BUFFER:
	    string = sct_buffer_to_string(CONST_BUFFER(var_val), 
					  BUFFER_SIZE(var_val)); 
	    rc = tracker_stringify_string(string, buf, buflen, 1); 
	    mem_free(string); 
	    string = NULL; 

	    pb = buf + rc;
	    if (pb - buf < buflen)
		*pb = '\0';
	    else
		*(pb - 1) = '\0';
	    break;
	default:
	    break;
	}
    }
    
    if (rc)
	len = strlen(buf);

    return (rc ? len : 0);
}

static int
tracker_stringify_var(char* buf, int buflen, sct_var_type_t var)
{
    return tracker_stringify_var_node(buf, buflen, get_var(var));
}

static int
tracker_stringify_var_indexed(char* buf, int buflen, sct_var_type_t var, int var_index)
{
    sct_filter_node_t* t = get_var_indexed(var, var_index); 
    return tracker_stringify_var_node(buf, buflen, t);
}

/* stringify the parameters of the syscall. */
static int 
tracker_stringify_params(char* buf,
			 int buflen,
			 const char* param_spec,
			 va_list ap)
{
    char* pb = buf;
    char c;
    unsigned char uc;
    long l;
    unsigned long ul;
    int rc;
    void* v, *p;

    if (param_spec) {
	int next_is_unsigned = 0;
	char* string = NULL; 
	char this_char = 0; 
	char next_char = 0; 
	int is_string; /* strings require a bit of special handling */ 
	int skip_params = 0; /* are we in 'parameters skipping' mode? */
	int logger_cb_id = -1;

	while (*param_spec && pb - buf < buflen) {
	    this_char = *param_spec++;
	    /* handle logger callbacks first. */
	    logger_cb_id = get_logger_callback_index(this_char);
	    if (logger_cb_id > -1) {
		void* p = va_arg(ap, void*);
		void* p_data = va_arg(ap, void*);
		rc = (*logger_callbacks[logger_cb_id].cb)(
				this_char,
				pb,
				buflen - (pb - buf),
				p,
				p_data,
				logger_callbacks[logger_cb_id].cb_data);
		pb += rc;

		goto tracker_stringify_params_next_char;
	    }

	    is_string = (this_char == 's');

	    switch (this_char){
	    case ' ': break; /* skip spaces */
	    case 'u': next_is_unsigned = 1; break;
	    case 'c':
		if (next_is_unsigned) {
		    uc = (unsigned char)va_arg(ap, unsigned int);
		    if (!skip_params) {
		        if (pb - buf < buflen)
		            *pb++ = uc;
		    }
		    next_is_unsigned = 0;
		} else {
		    c = (char)va_arg(ap, int);
		    if (!skip_params) {
		        if (pb - buf < buflen)
		            *pb++ = c;
		    }
		}
		break;
            case 'h':
		if (next_is_unsigned) {
		    ul = (unsigned short)va_arg(ap, unsigned int);
		    if (!skip_params) {
		        rc = tracker_stringify_unum(ul, pb, buflen-(pb-buf));
		        pb += rc;
		    }
		    next_is_unsigned = 0;
		} else {
		    l = (short)va_arg(ap, int);
		    if (!skip_params) {
		        rc = tracker_stringify_num(l, pb, buflen-(pb-buf));
		        pb += rc;
		    }
		}
		break;
	    case 'i':
		if (next_is_unsigned) {
		    ul = va_arg(ap, unsigned int);
		    if (!skip_params) {
		        rc = tracker_stringify_unum(ul, pb, buflen-(pb-buf));
		        pb += rc;
		    }
		    next_is_unsigned = 0;
		} else {
		    l = va_arg(ap, int);
		    if (!skip_params) {
		        rc = tracker_stringify_num(l, pb, buflen-(pb-buf));
		        pb += rc;
		    }
		}
		break;
	    case 'l':
		if (next_is_unsigned) {
		    ul = va_arg(ap, unsigned long);
		    if (!skip_params) {
		        rc = tracker_stringify_unum(ul, pb, buflen-(pb-buf));
		        pb += rc;
		    }
		    next_is_unsigned = 0;
		} else {
		    l = va_arg(ap, long);
		    if (!skip_params) {
		        rc = tracker_stringify_num(l, pb, buflen-(pb-buf));
		        pb += rc;
		    }
		}
		break;
	    case 'p':
		p = va_arg(ap, void*);
		if (!skip_params) {
		    rc = tracker_stringify_pointer(p, pb, buflen-(pb-buf));
		    pb += rc;
		}
		break;
	    case 'P': /* indexed var */
		l = va_arg(ap, int);
		if (!skip_params) {
		    rc = tracker_stringify_var_indexed(pb, buflen-(pb-buf), VT_PARAM, l);
		    pb += rc;
		}
		break;
	    case 'B': /* indexed var */ 
		l = va_arg(ap, int);
		if (!skip_params) {
		    rc = tracker_stringify_var_indexed(pb, buflen-(pb-buf), VT_PARAM, l);
		    pb += rc;
		}
		break; 
	    case 'T': /* type-name */
		string = va_arg(ap, char*);
		if (!skip_params) {
		    rc = tracker_stringify_string(string,
						  pb,
						  buflen-(pb-buf),
						  0);
		    pb += rc;
		}
		break;
	    case 's': /* string */
	    case 'S': /* struct */
		string = va_arg(ap, char*);
		if (!skip_params) {
		    rc = tracker_stringify_string(string,
						  pb,
						  buflen-(pb-buf),
						  is_string);
		    pb += rc;

		    /* for a string, we want to output '"string"' */
		    if (is_string) {
			/* for strings, we are done */
			break;
		    }
		    /* for structures, the next arg is a pointer to the */
		    /* _user space_ structure we only care if it's NULL */
		    /* or not, and we never ever dereference it.        */
		    v = va_arg(ap, void*);
		    if (v == NULL)
			skip_params = 1;
		}
		break;
	    case '{':
		if (pb - buf < buflen)
		    *pb++ = '{';
		break;
	    case '}':
		if (pb - buf < buflen)
		    *pb++ = '}';
		if (skip_params)
		    skip_params = 0;
		break;
	    case 'v':
		if (!skip_params) {
		    if (strlen("void") < buflen-(pb-buf)) {
		        strcpy(pb, "void");
		        pb += strlen("void");
		    }
		}
		break;
	    default:
		break;
	    }

	  tracker_stringify_params_next_char:
	    /* if we just printed a value, print a comma after it */
	    next_char = *param_spec;
	    if (logger_cb_id > -1 && next_char && next_char != '}') {
		if (!skip_params) {
		    if (strlen(", ") < buflen-(pb-buf)) {
		        strcpy(pb, ", ");
		        pb += strlen(", ");
		    }
		}
	    }
	    else if ((this_char != ' ') && (this_char != 'u') && 
		     (this_char != 'S') && (this_char != 'T') && 
		     (this_char != '{') && (next_char) && next_char != '}') {  
		if (!skip_params) {
		    if (strlen(", ") < buflen-(pb-buf)) {
		        strcpy(pb, ", ");
		        pb += strlen(", ");
		    }
		}
	    }
	}
    }

    return pb - buf;
}

/* perform macro translation into the given buffer with the given size. */
/* returns the size of data writtern into the buffer.                   */
static int 
tracker_translate_macro(int rule_id, 
			int syscall_id, 
			const char* syscall_name, 
			const char* macro, 
			char* buf, 
			int buflen, 
			const char* param_spec, 
			va_list ap)
{
    int rc;
    int len = 0;

    if (!strcmp(macro, "ruleid")) {
	rc = tracker_stringify_num(rule_id, buf, buflen);
	if (rc)
	    len = strlen(buf);
    }
    else if (!strcmp(macro, "sid")) {
	rc = tracker_stringify_num(syscall_id, buf, buflen);
	if (rc)
	    len = strlen(buf);
    }
    else if (!strcmp(macro, "sname")) {
	if (strlen(syscall_name) < buflen - 1) {
	    strcpy(buf, syscall_name);
	    len = strlen(buf);
	}
    }
    else if (!strcmp(macro, "params")) {
	rc = tracker_stringify_params(buf, buflen, param_spec, ap);
	len = rc;
    }
    else if (!strcmp(macro, "pid")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_PID);
	len = rc;
    }
    else if (!strcmp(macro, "uid")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_UID);
	len = rc;
    }
    else if (!strcmp(macro, "euid")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_EUID);
	len = rc;
    }
    else if (!strcmp(macro, "suid")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_SUID);
	len = rc;
    }
    else if (!strcmp(macro, "gid")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_GID);
	len = rc;
    }
    else if (!strcmp(macro, "egid")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_EGID);
	len = rc;
    }
    else if (!strcmp(macro, "sgid")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_SGID);
	len = rc;
    }
    else if (!strcmp(macro, "comm")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_COMM);
	len = rc;
    }
    else if (!strcmp(macro, "ppid")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_PPID);
	len = rc;
    }
    else if (!strcmp(macro, "pcomm")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_PCOMM);
	len = rc;
    }
    else if (!strcmp(macro, "retval")) {
	rc = tracker_stringify_var(buf, buflen, VT_RETVAL);
	len = rc;
    }

    return len;
}

/* log the syscall that matched one of the rules in 'rules' */
static void 
tracker_do_logging(int rule_id, 
		   int syscall_id, 
		   const char* syscall_name, 
		   const char* log_format, 
		   const char* param_spec, 
		   va_list ap)
{
    char buf[MAX_LOG_RECORD+1];
    char sbuf[MAX_LOG_RECORD+1];
    const char* pf = log_format;
    char* pb = buf;
    char const * ps;
    int len, mlen;

    while (*pf && pb - buf < MAX_LOG_RECORD) {
	if (*pf == '%') {
	    /* find next semicolon. */
	    if (*++pf) {
		if (*pf == '%') {
		    /* translate '%%' to '%'. */
		    pf++;
		    if (pb - buf < MAX_LOG_RECORD) {
			*pb++ = '%';
			pf++;
			continue;
		    }
		}
		else {
		    /* translate macro name into value. */
		    ps = pf; 
		    /* parse out the macro name - from the current char
		       until the next non alphabetic character */
		    while (isalpha(*ps++))
			;
		    ps--;
		    mlen = sct_min(ps - pf, MAX_LOG_RECORD - (pb - buf));
		    strncpy(sbuf, pf, mlen);
		    sbuf[mlen] = '\0';
		    len = tracker_translate_macro(rule_id,
					          syscall_id,
					          syscall_name,
					          sbuf,
					          pb,
					          MAX_LOG_RECORD - (pb - buf),
					          param_spec,
					          ap);
		    pf = ps;
		    pb += len;
		    continue;
		}
	    }
	}
	else {
	    if (pb - buf < MAX_LOG_RECORD) {
		*pb++ = *pf++;
		continue;
	    }
	}
    }
    *pb = '\0';
    buf[MAX_LOG_RECORD-1] = '\n';
    if (pb > buf && pb-buf < MAX_LOG_RECORD) {
	*(pb-1) = '\n';
	*(pb) = '\0';
    }
    buf[MAX_LOG_RECORD-2] = '.';
    buf[MAX_LOG_RECORD-3] = '.';
    buf[MAX_LOG_RECORD-4] = '.';
    PPRINT(KERN_NOTICE "%s", buf);
}


/*
 * public functions.
 */

/* register the given logger calback function, that handles */
/* logging a given type of spec-char                        */
void
set_logger_callback(char spec_char, logger_callback_t cb, void* cb_data)
{
    int i;

    for (i=0; i < num_logger_callbacks; i++) {
	if (logger_callbacks[i].spec_char == spec_char)
	    break;
    }

    if (i == num_logger_callbacks) {
	if (i >= MAX_LOGGER_CALLBACKS) {
	    goto set_logger_callback_exit;
	}
	num_logger_callbacks++;
    }
    logger_callbacks[i].spec_char = spec_char;
    logger_callbacks[i].cb = cb;
    logger_callbacks[i].cb_data = cb_data;

  set_logger_callback_exit:
    return;
}


/* stringify the parameters of the syscall. */
int 
sct_tracker_stringify_params(char* buf,
			     int buflen,
			     const char* param_spec,
			     ...)
{
    int rc;
    va_list ap;
    va_start (ap, param_spec);

    rc = tracker_stringify_params(buf, buflen, param_spec, ap);
    va_end(ap);

    return rc;
}

/* perform an action in the current system call's context. */
/* accepts a system call ID and its (var_args) parameters. */
/* returns an 'sct_syscall_retval_policy'. if it's         */
/* SCT_OVERRIDE_RETVAL, the new retval to use will be put  */
/* in 'syscall_retval'. otherwise, it's left untouched.    */ 
sct_syscall_retval_policy
sct_tracker_perform_action(sct_action_t* action,
			   int rule_id, 
			   int syscall_id, 
			   int *syscall_retval, 
			   const char* syscall_name, 
			   const char* param_spec, 
			   ...)
{
    sct_syscall_retval_policy retval = SCT_USE_ORIG_RETVAL; 
    int i;

    if (action) {
	/* TODO TODO - handle other action types. */
	/* also handle logging with varargs.      */

	/* perform set attribute commands. */
	for (i=0; i < action->attr_set_cmds_len; i++) {
	    struct tracker_action_attr_set_data* atsd =
		&action->attr_set_cmds[i];
	    if (atsd->var_index == -1)
		set_var_attr(atsd->var_type, atsd->name, atsd->val);
	    else
		set_var_indexed_attr(atsd->var_type,
				     atsd->var_index,
				     atsd->name,
				     atsd->val);
	}

	switch (action->type) {
	case AT_LOG:
	{
	    const char* log_format = action->u.log.format;
	    va_list ap;
	    va_start (ap, param_spec);

	    if (!log_format || !*log_format)
		log_format =
		    "syscall: %pid[%comm]: %sid_%sname(%params) (rule %ruleid)\n";
	    tracker_do_logging(rule_id,
			       syscall_id,
			       syscall_name,
			       log_format,
			       param_spec,
			       ap);
	    va_end(ap);
	    break;
	}
	case AT_SYSCALL_FAIL:
	{
	    PPRINT(KERN_NOTICE "failing syscall '%s(%d)' with '%d'\n",
		   syscall_name, syscall_id, action->u.error_code);
	    retval = SCT_OVERRIDE_RETVAL; 
	    if (syscall_retval)
		    *syscall_retval = action->u.error_code;
	    break;
	}
	case AT_PROCESS_SUSPEND:
	{
	    PPRINT(KERN_NOTICE "(disabled) suspending pid '%d'\n",
		   action->u.pid);
	    break; 
	}
	case AT_PROCESS_KILL:
	{
	    PPRINT(KERN_NOTICE "(disabled) killing process '%d'\n",
		   action->u.pid);
	    break;
	}
	default:
	    PPRINT(KERN_NOTICE "syscall '%s(%d)' invoked - unknown action %d\n",
		   syscall_name, syscall_id, action->type);
	    break;
	}
    }
    return retval;
}

/* serialize a list of tracking rules into a buffer,        */
/* for export purposes. returns the size of the serialized  */
/* data, 0 if there was an error, -1 if the buffer was      */
/* too small.                                               */ 
int sct_serialize_tracking_rule_list(sct_rule_list_t* rules, 
				     sct_membuf_t* mbuf)
{
	int i; 
	int rule_len, rc; 
	char tempbuf[1000]; /* a bit of ugliness, until serialize/deserialize 
			       rules supports membufs */ 

	if (!rules || !mbuf)
		return 0; 

	/* first store the number of rules in the list */ 
	if ((rc = sct_membuf_add_chunk(mbuf, &rules->num_rules, 
				       sizeof(rules->num_rules))) < 0) {
		PPRINT("serialize_tracing_rule_list: adding the number "
		       "of rules in the buffer failed with error %d\n", rc); 
		return rc; 
	}
	
	/* for each rule in the list, first store its length, */ 
	/* and then store the rule itself */ 
	for (i = 0; i < rules->num_rules; i++) {
		struct tracker_rule* r;

		/* NULL rule, should never happen */ 
		if (!rules->rules[i]) {
			PPRINT("rule_list_t rules has a NULL rule at location '%d'\n", i); 
			continue; 
		}
		
		r = rules->rules[i]; 
		rule_len = sct_serialize_tracking_rule(r->rule_id, 
						       0, /* bogus when, we never use it */ 
						       0, /* bogus syscall id */ 
						       r->action, 
						       r->filter, 
						       tempbuf, 
						       sizeof(tempbuf)); 
		/* we have the serialized rule in tempbuf, now copy it to the membuf */ 
		if (rule_len <= 0) {
			PPRINT("sct_serialize_tracking_rule_list: "
			       "rule_len is %d, aborting\n", rule_len); 
			return rule_len; 
		}
		rc = sct_membuf_add_chunk(mbuf, &rule_len, sizeof(rule_len)); 
		if (rc < 0) {
			PPRINT("sct_serialize_tracking_rule_list: "
			       "adding rule_len returned %d\n", rc); 
			return rc; 
		} 
		rc = sct_membuf_add_chunk(mbuf, tempbuf, rule_len); 
		if (rc < 0) {
			PPRINT("sct_serialize_tracking_rule_list: "
			       "adding the tempbuf returned %d\n", rc); 
			PPRINT("tempbuf size is %d, membuf size is %d with "
			       "%d available space\n", rule_len, 
			       sct_membuf_get_size(mbuf),
			       sct_membuf_get_free(mbuf)); 
			return rc; 
		}
	}
	return sct_membuf_get_serialized_size(mbuf); 
}

/* deserialize a tracking rule_list from a buffer, for import      */
/* purposes. returns 1 for success, -1 for failure. the rule list  */
/* pointed to by p_rules is filled with rules and should be freed  */
/* by the caller (with sct_delete_tracking_rule_list)              */
int sct_deserialize_tracking_rule_list(sct_membuf_t* mbuf, 
				       sct_rule_list_t** p_rules)
{
	int i; 
	int num_rules; 
	int rule_len; 
	sct_rule_list_t* list; 
	char tempbuf[1000]; /* a bit of ugliness, until serialize/deserialize */
	                    /* rules supports membufs */ 

	int rule_id;
	sct_rule_when_t when; 
	int syscall_id; 
	sct_action_t* a; 
	sct_filter_node_t* f; 

	if (!p_rules || !mbuf)
		return -1; 

	/* get the number of rules stored in this list */ 
	if (sct_membuf_get_chunk(mbuf, &num_rules, sizeof(num_rules)) 
	    != sizeof(num_rules))
		return -1; 

	/* create a rule list with this number of rules */ 
	list = sct_create_tracking_rule_list(); 
	if (!list)
		return -1; 

	/* for each rule, we stored its length first and then the rule itself */ 
	for (i = 0; i < num_rules; i++) {
		if (sct_membuf_get_chunk(mbuf, (unsigned char*)&rule_len, sizeof(rule_len))
		    != sizeof(rule_len)) 
			goto free_rule_list; 

		/* some more ugliness, due to membuf/char* incompatibilities */ 
		if (sct_membuf_get_chunk(mbuf, tempbuf, rule_len) != rule_len)
			goto free_rule_list; 
		
		if (sct_deserialize_tracking_rule(tempbuf, rule_len, 
						  &rule_id, &when, &syscall_id, 
						  &a, &f) != 1)
			goto free_rule_elements; 

		/* we have succesfully deserialized a rule. now add it to the list of rules */ 
		if (sct_add_tracking_rule(list, rule_id, f, a) < 0)
			goto free_rule_elements; 
	}
	/* whee, success! */ 
	*p_rules = list; 
	return 1; 	

free_rule_elements:
	sct_free_filter_node(f); 
	sct_free_action(a); 
free_rule_list:
	sct_delete_tracking_rule_list(list);
	return -1;
}
