/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 guy keren, choo@actcom.co.il	            */
/* License: GNU Lesser General Public License                       */

/*
 *$Id: rule_engine_private.h,v 1.24 2002/06/15 15:16:14 mulix Exp $
 */

#ifndef _RULE_ENGINE_PRIVATE_H
#define _RULE_ENGINE_PRIVATE_H

/* printing macro. */
#include "archdep.h"

/*
 * Private data types of the rules matching engine.
 */

#include "rule_engine_public.h"

/*
 * constants.
 */

/* number of spaces for each indent level. */
#define	INDENT_LEVEL	4

/*
 * data types used by the tracking module.
 */

/* types of supported nodes. */
typedef enum {
    NT_NIL = 0,
    NT_VAR,
    NT_CONST,
    NT_COMPOUND,
    NT_VECTOR
} node_type_t;

/* types of supported operators. */
typedef enum {
    OT_EQUAL = 0,
    OT_LESS,
    OT_GREATER,
    OT_LESS_EQUAL,
    OT_GREATER_EQUAL,
    OT_NOT_EQUAL,
    OT_PLUS,
    OT_MINUS,
    OT_BIT_SHIFT_LEFT,
    OT_BIT_SHIFT_RIGHT,
    OT_BIT_AND,
    OT_BIT_OR,
    OT_BIT_XOR,
    OT_BIT_NOT,
    OT_LOGIC_AND,
    OT_LOGIC_OR,
    OT_LOGIC_NOT,
    OT_PATTERN_MATCH,
    OT_IN_MATCH,
    OT_LAST_OPERATOR
} oper_type_t;

/* operators symbols, used for printing. */
extern const char* operators_symbols[];

/* 'var' tracker filter node - contains a single variable. */
struct tracker_filter_node_var {
    sct_var_type_t type;	/* type of variable.              	       */
    int var_index;	/* index, if an indexed variable. 	       */
    int field_index;	/* field index, if a var's field.  	       */
    int struct_type;	/* type of struct for interpreting the vector. */
};

/* 'constant' tracker filter node - contains a single constant. */
struct tracker_filter_node_const {
    sct_data_type_t type;	/* type of constant. */
    union {
	int b;			    /* bool           */
	char c;	                    /* char           */
	int i;			    /* int            */
	unsigned int ui;	    /* unsigned int   */
	long l;			    /* long           */
	unsigned long ul;	    /* unsigned long  */
	short s;		    /* short          */
	unsigned short us;	    /* unsigned short */
	float f;		    /* float          */
	char* str;		    /* string         */
	struct {                    /* buffer         */ 
	    void* data;             /* buffer data    */  
	    unsigned int size;      /* buffer size    */ 
	} buf; 
    } u;
};

/* constant filter node access macros. */
#define	CONST_DATA_TYPE(cs)	((cs)->u.constant.type)
#define	CONST_BOOL(cs)		((cs)->u.constant.u.b)
#define	CONST_CHAR(cs)		((cs)->u.constant.u.c)
#define	CONST_SHORT(cs)		((cs)->u.constant.u.s)
#define	CONST_USHORT(cs)	((cs)->u.constant.u.us)
#define	CONST_INT(cs)		((cs)->u.constant.u.i)
#define	CONST_UINT(cs)		((cs)->u.constant.u.ui)
#define	CONST_LONG(cs)		((cs)->u.constant.u.l)
#define	CONST_ULONG(cs)		((cs)->u.constant.u.ul)
#define	CONST_FLOAT(cs)		((cs)->u.constant.u.f)
#define	CONST_STRING(cs)	((cs)->u.constant.u.str)
#define CONST_BUFFER(cs)        ((cs)->u.constant.u.buf.data)

/* for a buffer, we also need access to its size */ 
#define BUFFER_SIZE(cs)        ((cs)->u.constant.u.buf.size)


/* 'compound' tracker filter node - contains operator and one or more nodes. */
struct tracker_filter_node_compound {
    oper_type_t op;			   /* type of operator. 	      */
    struct tracker_filter_node* node1; /* first node.     		      */
    struct tracker_filter_node* node2; /* second node - binary operators. */
};

/* 'vector' tracker filter node - contains a vector of filter nodes. */
struct tracker_filter_node_vector {
    int len;		        /* length of vector.      */
    sct_filter_node_t** nodes;  /* array of filter nodes. */
};

/* tracker filter structure. */
struct tracker_filter_node {
    node_type_t type;		        /* type of node. */
    sct_data_type_t data_type;		/* data type of node, as inherited. */
					/* from its contained nodes.	    */
    union {
	struct tracker_filter_node_var var;
	struct tracker_filter_node_const constant;
	struct tracker_filter_node_compound compound;
	struct tracker_filter_node_vector vec;
    } u;
};

/* types of actions that may be taken. */
typedef enum {
    AT_NIL = 0,
    AT_LOG,
    AT_RULE_ENABLE,
    AT_RULE_DISABLE,
    AT_PROCESS_SUSPEND,
    AT_PROCESS_KILL,
    AT_SYSCALL_FAIL,
    AT_LAST_ACTION_TYPE
} sct_action_type_t;

/* maximum number of attribute set operations supported, per action. */
#define	MAX_ATTR_SET_PER_ACTION	3

/* tracker attr set command data */
struct tracker_action_attr_set_data {
    sct_var_type_t var_type;
    int var_index;		/* -1 for non-indexed vars. */
    sct_var_attr_name_t name;
    int val;
};

/* tracker action structure. */
struct tracker_action {
    sct_action_type_t type;
    struct tracker_action_attr_set_data attr_set_cmds[MAX_ATTR_SET_PER_ACTION];
    int attr_set_cmds_len;
    union {
	struct _log {
	    int priority;	/* priority for message logging. */
	    char* format;	/* format for message logging.   */
	} log;
	int rule_id;	/* rule ID to be enabled/disabled.    */
	int error_code; /* error code when failing a syscall. */
	int pid;        /* pid when killing or suspending a system call */ 
    } u;
};

/* tracker rule structure. */
struct tracker_rule {
    int rule_id;
    struct tracker_filter_node* filter;
    struct tracker_action* action;
};

/* logger callbacks data. */
struct logger_callback {
    char spec_char;
    logger_callback_t cb;
    void* cb_data;
};

/* node building function - used by various node-building routines. */
struct tracker_filter_node*
sct_make_compound_node_with_type(oper_type_t op,
		   	     struct tracker_filter_node* node1,
		   	     struct tracker_filter_node* node2,
		   	     sct_data_type_t data_type);

/* filter node serialize/deserialize forward declarations. */
int do_sct_serialize_filter_node(sct_filter_node_t* f,
		             unsigned char* buf,
		             int buflen,
		             unsigned char** pos);
int do_sct_deserialize_filter_node(sct_filter_node_t** pf,
				unsigned char* buf,
				int buflen,
				unsigned char** pos);

/* free the 'vec' part of a filter node. don't free the node itself. */
void free_vec_of_filter_node(struct tracker_filter_node* f);

/* free the given filter node contents, as well as all sub-nodes and */
/* heap memory it contains, directly and indirectly.		     */
/* but leave the node memory alloced				     */
void sct_free_filter_node_contents(sct_filter_node_t* f);

/* compare two data types. return 0 if they are equal, -1  */
/* if the first takes more space, 1 if the second takes    */
/* more space.                                             */
int typecmp(sct_data_type_t t1, sct_data_type_t t2);

/* check that an operation 'op' is allowed between the two */
/* filter nodes f1 and f2, based on their types. returns   */
/* 1 if allowed, 0 if not                                  */
int op_allowed(oper_type_t op, sct_filter_node_t* f1, sct_filter_node_t* f2);

/* recursively print the given filter node, beginning */
/* with the given indentation level.                  */
void do_sct_print_filter_node(struct tracker_filter_node* f, unsigned int indent);

/* print a vector node. */
void sct_print_filter_node_vec(struct tracker_filter_node* f, unsigned int indent);

/* serialize a 'vector' filter node. */
int sct_serialize_filter_node_vector(sct_filter_node_t* f,
				     unsigned char* buf,
				     int buflen,
				     unsigned char** pos);

/* de-serialize a 'vector' filter node. */
int sct_deserialize_filter_node_vector(sct_filter_node_t** pf,
				       unsigned char* buf,
				       int buflen,
				       unsigned char** pos);

/* print spaces in the ammount given. */
void print_indent(unsigned int len);

/* print the buffer into a string. caller must free the resulting string */ 
char* sct_buffer_to_string(void* buf, unsigned int size); 

/* cast a filter node to the various forms */ 

/* cast a constant value into a boolean value, similarly   */
/* to how this is doen in C. Exception for a string value. */
int cast_bool(struct tracker_filter_node* val);

/* cast a constant value into a char value */
char cast_char(struct tracker_filter_node* val);

/* cast a constant value into a short value */
short cast_short(struct tracker_filter_node* val);

/* cast a constant value into a unsigned short value */
unsigned short cast_ushort(struct tracker_filter_node* val); 

/* cast a constant value into a int value */
int cast_int(struct tracker_filter_node* val);

/* cast a constant value into a uint value */
unsigned int cast_uint(struct tracker_filter_node* val);

/* cast a constant value into a long value */
long cast_long(struct tracker_filter_node* val);

/* cast a constant value into a unsigned long value */
unsigned long cast_ulong(struct tracker_filter_node* val);

/* cast a constant value into a string value, similarly    */
/* to how this is doen in C. Strings are strings. numbers  */
/* are being stringified.                                  */
/* the caller should free the resulting string.            */
char* cast_string(struct tracker_filter_node* val); 

/* cast a constant value into a buffer value, the caller */
/* should free the resulting buffer.                     */
void cast_buffer(struct tracker_filter_node* val, void** buf, unsigned int* size);

#endif /* _RULE_ENGINE_PRIVATE_H */
