/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 guy keren, choo@actcom.co.il	            */
/* License: GNU Lesser General Public License                       */

/*
 *$Id: rule_engine_public.h,v 1.32 2002/06/15 15:16:14 mulix Exp $
 */

#ifndef _RULE_ENGINE_PUBLIC_H
#define _RULE_ENGINE_PUBLIC_H


/*
 * Public interface of the rules matching engine.
 */

/*
 * forward declerations of data types used by the tracking module.
 */

#ifdef __cplusplus 
extern "C" {
#endif

/* values used for booleans. */
#define	TF_TRUE		1
#define	TF_FALSE	0

/* opaque tracker filter structure. */
typedef struct tracker_filter_node sct_filter_node_t;

/* opaque tracker action structure. */
typedef struct tracker_action sct_action_t;

/* opaque tracker rules list structure. */
typedef struct rule_list sct_rule_list_t;

/* opaque memory buffer structure */ 
typedef struct sct_membuf sct_membuf_t; 

/* enum containing supported variable types. */
typedef enum {
	VT_UNDEF = 0,
	VT_SYSCALL_ID,
	VT_PROC_UID,
	VT_PROC_EUID,
	VT_PROC_SUID,
	VT_PROC_GID,
	VT_PROC_EGID,
	VT_PROC_SGID,
	VT_PROC_PID,
	VT_PROC_COMM,
	VT_RETVAL,
	VT_PROC_PPID,
	VT_PROC_PCOMM,
	VT_PROC_ANCESTORS,
	VT_PARAM,	/* indexed variable type. */
	VT_LAST_TYPE
} sct_var_type_t;

/* types of supported data constants. */
typedef enum {
	DT_BOOL = 0,
	DT_CHAR,
	DT_INT,
	DT_UINT,
	DT_SHORT,
	DT_USHORT,
	DT_LONG,
	DT_ULONG,
	DT_FLOAT,
	DT_STRING,
	DT_BUFFER, 
	DT_MULTI,
	DT_LAST_TYPE
} sct_data_type_t;

/* variable attribute names. */
typedef enum {
    VAN_DYN_TYPE = 0,
    VAN_LAST_TYPE
} sct_var_attr_name_t;

/* types of expression syntax/semantic errors. */
typedef enum {
	ERR_NONE = 0,
	ERR_INVALID_DATA_TYPE,
	ERR_INVALID_NODE,
	ERR_INVALID_OPERATOR,
	ERR_TYPE_MISMATCH,
	ERR_BUF_TOO_SMALL,
	ERR_INVALID_VAR
} sct_exp_error_t;

/* types of rules - 'before' or 'after'. */
typedef enum {
	RW_BEFORE = 0,
	RW_AFTER
} sct_rule_when_t;

/*
 * functions that are used to build a filtering expression.
 */

/* create primitive types - int, string, variable, indexed variable */
/* (e.g. parameter).                                                */
sct_filter_node_t* sct_make_bool_val(int num);
sct_filter_node_t* sct_make_int_val(int num);
sct_filter_node_t* sct_make_uint_val(unsigned int num);
sct_filter_node_t* sct_make_char_val(char c);
sct_filter_node_t* sct_make_short_val(short num);
sct_filter_node_t* sct_make_ushort_val(unsigned short num);
sct_filter_node_t* sct_make_long_val(long num);
sct_filter_node_t* sct_make_ulong_val(unsigned long num);
sct_filter_node_t* sct_make_float_val(float num);
/* str is copied into the node. */
sct_filter_node_t* sct_make_string_val(const char* str);
/* str is used in the node. */
sct_filter_node_t* sct_make_string_val_without_copy(char* str);
/* buffer is copied into the node */ 
sct_filter_node_t* sct_make_buffer_val(const void* buf, unsigned int sz); 
/* buffer is copied into the node */ 
sct_filter_node_t* sct_make_buffer_val_without_copy(void* buf, 
						    unsigned int sz);

/* make a filter node representing a non-indexed variable reference. */
sct_filter_node_t* sct_make_var_val(sct_var_type_t v, sct_data_type_t var_data_type);
/* make a filter node representing a non-indexed variable's field reference. */
sct_filter_node_t* sct_make_var_field_val(sct_var_type_t v,
				  int field_index,
				  sct_data_type_t var_data_type,
				  int struct_type);
/* make a filter node representing an indexed variable reference. */
sct_filter_node_t* sct_make_var_indexed_val(sct_var_type_t v,
				    int var_index,
				    sct_data_type_t var_data_type);
/* make a filter node representing an indexed variable's field reference. */
sct_filter_node_t* sct_make_var_indexed_field_val(sct_var_type_t,
					  int var_index,
					  int field_index,
					  sct_data_type_t field_data_type,
					  int struct_type);

/* create combined expressions using operators.                 */
/* checks for matching (e.g. can't comprate an int and an str). */
/* NOTE - parameters are "swalloed" by the generated node, and  */
/* will be freed when the generated node is freed.              */
sct_filter_node_t* sct_make_equal(sct_filter_node_t* f1, sct_filter_node_t* f2);
sct_filter_node_t* sct_make_less_then(sct_filter_node_t* f1, sct_filter_node_t* f2);
sct_filter_node_t* sct_make_less_or_equal(sct_filter_node_t* f1, sct_filter_node_t* f2);
sct_filter_node_t* sct_make_greater_then(sct_filter_node_t* f1, sct_filter_node_t* f2);
sct_filter_node_t* sct_make_greater_or_equal(sct_filter_node_t* f1, sct_filter_node_t* f2);
sct_filter_node_t* sct_make_not_equal(sct_filter_node_t* f1, sct_filter_node_t* f2);
sct_filter_node_t* sct_make_logical_and(sct_filter_node_t* f1, sct_filter_node_t* f2);
sct_filter_node_t* sct_make_logical_or(sct_filter_node_t* f1, sct_filter_node_t* f2);
sct_filter_node_t* sct_make_logical_not(sct_filter_node_t* f);
/* these would work only for numeric types. */
sct_filter_node_t* sct_make_bit_and(sct_filter_node_t* f1, sct_filter_node_t* f2);
sct_filter_node_t* sct_make_bit_or(sct_filter_node_t* f1, sct_filter_node_t* f2);
sct_filter_node_t* sct_make_bit_xor(sct_filter_node_t* f1, sct_filter_node_t* f2);
sct_filter_node_t* sct_make_bit_not(sct_filter_node_t* f);
sct_filter_node_t* sct_make_bit_left_shift(sct_filter_node_t* f1, sct_filter_node_t* f2);
sct_filter_node_t* sct_make_bit_right_shift(sct_filter_node_t* f1, sct_filter_node_t* f2);
sct_filter_node_t* sct_make_math_plus(sct_filter_node_t* f1, sct_filter_node_t* f2);
sct_filter_node_t* sct_make_math_minus(sct_filter_node_t* f1, sct_filter_node_t* f2);
sct_filter_node_t* sct_make_in_match(sct_filter_node_t* f1, sct_filter_node_t* f2);
/* these would work only for string or buffer types. */
sct_filter_node_t* sct_make_pattern_match(sct_filter_node_t* f1, sct_filter_node_t* f2);

/* vector handling functions. */
sct_filter_node_t* sct_make_filter_vector(int len);
/* inserts the given filter node into the given vector node. */
/* returns '1' on success, '0' if 'vec' is not a vector.     */
/* '-1' if index is out of range.                            */
int sct_insert_filter_node(sct_filter_node_t* vec, int index, sct_filter_node_t* f);
/* return the length of the given vector node. returns '-1'  */
/* if 'vec' is not a vector.				     */
int sct_get_filter_vector_len(sct_filter_node_t* vec);
/* return the filter node in the given index of the given vector. */
/* returns NULL if index is out of range, or vec is not a vector. */
sct_filter_node_t* sct_get_filter_node_at(sct_filter_node_t* vec, int index);

/* free the given filter node, as well as all sub-nodes and */
/* heap memory it contains, directly and indirectly.        */
void sct_free_filter_node(sct_filter_node_t* f);

/* serialize a filter node (and all its sub-nodes) into a */
/* buffer, for export purposes. returns the size of the   */
/* serialized data, 0 if there was an error, -1 if buffer */
/* is too small.                                          */
int sct_serialize_filter_node(sct_filter_node_t* f, unsigned char* buf, int buflen);

/* de-serialize a filter node (and all its sub-nodes)     */
/* from a buffer, for import purposes. returns the new    */
/* allocated filter node (the top of the tree), or NULL   */
/* if an error occured.                                   */
sct_filter_node_t* sct_deserialize_filter_node(unsigned char* buf, int buflen);

/* print a filter node (recursively) to a "standard" device */
void sct_print_filter_node(sct_filter_node_t* f);

/* evaluate the given filter node. */
/* since this is a generic function, it returns a */
/* filter node, containing a constant and the     */
/* proper result.                                 */
sct_exp_error_t sct_eval_filter_node(sct_filter_node_t* f, sct_filter_node_t* result);

/*
 * Functions for actions manipulation.
 */

/* allocate an action object for logging purposes. */
sct_action_t* sct_make_action_log(int priority, const char* log_format);

/* allocate an action object for rule enabling purposes. */
sct_action_t* sct_make_action_enable_rule(int rule_id);

/* allocate an action object for rule disabling purposes. */
sct_action_t* sct_make_action_disable_rule(int rule_id);

/* allocate an action object for process suspending. */
sct_action_t* sct_make_action_suspend_process(int pid);

/* allocate an action object for process killing. */
sct_action_t* sct_make_action_kill_process(int pid);

/* allocate an action object for system call failing. */
sct_action_t* sct_make_action_fail_syscall(int error_code);

/* set action variable attribute set command. */
void set_action_var_attr_set_command(sct_action_t* a,
				     sct_var_type_t var_type,
				     sct_var_attr_name_t name,
				     int val);
void set_action_var_indexed_attr_set_command(sct_action_t* a,
				 	     sct_var_type_t var_type,
				 	     int var_index,
				 	     sct_var_attr_name_t name,
				 	     int val);

/* free the action object */
void sct_free_action(sct_action_t* a);

/* serialize an action node into a buffer, for export     */
/* purposes. returns the size of the serialized data, 0   */
/* if there was an error, -1 if buffer is too small.      */
int sct_serialize_action_node(sct_action_t* a, unsigned char* buf, int buflen);

/* de-serialize an action node from a buffer, for import  */
/* purposes. returns the new allocated action node or     */
/* NULL if an error occuerd.                              */
sct_action_t* sct_deserialize_action_node(unsigned char* buf, int buflen);

/* get a string representation of this sct_action_t */
/* dont forget to free it */
char* sct_get_action_as_string(sct_action_t* a);

/* print an action node to a "standard" device */
void sct_print_action_node(sct_action_t* a);


/*
 * Functions for tracking rules manipulation.
 */

/* serialize a tracking rule into a buffer, for export   */
/* purposes. returns the size of the serialized data, 0  */
/* if there was an error, -1 if the buffer is too small. */
int sct_serialize_tracking_rule(int rule_id, sct_rule_when_t when,
			    int syscall_id,
			    sct_action_t* a, sct_filter_node_t* f,
		            unsigned char* buf, int buflen);

/* de-serialize a tracking rule from a buffer, for import */
/* purposes. returns the 1 for success, -1 for failure.   */
/* the given action rule and filter node are filled with  */
/* elements geenrated from the buffer, and should be      */
/* freed by the caller.                                   */
int sct_deserialize_tracking_rule(unsigned char* buf, int buflen, 
				  int* p_rule_id, sct_rule_when_t* p_when,
				  int* p_syscall_id,
				  sct_action_t** pa, sct_filter_node_t** pf);

/* serialize a list of tracking rules into a buffer,        */
/* for export purposes. returns the size of the serialized  */
/* data, 0 if there was an error, -1 if the buffer was      */
/* too small.                                               */ 
int sct_serialize_tracking_rule_list(sct_rule_list_t* rules, 
				     sct_membuf_t* mbuf); 

/* deserialize a tracking rule_list from a buffer, for import      */
/* purposes. returns 1 for success, -1 for failure. the rule list  */
/* pointed to by p_rules is filled with rules and should be freed  */
/* by the caller (with sct_delete_tracking_rule_list)              */
int sct_deserialize_tracking_rule_list(sct_membuf_t* mbuf, 
				       sct_rule_list_t** p_rules); 

/* create an empty tracking rules list. */
sct_rule_list_t* sct_create_tracking_rule_list(void);

/* delete the given tracking rule list, freeing any rule found in it. */
void sct_delete_tracking_rule_list(sct_rule_list_t* rules);

/* add a new tracking rule. returns the new rule's id */
/* returns '-1' if an error occured or if there's     */
/* already a rule with the given rule ID.             */
int sct_add_tracking_rule(sct_rule_list_t* rules,
		      int rule_id,
		      sct_filter_node_t* filter,
		      sct_action_t* action);

/* delete a tracking rule. */
void sct_del_tracking_rule(sct_rule_list_t* rules, int rule_id);

/* deletes all tracking rules in the list. */
void sct_del_all_tracking_rules_in_list(sct_rule_list_t* rules);

/* check if there is a rule with the given ID. */
/* returns '1' if there is, '0' otehrwise.     */
int sct_has_tracking_rule(sct_rule_list_t* rules, int rule_id);

/* returns the number of rules in the given rule list. */
int sct_get_tracking_rule_count(sct_rule_list_t* rules);

/* print out the list of current tracking rules. */
void sct_print_tracking_rules(sct_rule_list_t* rules);

/*
 * functions user to register logger callbacks.
 */

/* typedef for a logger callback. */
typedef int (*logger_callback_t)(char spec_char,
				 char* buf,	/* buf to emit data into. */
				 int buflen,
				 void* param,   /* the parameter to emit. */
				 void* param_data, /* extra parameter-context */
						   /* as given as a parameter */
                                                   /* to the logging function */
				 void* cb_data); /* callback data */

/* register the given logger calback function, that handles */
/* logging a given type of spec-char                        */
void set_logger_callback(char spec_char, logger_callback_t cb, void* cb_data);

/*
 * functions used by tracking functions to perform filtering.
 */

/* filter an incoming system call.                                          */
/* accepts a system call ID and its (var_args) parameters.                  */
/* returns a tracker_action structure if there was a match, NULL otherwise. */
/* on successfull match, the rule id is set into '*rule_id'.                */
sct_action_t* sct_filter_syscall(sct_rule_list_t* rules, int syscall_id, int* rule_id, ...);

/* when returning from a system call, should we use */
/* the original return value, or should we "inject" */
/* some other return value?                         */
typedef int sct_syscall_retval_policy;

#define SCT_USE_ORIG_RETVAL 0
#define SCT_OVERRIDE_RETVAL 1

/* perform an action in the current system call's context. */
/* accepts a system call ID and its (var_args) parameters. */
/* returns an 'sct_syscall_retval_policy'. if it's         */
/* SCT_OVERRIDE_RETVAL, the new retval to use will be put  */
/* in 'syscall_retval'. otherwise, it's left untouched.    */ 
sct_syscall_retval_policy
sct_tracker_perform_action(sct_action_t* action,
			   int rule_id, 
			   int syscall_id, 
			   int *syscall_retval, 
			   const char* syscall_name, 
			   const char* param_spec, 
			   ...);
	
/* utility functions for logging of data.     */
/* stringify a parameters list, using the     */
/* given param_spec, into the given buffer.   */
/* returns the stringified length on success, */
/* 0 on overflow.                             */
int 
sct_tracker_stringify_params(char* buf,
			     int buflen,
			     const char* param_spec,
			     ...);

/* 
 * Functions and macros for handling the syscall id
 */

/* get a string representation of this syscall id */ 
/* note that the buffer has to be AT LEAST 22 bytes long */ 
/* see the implementation if you're curious why */ 
void sct_get_syscall_id_as_str(int syscall_id, char* buf, unsigned int sz); 

/* extract native syscall id from a syscall id. */
#define	SYSCALL_PART_OF_SYSCALL_ID(syscall_id)	((syscall_id) & 0xffff)
/* extract multiplexed function id from a syscall id. */
#define	FUNC_PART_OF_SYSCALL_ID(syscall_id) (((syscall_id) >> 16) & 0xffff)
/* build from a func id and a syscall id the multiplexed syscall id */ 
#define MAKE_MULTIPLEX_SYSCALL_ID(sys, func) ((func << 16) | sys)

/* 
 * utility functions
 */ 
const char* sct_get_var_type_as_str(sct_var_type_t var); 

#ifdef __cplusplus 
}
#endif

#endif /* _RULE_ENGINE_PUBLIC_H */
