/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU General Public License                              */

/*
 * $Id: special_hijack.c,v 1.5 2002/06/17 23:30:34 mulix Exp $
 */

/* 
 * This file has the hijacking functions for special system calls, such 
 * as exit(), execve() or fork(). 
 */ 

#include "linux/sched.h"

/* exit, syscall 1 */ 
static asmlinkage long stub_exit(int error_code)
{
	typedef int (*stub_exit)(int error_code); 
	typedef int (*stub_before_exit)(int error_code); 
	stub_exit orig_syscall_func;
	stub_before_exit before_func;
	struct syscall_hijack_data* hdata = &hijack_data[1];

	down(&hdata->sem);
	hdata->invocation_ref_count++;
	up(&hdata->sem);
	before_func = (stub_before_exit)hdata->before_hijack_func;

	if (before_func)
		(*before_func)(error_code);

	orig_syscall_func = (stub_exit)hdata->orig_syscall;

	down(&hdata->sem);
	if (--hdata->invocation_ref_count == 0 && hdata->defered_release_flag == 1) {
	    /* we couldn't reduce the module's use count when the syscall  */
	    /* was originally released (as marked by the 'defered' flag) - */
	    /* so we do so now.                                            */
	    MOD_DEC_USE_COUNT;
	    hdata->defered_release_flag = 0;
	}
	up(&hdata->sem);
	/* WE HAVE A RACE FROM HERE ON - we're still executing on behalf of the stub_exit
	 * code, but no longer protected by the invocation_ref_count. Since the original 
	 * syscall never returns, I don't see a way to avoid this race (except for disallowing  
	 * unloading the module, which we already do)
	 */

	 /* finally, call the syscall itself */ 
	if (orig_syscall_func)
		(*orig_syscall_func)(error_code); 

	/* exit() must not return! */ 
	BUG(); 
	
	/* doesn't mean anything, silence gcc */ 
	return 0; 
}

/* stub function definitions for syscall 'fork' (2). */
/* my apologies in advance to the style police. */ 
#if defined __arch_um__
#define FORK_ARGS_DEF void
#define FORK_ARGS_CALL 
#elif defined __i386__
#define FORK_ARGS_DEF struct pt_regs regs
#define FORK_ARGS_CALL regs
#else
#error cannot compile fork for unknown arch
#endif /* fork multiplatform definitions */ 

asmlinkage static int
stub_syscall_2(FORK_ARGS_DEF) {
	typedef int
		(*stub_func_2)(FORK_ARGS_DEF);
	typedef int
		(*stub_before_func_2)(int* syscall_rc); 
	typedef int
		(*stub_after_func_2)(int syscall_rc);
	stub_func_2 orig_syscall_func;
	stub_before_func_2 before_func;
	stub_after_func_2 after_func;
	sct_syscall_retval_policy retval; /* after calling the before */
					  /* func, should we continue */
					  /* and call the original    */
					  /* syscall?                 */
	int fake_syscall_retval = 0;  /* the 'before' function will */
			              /* fill this variable for us  */
				      /* if it wants to 'fail' the  */
				      /* syscall                    */
	int orig_syscall_retval = 0;  /* the original syscall retval */

	struct syscall_hijack_data* hdata = &hijack_data[2];

	down(&hdata->sem);
	hdata->invocation_ref_count++;
	up(&hdata->sem);
	before_func = (stub_before_func_2)hdata->before_hijack_func;

	retval = (before_func
		    ? (*before_func)( &fake_syscall_retval)
		    : SCT_USE_ORIG_RETVAL);

	/* invoke the original syscall, if the hijack function returned 0. */
	if (retval == SCT_USE_ORIG_RETVAL) {
	    orig_syscall_func = (stub_func_2)hdata->orig_syscall;
	    orig_syscall_retval = (orig_syscall_func ? (*orig_syscall_func)(FORK_ARGS_CALL) : -ENOSYS);

	    after_func = (stub_after_func_2)hdata->after_hijack_func;
	    if (after_func)
		(*after_func)(orig_syscall_retval );
	}

	/* calling unconditional down() is never a good idea but we   */
	/* *must* grab this lock and lower the ref count, or we risk  */
	/* having it too high by one, leading to being unable to      */
	/* unload the module. also, none of the other holders of this */
	/* lock can block, so down() should be ok.                    */
	down(&hdata->sem);
	if (--hdata->invocation_ref_count == 0 &&
				hdata->defered_release_flag == 1) {
	    /* we couldn't reduce the module's use count when the syscall  */
	    /* was originally released (as marked by the 'defered' flag) - */
	    /* so we do so now.                                            */
	    MOD_DEC_USE_COUNT;
	    hdata->defered_release_flag = 0;
	}
	up(&hdata->sem);

	return (retval == SCT_USE_ORIG_RETVAL ? orig_syscall_retval : fake_syscall_retval);
}

void init_hijack_syscalls(void)
{
	stub_functions[1] = (func_ptr)stub_exit; 
 	stub_functions[2] = (func_ptr)stub_syscall_2;
}
