/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 guy keren, choo@actcom.co.il             */
/* License: GNU General Public License                              */

/*
 *$Id: syscall_hijack.c,v 1.19 2002/06/15 15:16:14 mulix Exp $
 */
#include <linux/module.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sys.h>
#include <linux/stddef.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <asm/semaphore.h>

#include "syscall_hijack_public.h"
#include "sct_compat.h" 

MODULE_AUTHOR("guy keren <choo@actcom.co.il>, "
	      "Muli Ben-Yehuda <mulix@actcom.co.il>");

MODULE_DESCRIPTION("System Call Hijacking and Releasing, "
		   "for 'syscalltrack'");

MODULE_LICENSE("GPL");

static int allow_unloading = 1; /* defaults to 1 while debugging */ 

MODULE_PARM(allow_unloading, "i"); 
MODULE_PARM_DESC(allow_unloading, "set to true to allow the module "
		 "to be unloaded"); 

/* type of info stored per syscall. */
struct syscall_hijack_data {
	func_ptr orig_syscall;	/* for syscall restore.                  */
	func_ptr before_hijack_func;/* hijacking function - before syscall.  */
	func_ptr after_hijack_func;	/* hijacking function - after syscall.   */
	struct semaphore sem;	/* per-syscall semaphore.                */
	int invocation_ref_count;	/* how many current active invocations?  */
	int defered_release_flag;	/* is there a pending (defered) release? */
};

/* hijack data for all syscalls. */
static struct syscall_hijack_data hijack_data[NR_syscalls];

/* when to invoke a hijack function - 'before' or 'after' syscall. */
typedef enum {
	HFW_BEFORE = 0,
	HFW_AFTER
} hijack_func_when_t;

extern func_ptr* sys_call_table[NR_syscalls];

/* unfortunately, we need to be aware of 'sct_syscall_retval_policy',  */
/* which is currently defined in sct_rules. we want to have it defined */
/* someplace neutral, so we dont have a dependancy on sct_rules. then  */
/* again, this would require reworking the entire stub mechanism...    */
#include "rule_engine_public.h"

/* ok, include the autogenerated stuff... */
#include "syscall_hijack_autogen.c"
#include "special_hijack.c"

/* hijack the given system call. returns '0' on success, */
/* '1' if the syscall is already hijacked, and           */
/* a negative int on error.                              */
int hijack_syscall(int syscall_id,
		   func_ptr function,
		   hijack_func_when_t when)
{
	int retval = -ENOSYS;
	int rc;
	int defered_flag = 0;
	struct syscall_hijack_data* hdata = NULL;

	if (syscall_id < 0 || syscall_id >= NR_syscalls) {
		printk (KERN_NOTICE "hijack_syscall: invalid syscall_id '%d'\n",
			syscall_id);
		return retval;
	}
    
	hdata = &hijack_data[syscall_id];

	printk (KERN_DEBUG "hijack_syscall: %d: before down_interruptible\n",
		syscall_id);

	rc = down_interruptible(&hdata->sem);
	if (rc < 0){
		return -ERESTARTSYS;
	}

	/* remove the 'defered release flag' if it was set,   */
	/* but remember it was set, for later use.            */
	/* this flag might be set here if we marked this call */
	/* for defered release, and before the release could  */
	/* happen, we are re-hijacking it.                    */
	if (hdata->defered_release_flag) {
		defered_flag = 1;
		hdata->defered_release_flag = 0;
	}

	/* if the syscall has active invocations, and a hijck function */
	/* is still registered, then we cannot hijack it again - have  */
	/* to wait until its unregistered, or until it has no active   */
	/* invocations.                                                */
	if (hdata->invocation_ref_count > 0 && hdata->orig_syscall) {
		printk (KERN_NOTICE "hijack_syscall: syscall %d busy\n",
			syscall_id);
		retval = 1;
		goto hijack_syscall_release_lock;
	} 
    
	printk (KERN_INFO "hijack_syscall: beginning %d hijack\n", syscall_id);
	if (!stub_functions[syscall_id]) {
		printk (KERN_NOTICE "hijack_syscall: %d: unsupported syscall\n",
			syscall_id);
		retval = -ENOENT;
		goto hijack_syscall_release_lock;
	}
    
	if (!hdata->orig_syscall) { /* only hijack if we dont have it already hijacked */
		func_ptr orig_syscall;

		func_ptr* sys_call_table_addr = (func_ptr*)
			get_module_symbol(NULL, versioned_string(sys_call_table)); 
		if (!sys_call_table_addr) {
			printk (KERN_NOTICE "hijack_syscall: %d: couldn't locate "
				"sys_call_table\n", syscall_id);
			retval = -ENOENT;
			goto hijack_syscall_release_lock;
		}
	
		orig_syscall = sys_call_table_addr[syscall_id];
	    
		/* if we had a defered release for this syscall earlier, */
		/* we didn't DEC the use count of the module, and thus   */
		/* shouldn't INC it now.                                 */
		if (!defered_flag) {
			MOD_INC_USE_COUNT;
		}
	
		/* put into the syscall table the stub function */
		sys_call_table_addr[syscall_id] =
			(func_ptr)stub_functions[syscall_id];
	
		/* put into the hijacking functions table the user */
		/* supplied function.				   */
		if (when == HFW_BEFORE)
			hdata->before_hijack_func = function;
		else
			hdata->after_hijack_func = function;
	
		/* put the original syscall into the orig syscalls table */
		hdata->orig_syscall = orig_syscall;
	
		retval = 0;
	}
	else if (when == HFW_BEFORE && !hdata->before_hijack_func) {
		hdata->before_hijack_func = function;
		retval = 0; 
	}
	else if (when == HFW_AFTER && !hdata->after_hijack_func) {
		hdata->after_hijack_func = function;
		retval = 0;
	}
	else { 
		printk (KERN_NOTICE "hijack_syscall: %d: already hijacked for %s\n",
			syscall_id, when == HFW_BEFORE ? "'before'" : "'after'");
		retval = 1;
	}

 hijack_syscall_release_lock:

	printk (KERN_DEBUG "hijack_syscall: %d: before up\n", syscall_id);
	up(&hdata->sem); 
	printk (KERN_DEBUG "hijack_syscall: %d: after up\n", syscall_id);

	return retval;
}

int hijack_syscall_before(int syscall_id,
			  func_ptr function)
{
	return hijack_syscall(syscall_id, function, HFW_BEFORE);
}

int hijack_syscall_after(int syscall_id,
			 func_ptr function)
{
	return hijack_syscall(syscall_id, function, HFW_AFTER);
}


/* releases a previously hijacked system call. */
int release_syscall(int syscall_id, hijack_func_when_t when)
{
	int retval = -ENOSYS;

	struct syscall_hijack_data* hdata = NULL;

	if (syscall_id < 0 || syscall_id >= NR_syscalls) {
		printk (KERN_NOTICE "release_syscall: invalid syscall_id '%d'\n",
			syscall_id);
		return retval;
	}

	hdata = &hijack_data[syscall_id];

	printk (KERN_DEBUG "release_syscall: %d: before down_interruptible\n",
		syscall_id);
	retval = down_interruptible(&hdata->sem);
	if (retval < 0)
		return -ERESTARTSYS;
    
	if (hdata->orig_syscall) {
		/* if we're asked to remove a 'before' function, but still got */
		/* an 'after' function, or vice versa, then just release that  */
		/* function, but keep the syscall itself hijacked.             */
		if (hdata->before_hijack_func && hdata->after_hijack_func) {
			if (when == HFW_BEFORE) {
				hdata->before_hijack_func = (func_ptr)NULL;
			}
			else if (when == HFW_AFTER) {
				hdata->after_hijack_func = (func_ptr)NULL;
			}
		} /* ok, only one of them is stil hijacked */
		else {
			func_ptr* sys_call_table_addr = (func_ptr*)
				get_module_symbol(NULL, versioned_string(sys_call_table));		
			if (sys_call_table_addr) {
				func_ptr orig_syscall = hdata->orig_syscall;
		
				sys_call_table_addr[syscall_id] = orig_syscall;
				hdata->orig_syscall = (func_ptr)NULL;
				hdata->before_hijack_func = (func_ptr)NULL;
				hdata->after_hijack_func = (func_ptr)NULL;
				if (hdata->invocation_ref_count > 0) {
					printk (KERN_DEBUG "release_syscall: %d: syscall busy "
						"(%d active invocations) - marking for defered "
						"release\n",
						syscall_id, hdata->invocation_ref_count);
					/* mark the 'defered release flag' for this syscall. */
					hdata->defered_release_flag = 1;
				}
				else {
					MOD_DEC_USE_COUNT;
				} 
			}
		}
	}
	printk (KERN_DEBUG "release_syscall: %d: before up\n", syscall_id);
	up(&hdata->sem); 
	printk (KERN_DEBUG "release_syscall: %d: after up\n", syscall_id);
    
	return retval;
}

int release_syscall_before(int syscall_id)
{
	return release_syscall(syscall_id, HFW_BEFORE);
}

int release_syscall_after(int syscall_id)
{
	return release_syscall(syscall_id, HFW_AFTER);
}

int
init_module(void)
{
        int i;

	memset(hijack_data, 0, sizeof(hijack_data));
	
        for (i=0; i<NR_syscalls; i++) {
		init_MUTEX(&hijack_data[i].sem);
        }

	init_hijack_syscalls(); 

	/* only allow unloading this module if the user explicitly said so */
	if (!allow_unloading)
	    MOD_INC_USE_COUNT;

        return 0;
}

void
clean_module(void)
{
        /* do nothing - this module cannot be unloaded. */
        printk(KERN_NOTICE "Error - syscall_hijack module should NOT be unloaded!!");
}


EXPORT_SYMBOL(hijack_syscall_before);
EXPORT_SYMBOL(hijack_syscall_after);
EXPORT_SYMBOL(release_syscall_before); 
EXPORT_SYMBOL(release_syscall_after); 
