/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 guy keren, choo@actcom.co.il             */
/* License: GNU General Public License                              */

/*
 * $Id: tracker_sysctl.c,v 1.7 2002/06/15 15:16:14 mulix Exp $
 */ 

#include <linux/kernel.h>
#include <linux/errno.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>

#include "rule_engine_public.h"
#include "sct_rules_private.h"
#include "tracker_sysctl.h"
#include "update_rules.h"
#include "membuf.h"

/* adds a new rule for a given system call. */
int tracker_sysctl_add_rule(int rule_id,
			    sct_rule_when_t when,
			    int syscall_id,
			    int mux_func_id,
			    sct_filter_node_t* f,
			    sct_action_t* a)
{
	return tracker_add_rule(rule_id, when, syscall_id, 
				mux_func_id, f, a);
}

/* deletes an existing rule for a given system call. */
int tracker_sysctl_del_rule(int rule_id,
			    sct_rule_when_t when,
			    int syscall_id,
			    int mux_func_id)
{
	return tracker_del_rule(rule_id, when, syscall_id, mux_func_id);
}

/* deletes all existing rules for a given system call. */
/* TODO - handle muxed functions as if they were real syscalls. */
int tracker_sysctl_del_all_rules_for_syscall(int syscall_id, int mux_func_id)
{
	return tracker_del_all_rules_for_syscall(syscall_id, mux_func_id);
}

/* deletes all existing rules in the module */
int tracker_sysctl_del_all_module_rules(void)
{
	return tracker_del_all_module_rules();
}

/* prints all existing rules in the module */
int tracker_sysctl_print_all_rules(void)
{
	return tracker_print_all_rules();
}

int tracker_sysctl_get_all_rule_count(void* oldval, size_t* oldlenp)
{
	size_t sz; 
	sct_membuf_t* mbuf; 
	int retval;
	int rc; 

	printk(KERN_DEBUG "sysctl_get_all_rule_count starting\n"); 
	if (copy_from_user(&sz, oldlenp, sizeof(sz)))
		return -EFAULT; 
	
	printk(KERN_DEBUG "the uspace buffer is size %d\n", sz); 
	
	printk(KERN_DEBUG "sysctl_get_all_rule_count making membuf\n"); 
	if (!(mbuf = sct_membuf_create(sz, NULL, cp_to_user)))
		return -ENOMEM; 

	printk(KERN_DEBUG "sysctl_get_all_rule_count gettting rule count\n"); 
	if ((retval = tracker_get_all_rule_count(mbuf)) < 0)
		goto free_membuf; 

	rc = sct_membuf_get_chunk(mbuf, oldval, sz); 
	printk(KERN_DEBUG "sysctl_get_all_rule_count - get_chunk returned %d\n", retval); 
	if (rc == sz) /* we got the entire chunk we were expecting*/ 
		retval = 0; /* success */ 
	else if ((rc >= 0) && (rc < sz)) /* the membuf had less than we were expecting */ 
		retval = -ENOSPC; 
	else 
		retval = -EINVAL; 
	
 free_membuf:
	sct_membuf_destroy(mbuf); 
	
	printk(KERN_DEBUG "sysctl_get_all_rule_count returning %d\n", retval); 
	return retval; 
}

int tracker_sysctl_get_all_rules(void* oldval, size_t* oldlenp)
{
	size_t ulen; 
	size_t len; 
	int retval; 
	sct_membuf_t* mbuf; 
	if (copy_from_user(&ulen, oldlenp, sizeof(ulen))) {
		printk(KERN_NOTICE "tracker_get_all_rules: failed to get "
		       "the buffer length from uspace\n"); 
		return -EFAULT; 
	}
	
	if (!(mbuf = sct_membuf_create(ulen, NULL, cp_to_user))) {
		printk(KERN_NOTICE "tracker_get_all_rules: failed to allocate %d "
		       "bytes for the membuf\n", len); 
		return -ENOMEM; 
	}
	if ((retval = tracker_get_all_rules(mbuf)) < 0) {
		printk(KERN_DEBUG "tracker_get-all_rules: failed call to "
		       "tracker_get_all_rules, retval %d\n", retval); 
		goto free_mbuf; 
	}

	/* copy them to user space */ 
	len = sct_membuf_get_serialized_size(mbuf); 
	
	/* update the chunk's len, so uspace will  know */ 
	if (copy_to_user(oldlenp, &len, sizeof(len))) {
		retval = -EFAULT; 
		printk(KERN_DEBUG "tracker_get_all_rules: failed to copy "
		       "the rule lists to uspace\n"); 
		goto free_mbuf; 
	}

	if ((sct_membuf_get_chunk(mbuf, oldval, len) != len)) {
		retval = -ENOSPC;  
		printk(KERN_DEBUG "tracker_get_all_rules: failed to copy "
		       "the rule lists to uspace\n"); 
		goto free_mbuf; 
	}
	retval = 0; 

 free_mbuf:
	sct_membuf_destroy(mbuf); 
	return retval; 
}
