/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 guy keren, choo@actcom.co.il             */
/* License: GNU General Public License                              */

/*
 * $Id: update_rules.c,v 1.15 2002/06/15 15:16:14 mulix Exp $
 */ 

#define __NO_VERSION__
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>

#include "rule_engine_public.h"
#include "sct_rules_private.h"
#include "membuf.h"
#include "module_interface.h"
#include "update_rules.h"

DECLARE_MUTEX(tracker_sem);

/* general rule handling functions */

/* build a rule from this buffer. */
/* return negative error on error, 0 for success */
int tracker_build_rule(unsigned char* buf, int len, int* rule_id, 
		       sct_rule_when_t* when, int* syscall_id, 
		       int* mux_func_id, sct_action_t** act, 
		       sct_filter_node_t** f)
{
	int rc = sct_deserialize_tracking_rule(buf, len, rule_id, when,
					       syscall_id, act, f);
	if (rc <= 0) {
		printk(KERN_NOTICE "tracker_build_rule: failed to "
		       "deserialize tracking rule, err '%d'\n", rc);
		return -EINVAL; /* todo: anything better? */
	}

	*mux_func_id = FUNC_PART_OF_SYSCALL_ID(*syscall_id);
	*syscall_id = SYSCALL_PART_OF_SYSCALL_ID(*syscall_id);
    
	return rc;
}

/* sanity checks */
int sanity_check_user_input(int syscall_id, int mux_func_id, 
			    sct_rule_when_t when)
{
	int mux_id = mux_data_index(syscall_id);
    
	if (syscall_id < 0 || syscall_id > NR_syscalls) {
		printk(KERN_NOTICE "sanity_check_user_input: syscall id "
		       "'%d; out of range\n", syscall_id);
		return -EINVAL;
	}
	if (mux_func_id != 0) { /* make these checks only for mux functions */

		if ((mux_func_id < 0) || /* there cannot be such a mux function */
		    (mux_id < 0) || /* or this syscall is not a mux syscall */
		    /* or it is a mux syscall, but without this mux function */
		    (mux_func_id >= mux_data[mux_id].num_muxed)) { 
			printk(KERN_NOTICE "syscall id %d , mux function id %d out of range\n",
			       syscall_id, mux_func_id);
			return -EINVAL;
		}
		if (mux_func_id > 0 &&
		    ((when == RW_BEFORE && !mux_data[mux_id].before_rules[mux_func_id]) ||
		     (when == RW_AFTER && !mux_data[mux_id].after_rules[mux_func_id]))) {
			printk(KERN_NOTICE "syscall id %d , mux function id %d not supported\n",
			       syscall_id, mux_func_id);
			return -EINVAL;
		}
	}
	if (when != RW_BEFORE && when != RW_AFTER) {
		printk(KERN_NOTICE "sanity_check_user_input: when='%d'\n", when);
		return -EINVAL;
	}
	
	return 0; /* success */
}



/* utility functions */

static inline int
syscall_get_rule_count_when(int syscall_id, sct_rule_when_t when)
{
	int count = 0;
	int id, i; 

	/* special handling for mux syscalls. */
	if ((id = mux_data_index(syscall_id)) >= 0) {
		struct mux_data_t* data = &mux_data[id];
		for (i = 0; i < data->num_muxed; ++i) {
			count += ((when == RW_BEFORE) ?
				  sct_get_tracking_rule_count(data->before_rules[i]) :
				  sct_get_tracking_rule_count(data->after_rules[i]));
		}
	} else { /* regular syscalls */
		count = ((when == RW_BEFORE) ?
			 sct_get_tracking_rule_count(track_data[syscall_id].before_rules) :
			 sct_get_tracking_rule_count(track_data[syscall_id].after_rules));
	}
	return count;
}

static inline int 
syscall_is_hijacked_when(int syscall_id, sct_rule_when_t when)
{
	return (syscall_get_rule_count_when(syscall_id, when) > 0);
}


static inline int 
syscall_is_hijacked(int syscall_id)
{
	return (syscall_is_hijacked_when(syscall_id, RW_BEFORE) ||
		syscall_is_hijacked_when(syscall_id, RW_AFTER));
}

/* which rule list will we be working on? this depends on whether this   */
/* is a mux function or a regular function, and whether this is a before */
/* or after rule                                                         */
static inline sct_rule_list_t*
syscall_get_rule_list(int syscall_id, int mux_func_id, sct_rule_when_t when)
{
	sct_rule_list_t* p_rules = NULL;   /* assume failure. */
	int mux_idx;

	if (mux_func_id == 0) { /* regular function, not muxed */
		p_rules = ((when == RW_BEFORE) ? 
			   track_data[syscall_id].before_rules :
			   track_data[syscall_id].after_rules);
	} else {
		if ((mux_idx = mux_data_index(syscall_id)) >= 0) {
			p_rules = ((when == RW_BEFORE) ? 
				   mux_data[mux_idx].before_rules[mux_func_id] :
				   mux_data[mux_idx].after_rules[mux_func_id]);
		}
	}

	return p_rules;
}



/* adds a new rule for a given system call. */
int tracker_add_rule(int rule_id,
		     sct_rule_when_t when,
		     int syscall_id,
		     int mux_func_id,
		     sct_filter_node_t* f,
		     sct_action_t* a)
{
	int rc = 0;
	int retval = 0;
	sct_rule_list_t* p_rules;
	const char* when_str = (when == RW_BEFORE ? "before" : "after");
	int when_not = (when == RW_BEFORE ? RW_AFTER : RW_BEFORE);
    
	printk(KERN_NOTICE "adding tracking rule %d (%s) for syscall %d, mux_func_id %d\n",
	       rule_id, when_str, syscall_id, mux_func_id);
	/* lock to avoid races with other invocations. */
	if (down_interruptible(&tracker_sem)) {
		retval = -EINTR;
		goto err_add_rule_lock;
	}
	/* which rule list will we be working on? this depends on whether this is 
	   a mux function or a regular function, and whether this is a before
	   or after rule */
	p_rules = syscall_get_rule_list(syscall_id, mux_func_id, when);
	if (!p_rules) {
		retval = -ENOENT; 
		goto err_add_rule_no_rule_list;
	}

	rc = sct_add_tracking_rule(p_rules, rule_id, f, a);
	if (rc < 0) {
		printk (KERN_NOTICE "add tracking_rule returned '%d'\n", rc);
		retval = -ENOMEM; /* todo: something better? */
		goto err_add_rule_add_rule;
	}

	/* make sure we don't release 'f', since it was consumed */
	/* by the 'sct_add_tracking_rule' function.                  */
	f = NULL;
	a = NULL;
	sct_print_tracking_rules(p_rules);
    
	/* if it's a first rule, need to hijack */
	if (syscall_get_rule_count_when(syscall_id, when) == 1) {
		printk(KERN_NOTICE "first '%s' rule for syscall - hijacking syscall\n",
		       when_str);
		if (when == RW_BEFORE)
			rc = hijack_syscall_before(
				syscall_id,
				(func_ptr)before_track_functions[syscall_id]);
		else
			rc = hijack_syscall_after(
			        syscall_id,
				(func_ptr)after_track_functions[syscall_id]);
		printk(KERN_NOTICE "hijack_syscall_%s returned '%d'\n", when_str, rc);
		if (rc < 0) {
			printk(KERN_NOTICE "hijack_syscall_%s failed, error '%d'\n",
			       when_str, rc);
			retval = (rc != -ERESTARTSYS ? rc : -EINTR);
			/* remove rule */
			sct_del_tracking_rule(p_rules, rule_id);
			goto err_add_rule_hijack;
		}
		/* if it's the first hijack */
		if (!syscall_is_hijacked_when(syscall_id, when_not)) {
			MOD_INC_USE_COUNT;
		}
	}

 err_add_rule_hijack:
 err_add_rule_add_rule:
 err_add_rule_no_rule_list:
	up(&tracker_sem);
 err_add_rule_lock:
	if (f)
		sct_free_filter_node(f);
	if (a)
		sct_free_action(a);
    
	return retval;
}

/* deletes an existing rule for a given system call. */
int
tracker_del_rule(int rule_id,
		 sct_rule_when_t when,
		 int syscall_id,
		 int mux_func_id)
{
	int rc;
	int retval = 0;
	sct_rule_list_t* p_rules;
	const char* when_str = (when == RW_BEFORE ? "before" : "after");
	int when_not = (when == RW_BEFORE ? RW_AFTER : RW_BEFORE);

	printk(KERN_NOTICE "tracker_sysctl: deleting tracking rule %d (%s) for syscall %d\n",
	       rule_id, when_str, syscall_id);

	/* lock to avoid races with other invocations. */
	if (down_interruptible(&tracker_sem)) {
		retval = -ERESTARTSYS;
		goto err_del_rule_lock;
	}

	p_rules = syscall_get_rule_list(syscall_id, mux_func_id, when);
	if (!p_rules)
		goto err_del_rule_no_rule_list;

	rc = sct_has_tracking_rule(p_rules, rule_id);
	if (!rc) {
		printk(KERN_NOTICE "tracker_sysctl: rule '%d' not found.\n", rule_id);
		goto err_del_rule_no_such_rule;
	}
    
	sct_del_tracking_rule(p_rules, rule_id);
	sct_print_tracking_rules(p_rules);

	/* if it's a the last rule, need to release */
	if (!syscall_is_hijacked_when(syscall_id, when)) {
		printk(KERN_NOTICE "last '%s' rule for syscall - releasing syscall\n",
		       when_str);
		if (when == RW_BEFORE)
			rc = release_syscall_before(syscall_id);
		else
			rc = release_syscall_after(syscall_id);
		if (rc < 0) {
			printk(KERN_NOTICE "release_syscall_%s failed, error '%d'\n",
			       when_str, rc);
			retval = (rc != -ERESTARTSYS ? rc : -EINTR);
			goto err_del_rule_release;
		}
		if (!syscall_is_hijacked_when(syscall_id, when_not)) {
			MOD_DEC_USE_COUNT;
		}
	}

 err_del_rule_release:
 err_del_rule_no_such_rule:
 err_del_rule_no_rule_list:
	up(&tracker_sem);
 err_del_rule_lock:
    
	return retval;
}

/* utility function to delete all rules for a syscall */ 
/* callers must lock! */ 
static int
del_syscall_rules(int syscall_id, int mux_func_id)
{
	int rc1, rc2; 
	int retval = 0; 
	sct_rule_list_t* list; 

	rc1 = rc2 = 0; 

	/* get the before rule list, taking into account multiplexing */ 
	list = syscall_get_rule_list(syscall_id, mux_func_id, RW_BEFORE); 
	if (list) {
		sct_del_all_tracking_rules_in_list(list); 
		sct_print_tracking_rules(list); 		
		rc1 = release_syscall_before(syscall_id);
	}
	list = syscall_get_rule_list(syscall_id, mux_func_id, RW_AFTER); 
	if (list){
		sct_del_all_tracking_rules_in_list(list);
		sct_print_tracking_rules(list);
		rc2 = release_syscall_after(syscall_id);
	}

	if (rc1 < 0) {
		printk(KERN_NOTICE "del_syscall_rules: release_syscall_before failed, error '%d'\n", rc1); 
		retval = (rc1 != -ERESTARTSYS ? rc1 : -EINTR);
	}
	if (rc2 < 0) {
		printk(KERN_NOTICE "tracker_sysctl: release_syscall_after failed, error '%d'\n", rc2); 
		if(!retval) 
			retval = (rc2 != -ERESTARTSYS ? rc2 : -EINTR);
	}

	return retval; 
}

/* deletes all existing rules for a given system call. */
/* TODO - handle muxed functions as if they were real syscalls. */
int
tracker_del_all_rules_for_syscall(int syscall_id, int mux_func_id)
{
	int retval = 0;
	
	printk(KERN_NOTICE "tracker_sysctl: deleting all tracking rules for syscall %d\n",
	       syscall_id);

	/* lock to avoid races with other invocations. */
	if (down_interruptible(&tracker_sem)) {
		retval = -ERESTARTSYS;
		goto ret_nolock; 
	}

	if (!syscall_is_hijacked(syscall_id))
		goto done; 
	
	retval = del_syscall_rules(syscall_id, mux_func_id); 

	/* the system calls could still be hijacked, if it's a mux syscall */ 
	if (!syscall_is_hijacked(syscall_id))
		MOD_DEC_USE_COUNT; 

 done:
	up(&tracker_sem);

 ret_nolock:
	return retval;
}

/* deletes all existing rules in the module */
int
tracker_del_all_module_rules(void)
{
	unsigned int i;
	int j;
	int rc; 
	int retval = 0;
	int mux_idx; 

	printk(KERN_NOTICE "tracker_sysctl: deleting all module tracking rules\n");

	/* lock to avoid races with other invocations. */
#if 0
	printk(KERN_DEBUG "locking tracker_sem\n"); 
#endif /* 0 */ 
	if (down_interruptible(&tracker_sem)) {
		retval = -ERESTARTSYS;
		goto ret_nolock; 
	}

	/* iterate over all syscalls and release the hijacked ones */
	for (i = 0; i < NR_syscalls; ++i){
		/* not hijacked? go on */
		if (!syscall_is_hijacked(i))
			continue;

		/* handle muxed syscalls. */
		if ((mux_idx = mux_data_index(i)) >= 0) {
			struct mux_data_t* data = &mux_data[mux_idx];
			if (data->num_muxed > 0 && data->before_rules) {
				for (j=0; j < data->num_muxed; j++) {
					rc = del_syscall_rules(i, j); 
					if (rc) {
						retval = rc; 
						goto done; 
					}
				}
			}
			if (data->num_muxed > 0 && data->after_rules) {
				for (j=0; j < data->num_muxed; j++) {
					rc = del_syscall_rules(i, j); 
					if (rc) {
						retval = rc; 
						goto done; 
					}
				}
			}
		}
		/* and now the syscall itself */ 
		rc = del_syscall_rules(i, 0);
		
		/* if it's still hijacked now, it's a bug */ 
		if (syscall_is_hijacked(i))
			BUG(); 
		   
		MOD_DEC_USE_COUNT; 
	}
 done:
	up(&tracker_sem);
	
 ret_nolock:
	return retval;
}

/* prints all existing rules in the module */
int tracker_print_all_rules(void)
{
	unsigned int i;
	int retval = 0;
	int printed = 0;
    
	/* lock to avoid races with other invocations. */
	if (down_interruptible(&tracker_sem)) {
		retval = -ERESTARTSYS;
		goto err_print_module_rules_lock;
	}
    
	printk (KERN_NOTICE "**************************************\n");
	printk (KERN_NOTICE "syscalltrack - printing defined rules\n");
	for (i = 0; i < NR_syscalls; ++i)
	{
		if (syscall_is_hijacked(i)){
			int got_any_before = 0;
			int mux_idx;

			if (track_data[i].before_rules) {
				printk(KERN_NOTICE "rules for syscall '%d':\n", i);
				printk(KERN_NOTICE " 'before' rules:\n");
				sct_print_tracking_rules(track_data[i].before_rules);
				printed = 1;
				got_any_before = 1;
			}
			if (track_data[i].after_rules) {
				if (!got_any_before)
					printk(KERN_NOTICE "rules for syscall '%d':\n", i);
				printk(KERN_NOTICE " 'after' rules:\n");
				sct_print_tracking_rules(track_data[i].after_rules);
				printed = 1;
			}
			/* handled muxed syscalls. */
			if ((mux_idx = mux_data_index(i)) >= 0) {
				int got_any_before = 0;
				struct mux_data_t* data = &mux_data[mux_idx];

				if (data->before_rules) {
					int j;
					for (j=0; j<data->num_muxed; j++) {
						if (data->before_rules[j]) {
							printk(KERN_NOTICE "rules for syscall '%d:%d':\n", i,j);
							printk(KERN_NOTICE " 'before' rules:\n");
							sct_print_tracking_rules(data->before_rules[j]);
							printed = 1;
							got_any_before = 1;
						}
					}
				}
				if (data->after_rules) {
					int j;
					for (j=0; j<data->num_muxed; j++) {
						if (data->after_rules[j]) {
							if (!got_any_before)
								printk(KERN_NOTICE "rules for syscall '%d:%d':\n", i,j);
							printk(KERN_NOTICE " 'after' rules:\n");
							sct_print_tracking_rules(data->after_rules[j]);
							printed = 1;
						}
					}
				}
			}
		}
	}
	if (!printed){
		printk (KERN_NOTICE "no tracking rules registered currently\n");
	}
	printk (KERN_NOTICE "**************************************\n");
    
	up(&tracker_sem);
    
 err_print_module_rules_lock:
	return retval;
}

int tracker_get_all_rule_count(sct_membuf_t* mbuf)
{
	int i; 
	int retval = 0; 
	int rc; 
	int sum = 0; 

	retval = -ERESTARTSYS;
	/* lock to avoid races with other invocations. */
	if (down_interruptible(&tracker_sem))
		goto out_nolock;

	for (i = 0; i < NR_syscalls; ++i)
	{
		sum += syscall_get_rule_count_when(i, RW_BEFORE); 
		sum += syscall_get_rule_count_when(i, RW_AFTER); 
	}
#if 0
	printk(KERN_DEBUG "get_all_rule_count: sum is %d\n", sum); 
#endif /* 0 */ 
	up(&tracker_sem); 

	rc = sct_membuf_add_chunk(mbuf, &sum, sizeof(sum)); 
#if 0
	printk(KERN_DEBUG "membuf add chunk returned %d, data is now %d\n", 
	       rc, *(int*)sct_membuf_get_data(mbuf)); 
#endif /* 0 */ 
	if (rc != sizeof(sum)) {
		if (rc < 0) 
			return -EINVAL; 

		/* mbuf wasn't big enough */ 
		return -ENOSPC; 
	}
	retval = 0; 

 out_nolock:
	return retval; 
}

/* return -1 on buf too small, 0 on error, 1 on success */ 
static int ser_rule_list(sct_membuf_t* mbuf, int syscall_id, 
			 sct_rule_when_t when, sct_rule_list_t* list)
{
	int rc;
	int magic = GET_RULES_MAGIC_COOKIE; 

	if ((rc = sct_membuf_add_chunk(mbuf, &magic, 
				       sizeof(magic))) < 0) {
		printk(KERN_DEBUG "adding magic for %d returned %d\n", 
		       syscall_id, rc); 
		return -EINVAL; 
	}

	if ((rc = sct_membuf_add_chunk(mbuf, &syscall_id, 
				       sizeof(syscall_id))) < 0) {
		printk(KERN_DEBUG "adding syscall id for %d returned %d\n", 
		       syscall_id, rc); 
		return -EINVAL; 
	}

	if ((rc = sct_membuf_add_chunk(mbuf, &when, sizeof(when))) < 0) {
		printk(KERN_DEBUG "adding when for %d returned %d\n", 
		       syscall_id, rc); 
		return -EINVAL; 
	}
	
	if ((rc = sct_serialize_tracking_rule_list(list, mbuf)) <= 0) {
		printk(KERN_DEBUG "adding rule list for %d returned %d\n", 
		       syscall_id, rc); 
		return -EINVAL; 
	}

	return 1; 	
}

int tracker_get_all_rules(sct_membuf_t* mbuf)
{
	int retval, rc; 
	int i; 
	unsigned int marker = 0; 
	
	/* lock the tracker semaphore to prevent concurrent modification */ 
	retval = -ERESTARTSYS;
	if (down_interruptible(&tracker_sem)) {
		goto err_nolock; 
	}
	
#if 0 
	printk(KERN_DEBUG "tracker_get_all_rules: starting to serialize\n"); 
#endif /* 0 */ 
	/* for each of our syscalls, traverse over all of its rules */ 
	for (i = 0; i < NR_syscalls; ++i)
	{
		int mux_idx;
		
		if (!syscall_is_hijacked(i))
			continue; 

#if 0
		printk(KERN_DEBUG "working on syscall %d\n", i); 
#endif /* 0 */ 
		
		if (track_data[i].before_rules) {
			if ((rc = ser_rule_list(mbuf, i, RW_BEFORE, track_data[i].before_rules)) <= 0) {
				printk(KERN_DEBUG "ser_rule_list(%d) for before rules returned %d\n", 
				       i, rc); 
				retval = (rc == -1 ? -ENOSPC : -EINVAL); 
				goto err_release_lock; 
			}
		}
		if (track_data[i].after_rules) {
			if ((rc = ser_rule_list(mbuf, i, RW_AFTER, track_data[i].after_rules)) <= 0) {
				printk(KERN_DEBUG "ser_rule_list(%d) for after rules returned %d\n", 
				       i, rc); 
				retval = (rc == -1 ? -ENOSPC : -EINVAL); 
				goto err_release_lock; 
			}
		}
		/* handled muxed syscalls. */
		if ((mux_idx = mux_data_index(i)) >= 0) {
			struct mux_data_t* data = &mux_data[mux_idx];
#if 0 
			printk(KERN_DEBUG "serializing rules for mux syscall %d\n", i); 
#endif /* 0 */ 
			if (data->before_rules) {
				int j;
				for (j=0; j<data->num_muxed; j++) {
					if (!data->before_rules[j])
						continue; 
					if ((rc = ser_rule_list(mbuf, MAKE_MULTIPLEX_SYSCALL_ID(i, j), 
								RW_BEFORE, data->before_rules[j])) <= 0) {
						retval = (rc == -1 ? -ENOSPC : -EINVAL); 
						goto err_release_lock; 
					}
				}
			}
			if (data->after_rules) {
				int j;
				for (j=0; j<data->num_muxed; j++) {
					if (!data->after_rules[j])
						continue; 
					if ((rc = ser_rule_list(mbuf, MAKE_MULTIPLEX_SYSCALL_ID(i, j), 
								RW_AFTER, data->after_rules[j])) <= 0)
						goto err_release_lock; 
				}
			}
		}
	}

#if 0 
	printk(KERN_DEBUG "woo, serializing finished, now adding marker\n"); 
#endif /* 0 */ 
	/* we now have in the mbuf the serialized lists of all rules */
	/* add a NULL byte, to mark the ending */ 
	if ((rc = sct_membuf_add_chunk(mbuf, &marker, sizeof(marker)) < 0)) {
		printk(KERN_DEBUG "adding the marker returned %d\n", rc); 
		retval = (rc != sizeof(marker) ? -ENOSPC : -EINVAL); 
		goto err_release_lock; 
	}

	/* viva la victor! */
	printk(KERN_DEBUG "joy and victory, get_all_rules finished succesfully!\n"); 
	retval = 0; 

 err_release_lock:
	up(&tracker_sem); 
 err_nolock:
	return retval; 
}
